//********************************************************************************
//
// A class used to define B spline curves.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Construct a B spline containing the given number of points.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE BasisSpline<storage, numElements>::BasisSpline(S32 numPoints) :
_numPoints(numPoints),
_pointsAdded(0),
_assembled(false)
{
	_pointIndices = new S32[_numPoints + 2];
	_points = new storage[_numPoints];
	_normals = new storage[_numPoints];
	_lengths = new F32[_numPoints];
}


//********************************************************************************
// Destroy a B spline.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE BasisSpline<storage, numElements>::~BasisSpline(void)
{
	delete[] _lengths;
	delete[] _normals;
	delete[] _points;
	delete[] _pointIndices;
}


//********************************************************************************
// Get the number of control points.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE S32 BasisSpline<storage, numElements>::NumControlPoints(void)
{
	return _pointsAdded;
}


//********************************************************************************
// Get a particular control point.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE storage BasisSpline<storage, numElements>::ControlPoint(S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex < _pointsAdded);

	return _points[pointIndex];
}


//********************************************************************************
// Get a particular control normal.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE storage BasisSpline<storage, numElements>::ControlNormal(S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex < _pointsAdded);

	return _normals[pointIndex];
}


//********************************************************************************
// Add a point to the spline, through which it will interpolate.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::ControlPoint(const storage& position, S32 pointIndex)
{
	_assembled = false;

	if (pointIndex >= 0)
	{
		LIBMATHS_ASSERT(pointIndex < _pointsAdded);

		_points[pointIndex] = position;
	}
	else
	{
		LIBMATHS_ASSERT(_pointsAdded < _numPoints);

		_points[_pointsAdded++] = position;
	}
}


//********************************************************************************
// Add a point and normal to the spline, through which it will interpolate.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::ControlPointNormal(const storage& position, const storage& normal, S32 pointIndex)
{
	_assembled = false;

	if (pointIndex >= 0)
	{
		LIBMATHS_ASSERT(pointIndex < _pointsAdded);

		_points[pointIndex] = position;
		_normals[pointIndex] = normal;
	}
	else
	{
		LIBMATHS_ASSERT(_pointsAdded < _numPoints);

		_points[_pointsAdded] = position;
		_normals[_pointsAdded++] = normal;
	}
}


//********************************************************************************
// Insert a point into the spline, through which it will interpolate.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::InsertControlPoint(const storage& position, S32 pointIndex)
{
	_assembled = false;

	LIBMATHS_ASSERT(pointIndex >= 0);
	LIBMATHS_ASSERT(pointIndex <= _pointsAdded);
	LIBMATHS_ASSERT(_pointsAdded < _numPoints);

	for (S32 i = _pointsAdded; i > pointIndex; i--)
	{
		_points[i] = _points[i - 1];
	}

	_points[pointIndex] = position;

	_pointsAdded++;
}


//********************************************************************************
// Insert a point and normal into the spline, through which it will interpolate.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::InsertControlPointNormal(const storage& position, const storage& normal, S32 pointIndex)
{
	_assembled = false;

	LIBMATHS_ASSERT(pointIndex >= 0);
	LIBMATHS_ASSERT(pointIndex <= _pointsAdded);
	LIBMATHS_ASSERT(_pointsAdded < _numPoints);

	for (S32 i = _pointsAdded; i > pointIndex; i--)
	{
		_points[i] = _points[i - 1];
		_normals[i] = _normals[i - 1];
	}

	_points[pointIndex] = position;
	_normals[pointIndex] = normal;

	_pointsAdded++;
}


//********************************************************************************
// Remove a particular control point.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::RemoveControlPoint(S32 pointIndex)
{
	if (pointIndex >= 0 &&
		pointIndex < _pointsAdded)
	{
		_pointsAdded--;

		for (S32 i = pointIndex; i < _pointsAdded; i++)
		{
			_points[i] = _points[i + 1];
		}

		_assembled = false;
	}
}


//********************************************************************************
// Remove a particular control point and normal.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::RemoveControlPointNormal(S32 pointIndex)
{
	if (pointIndex >= 0 &&
		pointIndex < _pointsAdded)
	{
		_pointsAdded--;

		for (S32 i = pointIndex; i < _pointsAdded; i++)
		{
			_points[i] = _points[i + 1];
			_normals[i] = _normals[i + 1];
		}

		_assembled = false;
	}
}


//********************************************************************************
// Remove all control points.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::RemoveControlPoints(void)
{
	_pointsAdded = 0;

	_assembled = false;
}


//********************************************************************************
// Linear interpolation of position along the spline by the given parameter.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE storage BasisSpline<storage, numElements>::InterpolatePoint(F32 t)
{
	storage position;

	InterpolatePoint(t, position);

	return position;
}


//********************************************************************************
// Linear interpolation of position along the spline by the given parameter.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::InterpolatePoint(F32 t, storage& position)
{
	if (_assembled == false)
	{
		Assemble();
	}

	S32 numPoints = VirtualPoints() - 3;
	F32 max = _lengths[_pointsAdded - 1];

	t = Clamp(t, 0.0f, max);

	for (S32 i = 0; i < numPoints; i++)
	{
		storage& p1 = _points[_pointIndices[i + 1]];
		storage& p2 = _points[_pointIndices[i + 2]];
		F32 l1 = _lengths[_pointIndices[i + 1]];
		F32 l2 = _lengths[_pointIndices[i + 2]];

		if (t >= l1 && t <= l2)
		{
			F32 t2 = (t - l1) / (l2 - l1);

			position = Evaluate(t2, _points[_pointIndices[i]], p1, p2, _points[_pointIndices[i + 3]]);

			return;
		}
	}

	position = storage::ZERO();
}


//********************************************************************************
// Linear interpolation of position and normal along the spline by the given
// parameter.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::InterpolatePointNormal(F32 t, storage& position, storage& normal)
{
	if (_assembled == false)
	{
		Assemble();
	}

	S32 numPoints = VirtualPoints() - 3;
	F32 max = _lengths[_pointsAdded - 1];

	t = Clamp(t, 0.0f, max);

	for (S32 i = 0; i < numPoints; i++)
	{
		storage& p1 = _points[_pointIndices[i + 1]];
		storage& p2 = _points[_pointIndices[i + 2]];
		F32 l1 = _lengths[_pointIndices[i + 1]];
		F32 l2 = _lengths[_pointIndices[i + 2]];

		if (t >= l1 && t <= l2)
		{
			F32 t2 = (t - l1) / (l2 - l1);

			position = Evaluate(t2, _points[_pointIndices[i]], p1, p2, _points[_pointIndices[i + 3]]);
			normal = Evaluate(t2, _normals[_pointIndices[i]], _normals[_pointIndices[i + 1]], _normals[_pointIndices[i + 2]], _normals[_pointIndices[i + 3]]);

			return;
		}
	}

	position = storage::ZERO();
	normal = storage::ZERO();
}


//********************************************************************************
// Return the approximate tangent to the spline at the point given by the supplied
// interpolation parameter.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE storage BasisSpline<storage, numElements>::InterpolateDirection(F32 t, F32 eps)
{
	storage direction;

	InterpolateDirection(t, direction);

	return direction;
}


//********************************************************************************
// Return the approximate tangent to the spline at the point given by the supplied
// interpolation parameter.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::InterpolateDirection(F32 t, storage& direction, F32 eps)
{
	storage v0;
	storage v1;

	t = Clamp(t, 0.0f, 1.0f);

	if (t >= 1.0f - eps)
	{
		InterpolatePoint(t, v0);
		InterpolatePoint(t - eps, v1);

		direction = (v0 - v1).Normalised();
	}
	else
	{
		InterpolatePoint(t, v0);
		InterpolatePoint(t + eps, v1);

		direction = (v1 - v0).Normalised();
	}
}


//********************************************************************************
// Evaluate a set of points to return a point at location t.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE storage BasisSpline<storage, numElements>::Evaluate(F32 t, storage& p0, storage& p1, storage& p2, storage& p3)
{
	F32 t2 = t * t;
	F32 t3 = t * t2;
	storage result(storage::ZERO());

	for (S32 i = 0; i < numElements; i++)
	{
		F32 a = -p0[i] + 3 * p1[i] - 3 * p2[i] + p3[i];
		F32 b = 3 * p0[i] - 6 * p1[i] + 3 * p2[i];
		F32 c = -3 * p0[i] + 3 * p2[i];
		F32 d = p0[i] + 4 * p1[i] + p2[i];

		result[i] = ((a * t3) + (b * t2) + (c * t) + d) * (1.0f / 6.0f);
	}

	return result;
}


//********************************************************************************
// Assemble the spline into a form that can be queried.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE void BasisSpline<storage, numElements>::Assemble(void)
{
	if (_assembled == false)
	{
		for (S32 i = 0; i < _pointsAdded; i++)
		{
			_pointIndices[i + 1] = i;
		}

		if (_points[0] == _points[_pointsAdded - 1])
		{
			// This is a closed spline so we'll try and loop it.

			_pointIndices[0] = _pointsAdded - 2;
			_pointIndices[_pointsAdded + 1] = 1;
		}
		else
		{
			// Replicate start point at start and end point at end.

			_pointIndices[0] = 0;
			_pointIndices[_pointsAdded + 1] = _pointsAdded - 1;
		}

		const S32 maxSegments = 64;
		F32 lengths[maxSegments];
		F32 totalLength = 0.0f;

		LIBMATHS_ASSERT(VirtualPoints() < maxSegments);

		for (S32 i = 0; i < VirtualPoints() - 3; i++)
		{
			storage& p1 = _points[_pointIndices[i + 1]];
			storage& p2 = _points[_pointIndices[i + 2]];
			storage mid = Evaluate(0.5f, _points[_pointIndices[i]], p1, p2, _points[_pointIndices[i + 3]]);

			// An approximate mid-point length calculation

			F32 distance = (mid - p1).Length() + (mid - p2).Length();

			lengths[i] = distance;

			totalLength += distance;
		}

		F32 sum = 0.0f;

		_lengths[0] = 0.0f;

		for (S32 i = 0; i < VirtualPoints() - 3; i++)
		{
			sum += lengths[i];

			_lengths[i + 1] = sum / totalLength;
		}

		_assembled = true;
	}
}


//********************************************************************************
// Copy an existing spline.
//********************************************************************************

template<typename storage, S32 numElements>
LIBMATHS_INLINE BasisSpline<storage, numElements>& BasisSpline<storage, numElements>::operator = (const BasisSpline<storage, numElements>& rhs)
{
	if (this != &rhs)
	{
		_numPoints = rhs._numPoints;
		_pointsAdded = rhs._pointsAdded;

		_pointIndices = new S32[_numPoints + 2];
		_points = new storage[_numPoints];
		_normals = new storage[_numPoints];
		_lengths = new F32[_numPoints];

		for (S32 i = 0; i < _pointsAdded; i++)
		{
			_points[i] = rhs._points[i];
			_normals[i] = rhs._normals[i];
			_lengths[i] = rhs._lengths[i];
		}

		_assembled = false;
	}

	return *this;
}
