//********************************************************************************
//
// A class used to define Bezier curves.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"
#include "vector3.h"
#include "vector4.h"


namespace libmaths
{

	//********************************************************************************
	// Multi dimensional Bezier spline.
	//
	// A Bezier spline is a local interpolating spline used to define curves.
	//
	// This implementation supports interpolation of position along the spline and
	// optionally a second interpolation which may represent the normal to the spline
	// at any given point.
	//
	// Each spline is a curved line between control points.
	//********************************************************************************

	template<typename storage>
	class BezierSpline
	{

	public:

		// Construct a Bezier spline containing the given number of points.
		BezierSpline(S32 numPoints);

		// Destroy a Bezier spline.
		~BezierSpline(void);

		// Get the number of control points.
		S32 NumControlPoints(void);

		// Get a particular control point.
		storage ControlPoint(S32 pointIndex);

		// Get a particular control normal.
		storage ControlNormal(S32 pointIndex);

		// Add a point to the spline, through which it will interpolate.
		void ControlPoint(const storage& position, S32 pointIndex = -1);

		// Add a point and normal to the spline, through which it will interpolate.
		void ControlPointNormal(const storage& position, const storage& normal, S32 pointIndex = -1);

		// Insert a point into the spline, through which it will interpolate.
		void InsertControlPoint(const storage& position, S32 pointIndex);

		// Insert a point and normal into the spline, through which it will interpolate.
		void InsertControlPointNormal(const storage& position, const storage& normal, S32 pointIndex);

		// Remove a particular control point.
		void RemoveControlPoint(S32 pointIndex);

		// Remove a particular control point and normal.
		void RemoveControlPointNormal(S32 pointIndex);

		// Remove all control points.
		void RemoveControlPoints(void);

		// Linear interpolation of position along the spline by the given parameter.
		storage InterpolatePoint(F32 t);

		// Linear interpolation of position along the spline by the given parameter.
		void InterpolatePoint(F32 t, storage& position);

		// Linear interpolation of position and Normal along the spline by the given parameter.
		void InterpolatePointNormal(F32 t, storage& position, storage& normal);

		// Return the approximate tangent to the spline at the point given by the supplied interpolation parameter.
		storage InterpolateDirection(F32 t, F32 eps = 0.0001f);

		// Return the approximate tangent to the spline at the point given by the supplied interpolation parameter.
		void InterpolateDirection(F32 t, storage& direction, F32 eps = 0.0001f);

		BezierSpline& operator = (const BezierSpline& rhs);

	private:

		// Private default constructor to prevent construction of uninitialised splines.
		BezierSpline(void);

		// Get the factorial for a number (< 32)
		F32 Factorial(S32 Number);

		// Raise a number to an integer power
		F32 Power(F32 value, S32 power);

		// The maximum number of points the spline can interpolate.
		S32 _numPoints;

		// The current number of points the spline can interpolate.
		S32 _pointsAdded;

		// Points the spline can interpolate.
		storage* _points;

		// Additional normals the spline can interpolate.
		storage* _normals;
	};

	// Now include the inline definitions of the class members

#include "bezierspline.inl"

	// Vector2 Bezier spline
	typedef BezierSpline<Vector2> BezierSplineVector2;
	// Vector3x Bezier spline
	typedef BezierSpline<Vector3x> BezierSplineVector3x;
	// Vector3 Bezier spline
	typedef BezierSpline<Vector3> BezierSplineVector3;
	// Vector4 Bezier spline
	typedef BezierSpline<Vector4> BezierSplineVector4;

// namespace libmaths
}

#pragma managed(pop)
