//********************************************************************************
//
// A class used to define Bezier curves.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


#pragma once


//********************************************************************************
// Construct a Bezier spline containing the given number of points.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE BezierSpline<storage>::BezierSpline(S32 numPoints) :
_numPoints(numPoints),
_pointsAdded(0)
{
	LIBMATHS_ASSERT(_numPoints <= 31);

	_points = new storage[numPoints];
	_normals = new storage[numPoints];
}


//********************************************************************************
// Destroy a Bezier spline.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE BezierSpline<storage>::~BezierSpline(void)
{
	delete[] _normals;
	delete[] _points;
}


//********************************************************************************
// Get the number of control points.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE S32 BezierSpline<storage>::NumControlPoints(void)
{
	return _pointsAdded;
}


//********************************************************************************
// Get a particular control point.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE storage BezierSpline<storage>::ControlPoint(S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex < _pointsAdded);

	return _points[pointIndex];
}


//********************************************************************************
// Get a particular control point.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE storage BezierSpline<storage>::ControlNormal(S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex < _pointsAdded);

	return _normals[pointIndex];
}


//********************************************************************************
// Add a point to the spline, through which it will interpolate.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::ControlPoint(const storage& position, S32 pointIndex)
{
	if (pointIndex >= 0)
	{
		LIBMATHS_ASSERT(pointIndex < _pointsAdded);

		_points[pointIndex] = position;
	}
	else
	{
		LIBMATHS_ASSERT(_pointsAdded < _numPoints);

		_points[_pointsAdded++] = position;
	}
}


//********************************************************************************
// Add a point and normal to the spline, through which it will interpolate.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::ControlPointNormal(const storage& position, const storage& normal, S32 pointIndex)
{
	if (pointIndex >= 0)
	{
		LIBMATHS_ASSERT(pointIndex < _pointsAdded);

		_points[pointIndex] = position;
		_normals[pointIndex] = normal;
	}
	else
	{
		LIBMATHS_ASSERT(_pointsAdded < _numPoints);

		_points[_pointsAdded] = position;
		_normals[_pointsAdded++] = normal;
	}
}


//********************************************************************************
// Insert a point into the spline, through which it will interpolate.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::InsertControlPoint(const storage& position, S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex >= 0);
	LIBMATHS_ASSERT(pointIndex <= _pointsAdded);
	LIBMATHS_ASSERT(_pointsAdded < _numPoints);

	for (S32 i = _pointsAdded; i > pointIndex; i--)
	{
		_points[i] = _points[i - 1];
	}

	_points[pointIndex] = position;

	_pointsAdded++;
}


//********************************************************************************
// Insert a point and normal into the spline, through which it will interpolate.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::InsertControlPointNormal(const storage& position, const storage& normal, S32 pointIndex)
{
	LIBMATHS_ASSERT(pointIndex >= 0);
	LIBMATHS_ASSERT(pointIndex <= _pointsAdded);
	LIBMATHS_ASSERT(_pointsAdded < _numPoints);

	for (S32 i = _pointsAdded; i > pointIndex; i--)
	{
		_points[i] = _points[i - 1];
		_normals[i] = _normals[i - 1];
	}

	_points[pointIndex] = position;
	_normals[pointIndex] = normal;

	_pointsAdded++;
}


//********************************************************************************
// Remove a particular control point.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::RemoveControlPoint(S32 pointIndex)
{
	if (pointIndex >= 0 &&
		pointIndex < _pointsAdded)
	{
		_pointsAdded--;

		for (S32 i = pointIndex; i < _pointsAdded; i++)
		{
			_points[i] = _points[i + 1];
		}
	}
}


//********************************************************************************
// Remove a particular control point and normal.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::RemoveControlPointNormal(S32 pointIndex)
{
	if (pointIndex >= 0 &&
		pointIndex < _pointsAdded)
	{
		_pointsAdded--;

		for (S32 i = pointIndex; i < _pointsAdded; i++)
		{
			_points[i] = _points[i + 1];
			_normals[i] = _normals[i + 1];
		}
	}
}


//********************************************************************************
// Remove all control points.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::RemoveControlPoints(void)
{
	_pointsAdded = 0;
}


//********************************************************************************
// Linear interpolation of position along the spline by the given parameter.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE storage BezierSpline<storage>::InterpolatePoint(F32 t)
{
	storage position;

	InterpolatePoint(t, position);

	return position;
}


//********************************************************************************
// Linear interpolation of position along the spline by the given parameter.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::InterpolatePoint(F32 t, storage& position)
{
	S32 i, j;
	F32 Temp;

	j = _pointsAdded - 1;

	position = storage::ZERO();

	t = Clamp(t, 0.0f, 1.0f);

	for (i = 0; i < (j + 1); i++)
	{
		Temp = Factorial(j);
		Temp /= Factorial(i);
		Temp /= Factorial(j - i);

		Temp *= Power(t, i);
		Temp *= Power(1.0f - t, j-i);

		position += _points[i] * Temp;
	}
}


//********************************************************************************
// Linear interpolation of position and Normal along the spline by the given
// parameter.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::InterpolatePointNormal(F32 t, storage& position, storage& normal)
{
	S32 i, j;
	F32 Temp;

	j = _pointsAdded - 1;

	position = storage::ZERO();
	normal = storage::ZERO();

	t = Clamp(t, 0.0f, 1.0f);

	for (i = 0; i < (j + 1); i++)
	{
		Temp = Factorial(j);
		Temp /= Factorial(i);
		Temp /= Factorial(j - i);

		Temp *= Power(t, i);
		Temp *= Power(1.0f - t, j-i);

		position += _points[i] * Temp;
		normal += _normals[i] * Temp;
	}
}


//********************************************************************************
// Return the approximate tangent to the spline at the point given by the supplied
// interpolation parameter.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE storage BezierSpline<storage>::InterpolateDirection(F32 t, F32 eps)
{
	storage direction;

	InterpolateDirection(t, direction);

	return direction;
}


//********************************************************************************
// Return the approximate tangent to the spline at the point given by the supplied
// interpolation parameter.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE void BezierSpline<storage>::InterpolateDirection(F32 t, storage& direction, F32 eps)
{
	storage v0;
	storage v1;

	t = Clamp(t, 0.0f, 1.0f);

	if (t >= 1.0f - eps)
	{
		InterpolatePoint(t, v0);
		InterpolatePoint(t - eps, v1);

		direction = (v0 - v1).Normalised();
	}
	else
	{
		InterpolatePoint(t, v0);
		InterpolatePoint(t + eps, v1);

		direction = (v1 - v0).Normalised();
	}
}


//********************************************************************************
// Copy an existing spline.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE BezierSpline<storage>& BezierSpline<storage>::operator = (const BezierSpline<storage>& rhs)
{
	if (this != &rhs)
	{
		_numPoints = rhs._numPoints;
		_pointsAdded = rhs._pointsAdded;

		_points = new storage[_numPoints];
		_normals = new storage[_numPoints];

		for (S32 i = 0; i < _pointsAdded; i++)
		{
			_points[i] = rhs._points[i];
			_normals[i] = rhs._normals[i];
		}
	}

	return *this;
}


//********************************************************************************
// Return a factorial.
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE F32 BezierSpline<storage>::Factorial(S32 Number)
{
	static F32 table[] =
	{
		1.0f, table[0] * 1, table[1] * 2, table[2] * 3,
		table[3] * 4, table[4] * 5, table[5] * 6, table[6] * 7,
		table[7] * 8, table[8] * 9, table[9] * 10, table[10] * 11,
		table[11] * 12, table[12] * 13, table[13] * 14, table[14] * 15,
		table[15] * 16, table[16] * 17, table[17] * 18, table[18] * 19,
		table[19] * 20, table[20] * 21, table[21] * 22, table[22] * 23,
		table[23] * 24, table[24] * 25, table[25] * 26, table[26] * 27,
		table[27] * 28, table[28] * 29, table[29] * 30, table[30] * 31,
	};

	const S32 maxTable = sizeof(table) / sizeof(table[0]);

	if (Number < maxTable)
	{
		return table[Number];
	}

	LIBMATHS_ASSERT(0);

	return 0.0f;
}


//********************************************************************************
// Return the first number raised to an integer power
//********************************************************************************

template<typename storage>
LIBMATHS_INLINE F32 BezierSpline<storage>::Power(F32 value, S32 power)
{
	F32 result = 1.0f;

	for (S32 i = 0; i < power; i++)
	{
		result *= value;
	}

	return result;
}
