//********************************************************************************
//
// Axis aligned 3D bounding boxes for collision checks.
//
// These can be tested for collisions with other boxes, points, spheres or rays
// through the game world.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "matrix44.h"
#include "sphere.h"
#include "triangle3.h"


namespace libmaths
{

	//********************************************************************************
	// Three dimensional bounding box (Vector3x components).
	//
	// Three dimensional bounding box formed from a centre point and an extent
	// representing the half width of the box in each axis.
	//
	// These can be tested for collisions with other boxes,
	// points, spheres or rays through the game world.
	//
	// The box centre and extent are stored in Vector3x format.
	//
	// Two additional 32-bit data items may also be stored along with the box
	// dimensions (GeAABoundingBox::Data0 etc.).
	//********************************************************************************

	class AxisAlignedBox
	{

	public:

		// Construct a bounding box (NOTE: No default values).
		AxisAlignedBox(void)
		{ }

		// Construct a bounding box with the given centre and extent.
		AxisAlignedBox(const Vector3x& centre, const Vector3x& extent) :
			_centre(centre),
			_extent(extent)
		{ }

		// Construct a bounding box around a sphere.
		AxisAlignedBox(const Sphere& sphere)
		{
			F32 extent = sphere.Radius();

			_centre = sphere.Point();
			_extent = Vector3x(extent, extent, extent);
		}

		// Construct a bounding box with the given centre and uniform extent radius in each axis.
		AxisAlignedBox(const Vector3x& centre, F32 radius) :
			_centre(centre),
			_extent(radius, radius, radius)
		{ }

		// Destroy a bounding box.
		~AxisAlignedBox(void)
		{ }

		// Set the bounding box centre.
		void Centre(const Vector3x& centre)
		{ _centre = centre; }

		// Set the bounding box extent.
		void Extent(const Vector3x& extent)
		{ _extent = extent; }

		// Return the bounding box centre.
		const Vector3x& Centre(void) const
		{ return _centre; }

		// Return the bounding box extent.
		const Vector3x& Extent(void) const
		{ return _extent; }

		// Sets the minimum and maximum corner points of the bounding box.
		void MinMax(const Vector3x& min, const Vector3x& max)
		{
			_extent = (max - min) * 0.5f;
			_centre = min + _extent;
		}

		// Return the minimum corner point of the bounding box.
		Vector3x Minimum(void) const
		{ return _centre - _extent; }

		// Return the maximum corner point of the bounding box.
		Vector3x Maximum(void) const
		{ return _centre + _extent; }

		// Return the component of the minimum corner point of the bounding box in the given axis.
		F32 Minimum(const S32 axis) const
		{ return _centre[axis] - _extent[axis]; }

		// Return the component of the maximum corner point of the bounding box in the given axis.
		F32 Maximum(const S32 axis) const
		{ return _centre[axis] + _extent[axis]; }

		// Set the bounding box to empty. i.e. enclosing no volume.
		void Empty(void)
		{
			_centre = Vector3x(0.0f, 0.0f, 0.0f);
			_extent = Vector3x(-1.0e20f, -1.0e20f, -1.0e20f);
		}

		// Build a bounding box surrounding the two given bounding boxes.
		void Surround(const AxisAlignedBox& b1, const AxisAlignedBox& b2)
		{
			Vector3x min, max;

			min.MinOf(b1.Minimum(), b2.Minimum());
			max.MaxOf(b1.Maximum(), b2.Maximum());

			MinMax(min, max);
		}

		// Extend the bounding box to additionally enclose the given point.
		void Add(const Vector3x& point)
		{
			Vector3x min, max;

			min.MinOf(Minimum(), point);
			max.MaxOf(Maximum(), point);

			MinMax(min, max);
		}

		// Extend the bounding box to additionally enclose the given bounding box.
		void Add(const AxisAlignedBox& box)
		{ Surround(*this, box); }

		// Extend the bounding box to additionally enclose the given triangle.
		void Add(const Triangle3& triangle)
		{
			Add(triangle[0]);
			Add(triangle[1]);
			Add(triangle[2]);
		}

		// Extend the bounding box to additionally enclose the given sphere.
		void Add(const Sphere& sphere)
		{ Surround(*this, AxisAlignedBox(sphere)); }

		// Return the maximum _extent of the bounding box in all axes.
		F32 Size(void) const
		{ return Max3(_extent[0], _extent[1], _extent[2]); }

		// Return the axis of the bounding box with the maximum _extent.
		S32 MaxAxis(void) const
		{
			F32 a[2];
			S32 axis = (a[0] = Abs(_extent[0])) < (a[1] = Abs(_extent[1])) ? 1 : 0;
			return a[axis] < Abs(_extent[2]) ? 2 : axis;
		}

		// Check if the bounding box intersects the given bounding box.
		Bool Intersects(const AxisAlignedBox& box) const
		{
			const Vector3x centre(Abs(_centre - box._centre));
			const Vector3x extent(_extent + box._extent);

			if (centre.X() > extent.X() ||
				centre.Y() > extent.Y() ||
				centre.Z() > extent.Z())
			{
				return false;
			}

			return true;
		}

		// Check if the bounding box intersects (contains) the given point.
		Bool Intersects(const Vector3x& point) const
		{
			Vector3x centre(Abs(_centre - point));

			if (centre.X() > _extent.X() ||
				centre.Y() > _extent.Y() ||
				centre.Z() > _extent.Z())
			{
				return false;
			}

			return true;
		}

		// Check if the bounding box intersects the given sphere.
		Bool Intersects(const Sphere& sphere) const
		{
			Vector3x centre(Abs(_centre - sphere.Point()));
			Vector3x extent(_extent + sphere.Extent());

			if (centre.X() > extent.X() ||
				centre.Y() > extent.Y() ||
				centre.Z() > extent.Z())
			{
				return false;
			}

			return true;
		}

		// Check if the bounding box contains a point in space.
		Bool Contains(const Vector3x& point) const
		{
			Vector3x topLeft(Minimum());
			Vector3x bottomRight(Maximum());

			return (point.X() >= topLeft.X() && point.X() <= bottomRight.X() &&
					point.Y() >= topLeft.Y() && point.Y() <= bottomRight.Y() &&
					point.Z() >= topLeft.Z() && point.Z() <= bottomRight.Z());
		}

		// Check if the line intersects this bounding box.
		// Does not allow intersections from within the bounding box.
		Bool Intersects(const Ray3& ray) const
		{
			Vector3x point, normal;

			return Intersects(ray, point, normal);
		}

		// Check if the line intersects this bounding box.
		// Does not allow intersections from within the bounding box.
		Bool Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const;

		// Check if the line intersects this bounding box.
		// Does not allow intersections from within the bounding box.
		Bool Intersects(const LineSegment3& line) const;

		// Check if the line intersects this bounding box.
		// Does not allow intersections from within the bounding box.
		Bool Intersects(const LineSegment3& line, Vector3x& point, Vector3x& normal, F32& t) const
		{
			if (Intersects(Ray3(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the volume of the box.
		F32 Volume(void)
		{ return (_extent.x * 2) * (_extent.y * 2) * (_extent.z * 2); }

		// Transform the bounding box with the given matrix
		AxisAlignedBox Transform(const Matrix44& matrix) const;

		// Retrieve the first extra integer data item stored with the bounding box.
		U32 Data0(void) const
		{ return _centre.WInt(); }

		// Set the first extra integer data item stored with the bounding box.
		void Data0(U32 data)
		{ _centre.WInt(data); }

		// Retrieve the second extra integer data item stored with the bounding box.
		U32 Data1(void) const
		{ return _extent.WInt(); }

		// Set the second extra integer data item stored with the bounding box.
		void Data1(U32 data)
		{ _extent.WInt(data); }

	private:

		// Centre of the bounding box.
		Vector3x _centre;
		
		// Extent of the bounding box in each axis.
		Vector3x _extent;
	};

// namespace libmaths
}

#pragma managed(pop)
