//********************************************************************************
//
// A class used to define Catmull Rom spline curves.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"
#include "vector3.h"
#include "vector4.h"


namespace libmaths
{

	//********************************************************************************
	// Multi dimensional Catmull Rom spline.
	//
	// A Catmull Rom spline is a local interpolating spline used to define curves.
	//
	// This implementation supports interpolation of position along the spline and
	// optionally a second interpolation which may represent the normal to the spline
	// at any given point.
	//
	// Each spline is a curved line between control points. Unlike many other types of
	// spline, the Catmull Rom curve passes through all the control points. The path
	// between each pair of points is calculated using a cubic function of just the
	// four surrounding points, so other points further along the spline do not affect
	// that part of the curve.
	//********************************************************************************

	template<typename storage, S32 numElements>
	class CatmullRomSpline
	{

	public:

		// Construct a Catmull Rom spline containing the given number of points.
		CatmullRomSpline(S32 numPoints);

		// Destroy a Catmull Rom spline.
		~CatmullRomSpline(void);

		// Get the number of control points.
		S32 NumControlPoints(void);

		// Get a particular control point.
		storage ControlPoint(S32 pointIndex);

		// Get a particular control normal.
		storage ControlNormal(S32 pointIndex);

		// Add a point to the spline, through which it will interpolate.
		void ControlPoint(const storage& position, S32 pointIndex = -1);

		// Add a point and normal to the spline, through which it will interpolate.
		void ControlPointNormal(const storage& position, const storage& normal, S32 pointIndex = -1);

		// Insert a point into the spline, through which it will interpolate.
		void InsertControlPoint(const storage& position, S32 pointIndex);

		// Insert a point and normal into the spline, through which it will interpolate.
		void InsertControlPointNormal(const storage& position, const storage& normal, S32 pointIndex);

		// Remove a particular control point.
		void RemoveControlPoint(S32 pointIndex);

		// Remove a particular control point and normal.
		void RemoveControlPointNormal(S32 pointIndex);

		// Remove all control points.
		void RemoveControlPoints(void);

		// Linear interpolation of position along the spline by the given parameter.
		storage InterpolatePoint(F32 t);

		// Linear interpolation of position along the spline by the given parameter.
		void InterpolatePoint(F32 t, storage& position);

		// Linear interpolation of position and normal along the spline by the given parameter.
		void InterpolatePointNormal(F32 t, storage& position, storage& normal);

		// Return the approximate tangent to the spline at the point given by the supplied interpolation parameter.
		storage InterpolateDirection(F32 t, F32 eps = 0.0001f);

		// Return the approximate tangent to the spline at the point given by the supplied interpolation parameter.
		void InterpolateDirection(F32 t, storage& direction, F32 eps = 0.0001f);

		// Get the length for a point on the spline.
		F32 Length(S32 index)
		{ LIBCORE_ASSERT(index >= 0 && index < _pointsAdded); if (_assembled == false) Assemble(); return _lengths[index]; }

		// Copy an existing spline.
		CatmullRomSpline& operator = (const CatmullRomSpline& rhs);

	private:

		// Private default constructor to prevent construction of uninitialised splines.
		CatmullRomSpline(void);

		// Evaluate a set of points to return a point at location t.
		storage Evaluate(F32 t, storage& p0, storage& p1, storage& p2, storage& p3);

		// Assemble the spline into a form that can be queried.
		void Assemble(void);

		// Get the number of "virtual" points, always 2 more than the number of real points.
		S32 VirtualPoints(void)
		{ return _pointsAdded + 2; }

		// The maximum number of points the spline can interpolate.
		S32 _numPoints;

		// The current number of points the spline can interpolate.
		S32 _pointsAdded;

		// Indices to points the spline can interpolate.
		S32* _pointIndices;

		// Points the spline can interpolate.
		storage* _points;

		// Additional normals the spline can interpolate.
		storage* _normals;

		// Lengths for each point along the spline.
		F32* _lengths;

		// Has the spline information been assembled ready for use?
		Bool _assembled;
	};

	// Now include the inline definitions of the class members

#include "catmullromspline.inl"

	// Vector2 Catmull Rom spline
	typedef CatmullRomSpline<Vector2, 2> CatmullRomSplineVector2;
	// Vector3x Catmull Rom spline
	typedef CatmullRomSpline<Vector3x, 3> CatmullRomSplineVector3x;
	// Vector3 Catmull Rom spline
	typedef CatmullRomSpline<Vector3, 3> CatmullRomSplineVector3;
	// Vector4 Catmull Rom spline
	typedef CatmullRomSpline<Vector4, 4> CatmullRomSplineVector4;

// namespace libmaths
}

#pragma managed(pop)
