//********************************************************************************
//
// 2D circles.
//
// Circles can detect overlaps with other circles, points and line segments.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"
#include "linesegment2.h"
#include "ray2.h"


namespace libmaths
{

	class AxisAlignedRectangle;

	//********************************************************************************
	// A 2D circle.
	//********************************************************************************

	class Circle
	{

	public:

		// Construct an empty circle (enclosing no volume)
		Circle(void) :
			_centre(0, 0),
			_radius(0)
		 { }

		// Construct circle with the given centre and radius.
		Circle(const F32 x, const F32 y, const F32 radius) :
			_centre(x, y),
			_radius(radius)
		{ }

		// Construct circle with the given centre and radius.
		Circle(const Vector2& centre, const F32 radius) :
			_centre(centre),
			_radius(radius)
		{ }

		// Construct circle surrounding the given rectangle.
		Circle(const AxisAlignedRectangle& rectangle);

		// Destroy circle.
		~Circle(void)
		{ }

		// Empty the circle such that it contains no volume.
		void Empty(void)
		{ _centre.XY(0, 0); _radius = -Epsilon; }

		// Build circle enclosing the two given circles.
		void Surround(const Circle& circle0, const Circle& circle1);

		// Build circle enclosing the given circle and rectangle.
		void Surround(const Circle& circle, const AxisAlignedRectangle& rectangle);

		// Extend the circle to additionally enclose the given circle.
		void Add(const Circle& circle);

		// Extend the circle to additionally enclose the given rectangle.
		void Add(const AxisAlignedRectangle& rectangle);

		// Return the radius of the circle.
		const F32 Radius(void) const
		{ return _radius; }

		// Set the radius of the circle.
		void Radius(F32 radius)
		{ _radius = radius; }

		// Return the radius squared of the circle.
		const F32 RadiusSqr(void) const
		{ return _radius * _radius; }

		// Return the centre of the circle.
		const Vector2& Point(void) const
		{ return _centre; }

		// Set the centre of the circle.
		void Point(const Vector2& point)
		{ _centre = point; }

		// Move a circle in space, adding an offset to its current position.
		void Translate(const Vector2& offset)
		{ _centre += offset; }

		// Return a vector representing the uniform extent of the circle in all axes.
		const Vector2 Extent(void) const
		{ return Vector2(_radius, _radius); }

		// Get the extent of the circle.
		void Extent(Vector2& min, Vector2& max) const
		{
			Vector2 r(Extent());

			min = Point() - r;
			max = Point() + r;
		}

		// Check if the circle contains a point in space.
		Bool Contains(const Vector2& point) const
		{ return ((point - Point()).LengthSqr() <= RadiusSqr()); }

		// Check if the circle intersects the given circle.
		Bool Intersects(const Circle& circle) const
		{
			Vector2 diff = _centre - circle._centre;
			F32 radii = Radius() + circle.Radius();

			return (diff.LengthSqr() < radii * radii);
		}

		// Check if the ray intersects this circle.
		// Does not allow intersections from within the circle.
		Bool Intersects(const Ray2& ray) const;

		// Check if the ray intersects this circle.
		// Does not allow intersections from within the circle.
		Bool Intersects(const Ray2& ray, Vector2& point, Vector2& normal) const;

		// Check if the line intersects this circle.
		// Does not allow intersections from within the circle.
		Bool Intersects(const LineSegment2& line) const
		{
			Vector2 point, normal;

			if (Intersects(Ray2(line), point, normal))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this circle.
		// Does not allow intersections from within the circle.
		Bool Intersects(const LineSegment2& line, Vector2& point, Vector2& normal, F32& t) const
		{
			if (Intersects(Ray2(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the area of the circle.
		F32 Area(void)
		{ return PI * RadiusSqr(); }

	private:

		// The centre / radius of the circle.
		Vector2 _centre;
		F32 _radius;
	};

// namespace libmaths
}

#pragma managed(pop)
