//********************************************************************************
//
// 3D bounding cylinders for collision checks.
//
// Bounding cylinders can detect overlaps with other cylinders, points and
// bounding spheres.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"
#include "ray3.h"
#include "linesegment3.h"


namespace libmaths
{

	//********************************************************************************
	// A bounding cylinders.
	//
	// Supports intersection checks, bounding of point sets, boxes and other cylinders.
	//********************************************************************************

	class Cylinder
	{

	public:

		// Construct an empty bounding cylinder (enclosing no volume)
		Cylinder(void)
		{
			_endPoints[0] = Vector3x::ZERO();
			_endPoints[1] = Vector3x::ZERO();
			_endPoints[0].W(-Epsilon);
		}

		// Construct bounding cylinder with the given centre and radius.
		Cylinder(const Vector3x& p0, const Vector3x& p1, const F32 radius)
		{
			_endPoints[0] = p0;
			_endPoints[1] = p1;
			_endPoints[0].W(radius);
		}

		// Destroy bounding cylinder.
		~Cylinder(void)
		{ }

		// Empty the bounding cylinder such that it contains no volume.
		void Empty(void)
		{ _endPoints[0].XYZ(0, 0, 0); _endPoints[1].XYZ(0, 0, 0); _endPoints[0].W(-Epsilon); }

		// Return the radius of the bounding cylinder.
		F32 Radius(void) const
		{ return _endPoints[0].W(); }

		// Set the radius of the bounding cylinder.
		void Radius(F32 radius)
		{ _endPoints[0].W(radius); }

		// Return the radius squared of the bounding cylinder.
		F32 RadiusSqr(void) const
		{ return _endPoints[0].W() * _endPoints[0].W(); }

		// Return the length of the bounding cylinder.
		F32 Length(void) const
		{ return (_endPoints[1] - _endPoints[0]).Length(); }

		// Return an end point of the bounding cylinder.
		const Vector3x& Point(S32 index) const
		{ LIBMATHS_ASSERT(index >= 0 && index < 2); return _endPoints[index]; }

		// Set an end point of the bounding cylinder.
		void Point(S32 index, const Vector3x& point)
		{ LIBMATHS_ASSERT(index >= 0 && index < 2); _endPoints[index].XYZ(point.X(), point.Y(), point.Z()); }

		// Move a bounding cylinder in space, adding an offset to its current position.
		void Translate(const Vector3x& offset)
		{ _endPoints[0] += offset; _endPoints[1] += offset; }

		// Get the extent of the bounding cylinder.
		void Extent(Vector3x& min, Vector3x& max) const
		{
			// TODO: This extent is a little too large.

			Vector3x radius = Vector3x(Radius(), Radius(), Radius());

			min = Vector3x(+FLOAT_CMAX, +FLOAT_CMAX, +FLOAT_CMAX);
			max = Vector3x(-FLOAT_CMAX, -FLOAT_CMAX, -FLOAT_CMAX);

			min.MinOf(Point(0), Point(1)) - radius;
			max.MaxOf(Point(0), Point(1)) + radius;
		}

		// Get the distance of the cylinder from a point in space.
		F32 Distance(const Vector3x& point, F32& pointOnLine) const;

		// Get the squared distance of the cylinder from a point in space.
		F32 DistanceSqr(const Vector3x& point, F32& pointOnLine) const;

		// Check if the cylinder contains a point in space.
		Bool Contains(const Vector3x& point) const;

		// Does the cylinder intersect a cylinder?
		Bool Intersects(const Cylinder& cylinder) const;

		// Does the cylinder intersect a cylinder?
		Bool Intersects(const Cylinder& cylinder, Vector3x& ip0, Vector3x& ip1) const;

		// Check if the ray intersects this cylinder.
		// Does not allow intersections from within the cylinder.
		Bool Intersects(const Ray3& ray) const;

		// Check if the ray intersects this cylinder.
		// Does not allow intersections from within the cylinder.
		Bool Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const;

		// Check if the line intersects this cylinder.
		// Does not allow intersections from within the cylinder.
		Bool Intersects(const LineSegment3& line) const
		{
			Vector3x point, normal;

			if (Intersects(Ray3(line), point, normal))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this cylinder.
		// Does not allow intersections from within the cylinder.
		Bool Intersects(const LineSegment3& line, Vector3x& point, Vector3x& normal, F32& t) const
		{
			if (Intersects(Ray3(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the volume of the cylinder.
		F32 Volume(void)
		{ return PI * RadiusSqr() * Length(); }

	private:

		// The end points of the bounding cylinder.
		Vector3x _endPoints[2];
	};

// namespace libmaths
}

#pragma managed(pop)
