//********************************************************************************
//
// Some fixed point maths classes.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"


//********************************************************************************
// Templated fixed point math class. Implementations in 8.8, 4.12, 1.15, 16.16 and
// 24.8 are provided by default.
//********************************************************************************

namespace libmaths
{

	template <typename storage, S32 multiplier, U32 shift>
	class FixedPoint
	{

	public:

		// Construct from no value, defaulting to 0
		FixedPoint<storage, multiplier, shift>(void)
		{ _value = 0; }

		// Construct from a 16 bit integer
		FixedPoint<storage, multiplier, shift>(S16 rhs)
		{ _value = storage(rhs); }

		// Construct from a 16 bit integer
		FixedPoint<storage, multiplier, shift>(U16 rhs)
		{ _value = storage(rhs); }

		// Construct from a 32 bit integer
		FixedPoint<storage, multiplier, shift>(S32 rhs)
		{ _value = storage(rhs); }

		// Construct from a 32 bit integer
		FixedPoint<storage, multiplier, shift>(U32 rhs)
		{ _value = storage(rhs); }

		// Construct from a floating point value
		FixedPoint<storage, multiplier, shift>(F32 rhs)
		{ _value = storage(FToI(rhs * multiplier)); }

		// Return a floating point value, accurately representing the fixed point value
		operator F32() const
		{ return F32(_value) * (1.0f / multiplier); }

		// Return a 32 bit integer that is just a straight copy (including the fraction) of the fixed point value
		operator S32() const
		{ return S32(_value); }

		// Return a 32 bit integer that is just a straight copy (including the fraction) of the fixed point value
		operator U32() const
		{ return U32(_value); }

		// Return a 32 bit integer that is just the integer part of the fixed point value
		S32 Integer(void) const
		{ return S32(_value >> shift); }

		// Assign from a 16 bit integer
		FixedPoint<storage, multiplier, shift>& operator = (const S16 rhs)
		{ _value = storage(rhs); return *this; }

		// Assign from a 32 bit integer
		FixedPoint<storage, multiplier, shift>& operator = (const S32 rhs)
		{ _value = storage(rhs); return *this; }

		// Assign from a floating point value
		FixedPoint<storage, multiplier, shift>& operator = (const F32 rhs)
		{ _value = storage(FToI(rhs * multiplier)); return *this; }

		// += a floating point value
		FixedPoint<storage, multiplier, shift>& operator += (const F32 rhs)
		{ _value = storage((_value + FToI(rhs * multiplier))); return *this; }

		// -= a floating point value
		FixedPoint<storage, multiplier, shift>& operator -= (const F32 rhs)
		{ _value = storage((_value - FToI(rhs * multiplier))); return *this; }

		// *= a floating point value
		FixedPoint<storage, multiplier, shift>& operator *= (const F32 rhs)
		{ _value = storage((S32(_value) * FToI(rhs * multiplier)) >> shift); return *this; }

		// /= a floating point value
		FixedPoint<storage, multiplier, shift>& operator /= (const F32 rhs)
		{ _value = storage((S32(_value) << shift) / FToI(rhs * multiplier)); return *this; }

		// += another fixed point type
		FixedPoint<storage, multiplier, shift>& operator += (const FixedPoint<storage, multiplier, shift> rhs)
		{ _value = storage(_value + rhs._value); return *this; }

		// -= another fixed point type
		FixedPoint<storage, multiplier, shift>& operator -= (const FixedPoint<storage, multiplier, shift> rhs)
		{ _value = storage(_value - rhs._value); return *this; }

		// *= another fixed point type
		FixedPoint<storage, multiplier, shift>& operator *= (const FixedPoint<storage, multiplier, shift> rhs)
		{ _value = storage((S32(_value) * (S32)rhs._value) >> shift); return *this; }

		// /= another fixed point type
		FixedPoint<storage, multiplier, shift>& operator /= (const FixedPoint<storage, multiplier, shift> rhs)
		{ _value = storage((S32(_value) << shift) / rhs._value); return *this; }

		// + a floating point value
		FixedPoint<storage, multiplier, shift> operator + (const F32 rhs) const
		{ return FixedPoint<storage, multiplier, shift>((_value + FToI(rhs * multiplier))); }

		// - a floating point value
		FixedPoint<storage, multiplier, shift> operator - (const F32 rhs) const
		{ return FixedPoint<storage, multiplier, shift>((_value - FToI(rhs * multiplier))); }

		// * a floating point value
		FixedPoint<storage, multiplier, shift> operator * (const F32 rhs) const
		{ return FixedPoint<storage, multiplier, shift>((S32(_value) * FToI(rhs * multiplier)) >> shift); }

		// / a floating point value
		FixedPoint<storage, multiplier, shift> operator / (const F32 rhs) const
		{ return FixedPoint<storage, multiplier, shift>((S32(_value) << shift) / FToI(rhs * multiplier)); }

		// + another fixed point type
		FixedPoint<storage, multiplier, shift> operator + (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return FixedPoint<storage, multiplier, shift>(_value + rhs._value); }

		// - another fixed point type
		FixedPoint<storage, multiplier, shift> operator - (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return FixedPoint<storage, multiplier, shift>(_value - rhs._value); }

		// * another fixed point type
		FixedPoint<storage, multiplier, shift> operator * (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return FixedPoint<storage, multiplier, shift>((S32(_value) * (S32)rhs._value) >> shift); }

		// / another fixed point type
		FixedPoint<storage, multiplier, shift> operator / (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return FixedPoint<storage, multiplier, shift>((S32(_value) << shift) / rhs._value); }

		// < comparison
		Bool operator < (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value < rhs._value); }

		// > comparison
		Bool operator > (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value > rhs._value); }

		// <= comparison
		Bool operator <= (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value <= rhs._value); }

		// >= comparison
		Bool operator >= (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value >= rhs._value); }

		// == comparison
		Bool operator == (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value == rhs._value); }

		// != comparison
		Bool operator != (const FixedPoint<storage, multiplier, shift> rhs) const
		{ return (_value != rhs._value); }

	private:

		storage _value;
	};

	// 8.8 fixed point type
	typedef FixedPoint<S16, 256, 8> Fixed8p8;

	// 4.12 fixed point type
	typedef FixedPoint<S16, 4096, 12> Fixed4p12;

	// 1.15 fixed point type
	typedef FixedPoint<S16, 32768, 15> Fixed1p15;

	// 16.16 fixed point type
	typedef FixedPoint<S32, 65536, 16> Fixed16p16;

	// 24.8 fixed point type
	typedef FixedPoint<S32, 256, 8> Fixed24p8;

// namespace libmaths
}

#pragma managed(pop)
