//********************************************************************************
//
// Frustum - 5 to 6 sided viewing volume normally associated with a camera.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "sphere.h"
#include "matrix44.h"


namespace libmaths
{

	class Frustum
	{

	public:

		// Indices for the frustum planes.
		enum GeFrustumPlanes
		{
			FRUSTUM_LEFT,
			FRUSTUM_RIGHT,
			FRUSTUM_TOP,
			FRUSTUM_BOTTOM,
			FRUSTUM_FRONT,
			FRUSTUM_BACK,

			NUM_FRUSTUM_PLANES
		};

		// Construct an empty frustum
		Frustum(void) :
			_numPlanes(NUM_FRUSTUM_PLANES),
			_endNormal(&_normals[NUM_FRUSTUM_PLANES])
		{ }

		// Get the number of planes in the frustum.
		S32 NumPlanes(void) const
		{ return _numPlanes; }

		// Set the number of planes in the frustum.
		void NumPlanes(S32 numPlanes)
		{
			LIBMATHS_ASSERT(numPlanes >= FRUSTUM_BACK && numPlanes <= NUM_FRUSTUM_PLANES);

			_numPlanes = numPlanes;
			_endNormal = &_normals[_numPlanes];
		}

		// Get the position of a plane.
		const Vector3x& Position(S32 index) const
		{ LIBMATHS_ASSERT(index >= 0 && index < _numPlanes); return _positions[index]; }

		// Set the position of a plane.
		void Position(S32 index, Vector3x position)
		{ LIBMATHS_ASSERT(index >= 0 && index < _numPlanes); _positions[index] = position; }

		// Get the normal of a plane.
		const Vector3x& Normal(S32 index) const
		{ LIBMATHS_ASSERT(index >= 0 && index < _numPlanes); return _normals[index]; }

		// Set the normal of a plane.
		void Normal(S32 index, Vector3x normal)
		{ LIBMATHS_ASSERT(index >= 0 && index < _numPlanes); _normals[index] = normal; }

		// Does the frustum contain a point?
		Bool Contains(const Vector3x& point) const;

		// Does the frustum contain a sphere?
		Bool Contains(const Sphere& sphere) const;

		// Do the frustum planes intersect a sphere?
		Bool Intersects(const Sphere& sphere) const;

		// Transform a frustum by a matrix.
		void TransformFrustum(const Frustum& source, const Matrix44& transform);

		// Setup a frustum from a field of view and near / far distances.
		void FromFOVDistances(F32 verticalFOV, F32 aspectRatio, F32 nearDistance, F32 farDistance);

	private:

		// The normals for the frustum planes.
		Vector3x _normals[NUM_FRUSTUM_PLANES];

		// The positions of the frustum planes.
		Vector3x _positions[NUM_FRUSTUM_PLANES];

		// Pointer to the last normal.
		Vector3x* _endNormal;

		// The number of frustum planes used.
		S32 _numPlanes;
	};

	// Now include the inline definitions of the structure members

#include "frustum.inl"

// namespace libmaths
}

#pragma managed(pop)
