//********************************************************************************
//
// Frustum - 5 to 6 sided viewing volume normally associated with a camera.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Does the frustum contain a point?
//********************************************************************************

LIBMATHS_INLINE Bool Frustum::Contains(const libmaths::Vector3x& point) const
{
	const libmaths::Vector3x* pThisNormal = _normals;
	const libmaths::Vector3x* pThisPosition = _positions;

	for (; pThisNormal < _endNormal; pThisNormal++, pThisPosition++)
	{
		if (libmaths::Vector3x(point - (*pThisPosition)) * (*pThisNormal) <= 0.0f)
		{
			return false;
		}
	}

	return true;
}


//********************************************************************************
// Does the frustum contain a sphere?
//********************************************************************************

LIBMATHS_INLINE Bool Frustum::Contains(const libmaths::Sphere& sphere) const
{
	const libmaths::Vector3x* pThisNormal = _normals;
	const libmaths::Vector3x* pThisPosition = _positions;
	const libmaths::Vector3x position(sphere.Point());
	F32 radius(-sphere.Radius());

	for (; pThisNormal < _endNormal; pThisNormal++, pThisPosition++)
	{
		if (libmaths::Vector3x(position - (*pThisPosition)) * (*pThisNormal) <= radius)
		{
			return false;
		}
	}

	return true;
}


//********************************************************************************
// Do the frustum planes intersect a sphere?
//********************************************************************************

LIBMATHS_INLINE Bool Frustum::Intersects(const libmaths::Sphere& sphere) const
{
	const libmaths::Vector3x* pThisNormal = _normals;
	const libmaths::Vector3x* pThisPosition = _positions;
	const libmaths::Vector3x position(sphere.Point());

	for (; pThisNormal < _endNormal; pThisNormal++, pThisPosition++)
	{
		if (libmaths::Vector3x(position - (*pThisPosition)) * (*pThisNormal) < sphere.Radius())
		{
			return true;
		}
	}

	return false;
}


//********************************************************************************
// Transform a frustum by a matrix.
//********************************************************************************

LIBMATHS_INLINE void Frustum::TransformFrustum(const Frustum& source, const libmaths::Matrix44& transform)
{
	_numPlanes = source.NumPlanes();
	_endNormal = &_normals[_numPlanes];

	for (S32 i = 0; i < _numPlanes; i++)
	{
		Normal(i, transform.TransformVector(source.Normal(i)));
		Position(i, transform.TransformPoint(source.Position(i)));
	}
}
