//********************************************************************************
//
// Core maths implementation.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "configuration.h"

#define LIBMATHS_STD std
#define LIBMATHS_ASSERT(x) _ASSERTE(x)
#define LIBMATHS_UNUSED(x) (void)x

#ifdef _WIN32
#define LIBMATHS_WIN32_ANY 1
#endif

#if !LIBMATHS_GENERIC

#if (defined(_WIN32) && defined(_MSC_VER) && (_M_IX86 > 400))
#define LIBMATHS_WIN32_MSVC 1
#define LIBMATHS_32BIT 1
#define LIBMATHS_64BIT 0
#endif

#if (defined(_WIN32) && defined(_MSC_VER) && (_M_X64 > 100))
#define LIBMATHS_WIN64_MSVC 1
#define LIBMATHS_32BIT 0
#define LIBMATHS_64BIT 1
#endif

#endif // !LIBMATHS_GENERIC


// Inlining and calling convention macros.

#if LIBMATHS_WIN32_ANY
	#define LIBMATHS_FASTCALL __fastcall
	#define LIBMATHS_FORCEINLINE __forceinline
	#define LIBMATHS_INLINE __inline
	#pragma warning(disable:4561)
#else
	#define LIBMATHS_FASTCALL
	#define LIBMATHS_FORCEINLINE inline
	#define LIBMATHS_INLINE inline
#endif // LIBMATHS_WIN32_ANY

#include <cmath>

#include "basictypes.h"

#if LIBMATHS_WIN32_MSVC
	#include <cfloat>
#endif

#if LIBMATHS_WIN32_ANY
	#include <crtdbg.h>
#endif


namespace libmaths
{
	// Useful constants

	// The ubiquitous pi (3.141592654) - the ratio of the circumference of a circle to its diameter.
	const F32 PI = 3.141592654f;

	// pi/180 (0.017453292) The number of radians in a degree.
	const F32 PIO180 = PI/180.0f;

	// 180/pi (57.29577951) The number of degrees in a radian.
	const F32 ONE80OPI = 180.0f/PI;

	// 2*pi (6.283185307) The number of radians all round a circle.
	const F32 TWOPI = 2.0f*PI;

	// pi/2 (1.570796327) The number of radians in a quarter circle (90 degrees).
	const F32 PIO2 = PI/2.0f;

	// 2/pi (0.636619772).
	const F32 TWOOPI = 2.0f/PI;

	// Exponential constant e (2.718281828).
	const F32 EXPONENTIAL = 2.718281828f;

	// sqrt(2) (1.414213562) The diagonal length of a unit square.
	const F32 SQRT2 = 1.414213562f;

	// sqrt(3) (1.732050808).
	const F32 SQRT3 = 1.732050808f;

	// Smallest value such that 1+epsilon != 1.
	const F32 FLOAT_EPSILON = 1.19209290e-07f;

	// Smallest value that can be safely normalised.
	const F32 FLOAT_NORMALISE_EPSILON = 1.0e-35f;

	// Smallest representable single precision floating point value.
	const F32 FLOAT_MIN = 1.17549435e-38f;

	// Largest representable single precision floating point value.
	const F32 FLOAT_MAX = 3.40282346e+38f;

	// Large number for floating point maximum comparisons without precision problems
	const F32 FLOAT_CMAX = 1.e+10f;

	// A more generous epsilon value to use in floating point comparisons.
	const F32 Epsilon = 3*FLOAT_EPSILON;

	// IEEE floating point "Not a Number" bit mask.
	const U32 NaNMask = (255L<<23L);

	// IEEE floating point "Not a Number" bit mask.
	const F32 NaN = *((F32*)(&NaNMask));

	// Convert f from F32 to signed integer. Always rounds towards 0.
	// (i.e. equivalent to an integer cast).
	S32 LIBMATHS_FASTCALL FToI(const F32 f);

	// Return the sine of f.
	F32 LIBMATHS_FASTCALL Sin(F32 f);

	// Return the cosine of f.
	F32 LIBMATHS_FASTCALL Cos(F32 f);

	// Set c to the cosine of f, s to the sine of f.
	void LIBMATHS_FASTCALL CosSin(F32 f, F32& c, F32& s);

	// Set c to the cosine of f, s to the sine of f.
	void LIBMATHS_FASTCALL SinCos(F32 f, F32& s, F32& c);

	// Return the tangent of f. f must be in the range -pi/2 to +pi/2.
	F32 LIBMATHS_FASTCALL Tan(F32 f);

	// Return the arc-sine of f.
	F32 LIBMATHS_FASTCALL Asin(F32 f);

	// Return the arc-cosine of f.
	F32 LIBMATHS_FASTCALL Acos(F32 f);

	// Return the arc-tangent of f.
	F32 LIBMATHS_FASTCALL Atan(F32 f);

	// Return the arc-tangent of y/x.
	F32 LIBMATHS_FASTCALL Atan2(F32 y, F32 x);

	// Return the fast arc-tangent of y/x.
	// The max error is a little less than 0.07 rads (only at a few angles though).
	F32 LIBMATHS_FASTCALL Atan2Fast(F32 y, F32 x);

	// Return f raised to the power of f.
	F32 LIBMATHS_FASTCALL Exp(F32 f);

	// Return the square root of f.
	F32 LIBMATHS_FASTCALL Sqrt(F32 f);

	// Return the fast square root of f. (Up to 0.2% under, never over).
	// NB. Test the speed vs. normal Sqrt on your target platform, this NEEDS to be inlined.
	F32 LIBMATHS_FASTCALL SqrtFast(F32 f);

	// Return the super fast square root of f, but inaccurately. (Up to 6% over, never under).
	// NB. Test the speed vs. normal Sqrt on your target platform, this NEEDS to be inlined.
	F32 LIBMATHS_FASTCALL SqrtSuperFast(F32 f);

	// Return the inverse square root of f
	F32 LIBMATHS_FASTCALL InvSqrt(F32 f);

	// Return the fast inverse square root of f
	// NB. Test the speed vs. normal InvSqrt on your target platform, this NEEDS to be inlined.
	F32 LIBMATHS_FASTCALL InvSqrtFast(F32 f);

	// Return the super fast inverse square root of f
	// NB. Test the speed vs. normal InvSqrt on your target platform, this NEEDS to be inlined.
	F32 LIBMATHS_FASTCALL InvSqrtSuperFast(F32 f);

	// Return the absolute value of f.
	F32 LIBMATHS_FASTCALL Abs(F32 f);

	// Return the integer value immediately less than f.
	S32 LIBMATHS_FASTCALL Floor(F32 f);

	// Return the integer value immediately less than f.
	F32 LIBMATHS_FASTCALL Floorf(F32 f);

	// Return the integer value immediately greater than f.
	S32 LIBMATHS_FASTCALL Ceil(F32 f);

	// Return the integer value immediately greater than f.
	F32 LIBMATHS_FASTCALL Ceilf(F32 f);

	// Return the integer value closest to f.
	S32 LIBMATHS_FASTCALL Round(F32 f);

	// Return the remainder of x/y.
	F32 LIBMATHS_FASTCALL Fmod(F32 x, F32 y);

	// Return the sign of f, or 0 if f is zero.
	F32 LIBMATHS_FASTCALL Zsgn(F32 f);

	// Return the sign of f. If f is zero then return +1.
	F32 LIBMATHS_FASTCALL Sgn(F32 f);

	// Convert the angle a from degrees to radians.
	F32 LIBMATHS_FASTCALL DegToRad(F32 a);

	// Convert the angle a from radians to degrees.
	F32 LIBMATHS_FASTCALL RadToDeg(F32 a);

	// Return whether f is approximately zero.
	Bool LIBMATHS_FASTCALL Eqz(F32 f);

	// Return whether f is approximately zero by comparison with e.
	Bool LIBMATHS_FASTCALL Eqz(F32 f, F32 e);

	// Return whether a is approximately equal to b.
	Bool LIBMATHS_FASTCALL Approx(F32 a, F32 b);

	// Return whether a is approximately equal to b withing the range of e.
	Bool LIBMATHS_FASTCALL Approx(F32 a, F32 b, F32 e);

	// Return f*f. (Should only be used for scalar types).
	template <typename T> LIBMATHS_INLINE T Square(T f);

	// Return the minimum of a and b.
	template <typename T> LIBMATHS_INLINE T Min(T a, T b);

	// Return the minimum of a, b and c.
	template <typename T> LIBMATHS_INLINE T Min3(T a, T b, T c);

	// Return the minimum of a, b, c and d.
	template <typename T> LIBMATHS_INLINE T Min3(T a, T b, T c, T d);

	// Return the maximum of a and b.
	template <typename T> LIBMATHS_INLINE T Max(T a, T b);

	// Return the maximum of a, b and c.
	template <typename T> LIBMATHS_INLINE T Max3(T a, T b, T c);

	// Return the maximum of a, b, c and d.
	template <typename T> LIBMATHS_INLINE T Max3(T a, T b, T c, T d);

	// Swap the values of a and b. (May incur a performance hit for large structures).
	template <typename T> LIBMATHS_INLINE void Swap(T& a, T& b);

	// Linear interpolation from l to h by an amount f (0 to 1).
	template <typename T> LIBMATHS_INLINE T Lerp(F32 f, T l, T h);

	// Clamp v to the range l to h.
	template <typename T> LIBMATHS_INLINE T Clamp(T v, T l, T h);

	// Reduce angle a to the range -pi * 2 to +pi * 2.
	F32 ReduceAngle(F32 a);

	// Reduce angle a to the range -pi to +pi.
	F32 ReduceAnglePI(F32 a);

	// Reduce angle a to the range 0 to +pi.
	F32 ReduceAnglePI2(F32 a);

	// Get the difference between angle a and angle b.
	F32 AngleDifference(F32 a, F32 b, Bool clamp = true);

	// Return the closest power of 2 to f.
	S32 Log2(S32 f);

	// Return true if f is not a valid IEEE floating point value.
	Bool IsNaN(F32 f);

	// Return the quadratic roots of ax^2 + bx + c = 0.
	S32 QuadraticRoots(F32 a, F32 b, F32 c, F32* pRoots);

	// Get the order of points from a dominant plane
	const S32* GetDominantPlaneOrder(S32 plane);

	// Include inline definitions within the namespace!

#include "libmaths.inl"

// namespace libmaths
}

#pragma managed(pop)
