//********************************************************************************
//
// Core maths implementation.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Return the sine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Sin(F32 a)
{

#if LIBMATHS_WIN32_MSVC

	F32 s;

	__asm
	{
		fld a
		fsin
		fstp s
	}
	return s;

#else

	return LIBMATHS_STD::sinf(a);

#endif

}


//********************************************************************************
// Return the cosine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Cos(F32 a)
{

#if LIBMATHS_WIN32_MSVC

	F32 c;

	__asm
	{
		fld a
		fcos
		fstp c
	}
	return c;

#else

	return LIBMATHS_STD::cosf(a);

#endif

}


//********************************************************************************
// Set c to the cosine of f, s to the sine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE void LIBMATHS_FASTCALL SinCos(F32 f, F32& s, F32& c)
{

#if LIBMATHS_WIN32_MSVC

	__asm
	{
		fld f
		fsincos
		mov eax,[c]
		fstp dword ptr [eax]
		mov eax,[s]
		fstp dword ptr [eax]
	}

#else

	s = Sin(f);
	c = Cos(f);

#endif

}


//********************************************************************************
// Set c to the cosine of f, s to the sine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE void LIBMATHS_FASTCALL CosSin(F32 f, F32& c, F32& s)
{
	SinCos(f, s, c);
}


//********************************************************************************
// Return the tangent of f. f must be in the range -pi/2 to +pi/2.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL libmaths::Tan(F32 f)
{

#if LIBMATHS_WIN32_MSVC

	F32 t;

	__asm
	{
		fld f
		fptan
		fstp t
		fstp t
	}
	return t;

#else

	return LIBMATHS_STD::tanf(f);

#endif

}


//********************************************************************************
// Return the arc-sine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Asin(F32 f)
{

#if LIBMATHS_WIN32_MSVC

	F32 a;

	__asm
	{
		fld f
		fld st(0) ; Duplicate f on tos.
		fmul ; Compute f ** 2.
		fld st(0) ; Duplicate f ** 2 on tos.
		fld1 ; Compute 1 - f ** 2.
		fsubr
		fdiv ; Compute f ** 2 / (1 - f ** 2).
		fsqrt ; Compute sqrt((f ** 2) / (1 - f ** 2)).
		fld1 ; To compute full arctangent.
		fpatan ; Compute atan of the above
		fstp a
	}

	return a;

#else

	return LIBMATHS_STD::asinf(f);

#endif

}


//********************************************************************************
// Return the arc-cosine of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Acos(F32 f)
{

#if LIBMATHS_WIN32_MSVC

	F32 a;

	__asm
	{
		fld f
		fld st(0) ; Duplicate f on tos.
		fmul ; Compute f ** 2.
		fld st(0) ; Duplicate f ** 2 on tos.
		fld1 ; Compute 1 - f ** 2.
		fsubr
		fdivr ; Compute (1 - f ** 2) / f ** 2.
		fsqrt ; Compute sqrt((1 - f ** 2) / f ** 2).
		fld1 ; To compute full arctangent.
		fpatan ; Compute atan of the above
		fstp a
	}

	return a;

#else

	return LIBMATHS_STD::acosf(f);

#endif

}


//********************************************************************************
// Return the arc-tangent of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Atan(F32 f)
{

#if LIBMATHS_WIN32_MSVC

	F32 a;

	__asm
	{
		fld f
		fld1
		fpatan
		fstp a
	}
	return a;

#else

	return LIBMATHS_STD::atanf(f);

#endif

}


//********************************************************************************
// Return the arc-tangent of y/x.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Atan2(F32 y, F32 x)
{
	return LIBMATHS_STD::atan2f(y, x);
}


//********************************************************************************
// Return the fast arc-tangent of y/x.
// The max error is a little less than 0.07 rads (only at a few angles though).
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Atan2Fast(F32 y, F32 x)
{
	F32 angle;
	F32 coeff1 = libmaths::PI * 0.25f;
	F32 coeff2 = coeff1 * 3.0f;
	F32 absY = libmaths::Abs(y) + FLOAT_EPSILON;

	if (x >= 0.0f)
	{
		angle = coeff1 - coeff1 * ((x - absY) / (x + absY));
	}
	else
	{
		angle = coeff2 - coeff1 * ((x + absY) / (absY - x));
	}

	return ((y < 0.0f) ? -angle : angle);
}


//********************************************************************************
// Return f raised to the power of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Exp(F32 f)
{
	return LIBMATHS_STD::expf(f);
}


//********************************************************************************
// Return the square root of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Sqrt(F32 f)
{
	LIBMATHS_ASSERT(f >= 0.0f);

#if LIBMATHS_WIN32_MSVC

	F32 r;

	__asm
	{
		fld f
		fsqrt
		fstp r
	}

	return r;

#else

	return LIBMATHS_STD::sqrtf(f);

#endif

}


//********************************************************************************
// Inverse square root of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL InvSqrt(F32 f)
{
	return 1.0f / Sqrt(f);
}


//********************************************************************************
// Return the fast inverse square root of f
// NB. Test the speed vs. normal InvSqrt on your target platform, this NEEDS to be
// inlined.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL InvSqrtFast(F32 f)
{
	const F32 half = 0.5f * f;
	S32 i = *(S32*)&f;

	i = 0x5f3759df - (i >> 1);
	f = *(F32*)&i;

	return f * (1.5f - half * f * f);
}


//********************************************************************************
// Return the fast square root of f. (Up to 0.2% under, never over).
// NB. Test the speed vs. normal Sqrt on your target platform, this NEEDS to be
// inlined.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL SqrtFast(F32 f)
{
	LIBMATHS_ASSERT(f >= 0.0f);

	return f * InvSqrtFast(f);
}


//********************************************************************************
// Return the super fast square root of f, but inaccurately. (Up to 6% over, never
// under). NB. Test the speed vs. normal Sqrt on your target platform, this NEEDS
// to be inlined.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL SqrtSuperFast(F32 f)
{
	S32 i = (1 << 29) + (*(S32*)&f >> 1) - (1 << 22);

	return *(F32*)&i;
}


//********************************************************************************
// Return the super fast inverse square root of f
// NB. Test the speed vs. normal InvSqrt on your target platform, this NEEDS to be
// inlined.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL InvSqrtSuperFast(F32 f)
{
	return 1.0f / SqrtSuperFast(f);
}


//********************************************************************************
// Return the absolute value of f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Abs(F32 f)
{
	S32 a = (*(S32 *)&f) & 0x7FFFFFFF;
	return *(F32*)&a;
}


//********************************************************************************
// Return the integer value immediately less than f.
//********************************************************************************

LIBMATHS_FORCEINLINE S32 LIBMATHS_FASTCALL Floor(F32 f)
{
	return FToI(f) - ((*(U32 *)&f & 0x80000000) >> 31);
}


//********************************************************************************
// Return the integer value immediately less than f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Floorf(F32 f)
{
	return F32(Floor(f));
}


//********************************************************************************
// Return the integer value immediately greater than f.
//********************************************************************************

LIBMATHS_FORCEINLINE S32 LIBMATHS_FASTCALL Ceil(F32 f)
{
	S32 floor = Floor(f);

	return (((F32)floor == f) ? floor : floor + 1);
}


//********************************************************************************
// Return the integer value immediately greater than f.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Ceilf(F32 f)
{
	return F32(Ceil(f));
}


//********************************************************************************
// Return the integer value closest to f.
//********************************************************************************

LIBMATHS_FORCEINLINE S32 LIBMATHS_FASTCALL Round(F32 f)
{
	U32 offset = 0x3f000000 | (*(U32 *)&f & 0x80000000);

	return FToI(f + *(F32 *)&offset);
}


//********************************************************************************
// Return the remainder of x/y.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Fmod(F32 x, F32 y)
{
	return x - FToI(x / y) * y;
}


//********************************************************************************
// Return the sign of f, or 0 if f is zero.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Zsgn(F32 f)
{
	U32 t = *(U32*)(&f);

	if (t & 0x80000000)
		return -1.0f;
	else if (t == 0)
		return 0.0f;
	else
		return 1.0f;
}


//********************************************************************************
// Return the sign of f. If f is zero then return +1.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL Sgn(F32 f)
{
	return (*(U32*)(&f) & 0x80000000) ? -1.0f : 1.0f;
}


//********************************************************************************
// Convert the angle a from degrees to radians.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL DegToRad(F32 a)
{
	return a * PIO180;
}


//********************************************************************************
// Convert the angle a from radians to degrees.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 LIBMATHS_FASTCALL RadToDeg(F32 a)
{
	return a * ONE80OPI;
}


//********************************************************************************
// Return whether f is approximately zero.
//********************************************************************************

LIBMATHS_FORCEINLINE Bool LIBMATHS_FASTCALL Eqz(F32 f)
{
	return (Abs(f) < libmaths::Epsilon);
}


//********************************************************************************
// Return whether f is approximately zero by comparison with e.
//********************************************************************************

LIBMATHS_FORCEINLINE Bool LIBMATHS_FASTCALL Eqz(F32 f, F32 e)
{
	return (Abs(f) < e);
}


//********************************************************************************
// Return whether a is approximately equal to b.
//********************************************************************************

LIBMATHS_FORCEINLINE Bool LIBMATHS_FASTCALL Approx(F32 a, F32 b)
{
	return Eqz(a-b);
}


//********************************************************************************
// Return whether a is approximately equal to b withing the range of e.
//********************************************************************************

LIBMATHS_FORCEINLINE Bool LIBMATHS_FASTCALL Approx(F32 a, F32 b, F32 e)
{
	return Eqz(a-b, e);
}


//********************************************************************************
// Return f*f. (Should only be used for scalar types).
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Square(T f)
{
	return f * f;
}


//********************************************************************************
// Return the minimum of a and b.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Min(T a, T b)
{
	return (a < b) ? a : b;
}


//********************************************************************************
// Return the minimum of a, b and c.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Min3(T a, T b, T c)
{
	return Min(Min(a, b), c);
}


//********************************************************************************
// Return the minimum of a, b, c and d.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Min4(T a, T b, T c, T d)
{
	return Min(Min(Min(a, b), c), d);
}


//********************************************************************************
// Return the maximum of a and b.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Max(T a, T b)
{
	return (a > b) ? a : b;
}


//********************************************************************************
// Return the maximum of a, b and c.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Max3(T a, T b, T c)
{
	return Max(Max(a, b), c);
}


//********************************************************************************
// Return the maximum of a, b, c and d.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Max4(T a, T b, T c, T d)
{
	return Max(Max(Max(a, b), c), d);
}


//********************************************************************************
// Swap the values of a and b. (May incur a performance hit for large structures).
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE void Swap(T& a, T& b)
{
	T t = b;
	b = a;
	a = t;
}


//********************************************************************************
// Linear interpolation from l to h by an amount f (0 to 1).
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Lerp(F32 f, T l, T h)
{
	return (T)(l + ((h - l) * f));
}


//********************************************************************************
// Clamp v to the range l to h.
//********************************************************************************

template <typename T> LIBMATHS_FORCEINLINE T Clamp(T v, T l, T h)
{
	return (v < l) ? l : (v > h) ? h : v;
}


//********************************************************************************
// Reduce angle a to the range -pi * 2 to +pi * 2.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 ReduceAngle(F32 x)
{
	return Fmod(x, TWOPI);
}


//********************************************************************************
// Reduce angle a to the range -pi to +pi.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 ReduceAnglePI(F32 x)
{
	x = Fmod(x, TWOPI);

	if (Abs(x) > PI)
	{
		x = -(TWOPI - x);
	}

	return x;
}


//********************************************************************************
// Reduce angle a to the range 0 to +pi.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 ReduceAnglePI2(F32 x)
{
	x = Fmod(x, TWOPI);

	if (x < 0.0f)
	{
		x = TWOPI - x;
	}

	return x;
}


//********************************************************************************
// Get the difference between angle a and angle b.
//********************************************************************************

LIBMATHS_FORCEINLINE F32 AngleDifference(F32 a, F32 b, Bool clamp)
{
	if (clamp == true)
	{
		a = ReduceAnglePI2(a);
		b = ReduceAnglePI2(b);
	}

	return Min((TWOPI) - Abs(a - b), Abs(a - b));
}


//********************************************************************************
// Return the closest power of 2 to f.
//********************************************************************************

LIBMATHS_FORCEINLINE S32 Log2(S32 f)
{
	S32 ret = 0;
	while ((1<<ret) < f) ret++;
	return ret;
}


//********************************************************************************
// Return true if f is not a valid IEEE floating point value.
//********************************************************************************

LIBMATHS_FORCEINLINE Bool IsNaN(F32 f)
{
	return ((*(U32*)&f) & libmaths::NaNMask) == libmaths::NaNMask;
}


//********************************************************************************
// Convert f from F32 to signed integer. Always rounds towards 0.
// (i.e. equivalent to an integer cast).
//********************************************************************************

LIBMATHS_FORCEINLINE S32 LIBMATHS_FASTCALL FToI(const F32 f)
{
	// This is completely portable (for IEEE 754 compliant floats) and not much slower than a direct fistp on IX86!

	// take bit pattern of F32 into a register
	S32 a = *reinterpret_cast<const S32*>(&f);
	// extract the exponent
	S32 exponent = ((a&0x7fffffff)>>23)-127;
	// extract mantissa and add the hidden bit
	S32 mantissa = (a&((1<<23)-1))|(1<<23);
	// sign = 0xFFFFFFFF if original value is negative, 0 if positive
	S32 sign = (a>>31);
	// ((1<<exponent)*mantissa)>>24 -- (we know that mantissa > (1<<24))
	S32 r = (S32)(((U32)(mantissa)<<8)>>(31-exponent));
	// add original sign. If exponent was negative, make return value 0.
	return ((r ^ (sign)) - sign ) & ~(exponent>>31);
}


//********************************************************************************
// Get the order of points from a dominant plane
//********************************************************************************

LIBMATHS_FORCEINLINE const S32* GetDominantPlaneOrder(S32 plane)
{
	static S32 Orders[3][3] =
	{
		{ 1, 2, 0 },
		{ 0, 2, 1 },
		{ 0, 1, 2 }
	};

	return &Orders[plane][0];
}
