//********************************************************************************
//
// A class used to define line segments in 2D space, often used for detecting
// intersections with other geometric types such as circles and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"


namespace libmaths
{

	class AxisAlignedRectangle;
	class Circle;
	class Triangle2;

	//********************************************************************************
	// Class for a 2-dimensional line. It has a start position and a combined
	// direction and length.
	//********************************************************************************

	class LineSegment2
	{

	public:

		// Construct a line
		LineSegment2(void) :
			_origin(Vector2::ZERO()),
			_direction(Vector2::ZERO())
		{ }

		// Construct a line with a start point and a direction (should always be normalised)
		LineSegment2(const Vector2& start, const Vector2& direction) :
			_origin(start),
			_direction(direction)
		{ }

		// Access the start position of the line
		const Vector2& Origin(void) const
		{ return _origin; }

		// Access the start position of the line
		void Origin(const Vector2& origin)
		{ _origin = origin; }

		// Access the start direction of the line
		const Vector2& Direction(void) const
		{ return _direction; }

		// Access the start direction of the line
		void Direction(const Vector2& direction)
		{ _direction = direction; }

		// Retrieve the start or end point of the line
		Vector2 Point(U32 point) const
		{ return (point == 0) ? _origin : _origin + _direction; }

		// Get the distance of the line from a point in space
		F32 Distance(const Vector2& point, F32& pointOnLine, Bool clamp = true) const;

		// Get the squared distance of the line from a point in space
		F32 DistanceSqr(const Vector2& point, F32& pointOnLine, Bool clamp = true) const;

		// Get the distance of the line from a line in space
		F32 Distance(const LineSegment2& line, F32& pointOnLine, F32& pointOnOtherLine) const;

		// Get the squared distance of the line from a line in space
		F32 DistanceSqr(const LineSegment2& line, F32& pointOnLine, F32& pointOnOtherLine) const;

		// Find the point on the line closest to the given point
		Vector2 ClosestPoint(Vector2 point) const;

		// Find the point on the line closest to the given line
		Vector2 ClosestPoint(const LineSegment2& line) const;

		// Does the line intersect a rectangle?
		Bool Intersects(AxisAlignedRectangle& rectangle) const;

		// Does the line intersect a rectangle?
		Bool Intersects(AxisAlignedRectangle& rectangle, Vector2& point, Vector2& normal, F32& t) const;

		// Does the line intersect a sphere?
		Bool Intersects(Circle& circle) const;

		// Does the line intersect a sphere?
		Bool Intersects(Circle& circle, Vector2& point, Vector2& normal, F32& t) const;

		// Does the line intersect a triangle?
		Bool Intersects(Triangle2& triangle) const;

		// Does the line intersect a triangle?
		Bool Intersects(Triangle2& triangle, Vector2& point, Vector2& normal, F32& t) const;

	private:

		// The start position of the line
		Vector2 _origin;

		// The direction and length of the line
		Vector2 _direction;
	};

// namespace libmaths
}

#pragma managed(pop)
