//********************************************************************************
//
// A class used to define line segments in 3D space, often used for detecting
// intersections with other geometric types such as spheres and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"


namespace libmaths
{

	class AxisAlignedBox;
	class Capsule;
	class Sphere;
	class Triangle3;
	class Plane;


	//********************************************************************************
	// Class for a 3-dimensional line. It has a start position and a combined
	// direction and length.
	//********************************************************************************

	class LineSegment3
	{

	public:

		// Construct a line
		LineSegment3(void) :
			_origin(Vector3x::ZERO()),
			_direction(Vector3x::ZERO())
		{ }

		// Construct a line with a start point and a direction
		LineSegment3(const Vector3x& start, const Vector3x& direction) :
			_origin(start),
			_direction(direction)
		{ }

		// Access the start position of the line
		const Vector3x& Origin(void) const
		{ return _origin; }

		// Access the start position of the line
		void Origin(const Vector3x& origin)
		{ _origin = origin; }

		// Access the start direction of the line
		const Vector3x& Direction(void) const
		{ return _direction; }

		// Access the start direction of the line
		void Direction(const Vector3x& direction)
		{ _direction = direction; }

		// Retrieve the start or end point of the line
		Vector3x Point(U32 point) const
		{ return (point == 0) ? _origin : _origin + _direction; }

		// Get the distance of the line from a point in space
		F32 Distance(const Vector3x& point, F32& pointOnLine, Bool clamp = true) const;

		// Get the squared distance of the line from a point in space
		F32 DistanceSqr(const Vector3x& point, F32& pointOnLine, Bool clamp = true) const;

		// Get the distance of the line from a line in space
		F32 Distance(const LineSegment3& line, F32& pointOnLine, F32& pointOnOtherLine) const;

		// Get the squared distance of the line from a line in space
		F32 DistanceSqr(const LineSegment3& line, F32& pointOnLine, F32& pointOnOtherLine) const;

		// Find the point on the line closest to the given point
		Vector3x ClosestPoint(Vector3x point) const;

		// Find the point on the line closest to the given line
		Vector3x ClosestPoint(const LineSegment3& line) const;

		// Does the line intersect a box?
		Bool Intersects(const AxisAlignedBox& box) const;

		// Does the line intersect a box?
		Bool Intersects(const AxisAlignedBox& box, Vector3x& point, Vector3x& normal, F32& t) const;

		// Does the line intersect a capsule?
		Bool Intersects(const Capsule& capsule) const;

		// Does the line intersect a capsule?
		Bool Intersects(const Capsule& capsule, Vector3x& point, Vector3x& normal, F32& t) const;

		// Does the line intersect a sphere?
		Bool Intersects(const Sphere& sphere) const;

		// Does the line intersect a sphere?
		Bool Intersects(const Sphere& sphere, Vector3x& point, Vector3x& normal, F32& t) const;

		// Does the line intersect a triangle?
		Bool Intersects(const Triangle3& triangle, Bool twoSided = false) const;

		// Does the line intersect a triangle?
		Bool Intersects(const Triangle3& triangle, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided = false) const;

		// Does the line intersect a plane?
		Bool Intersects(const Plane& plane, Bool twoSided = false) const;

		// Does the line intersect a plane?
		Bool Intersects(const Plane& plane, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided = false) const;

	private:

		// The start position of the line
		Vector3x _origin;

		// The direction and length of the line
		Vector3x _direction;
	};

// namespace libmaths
}

#pragma managed(pop)
