//********************************************************************************
//
// 2x2 matrix class. Representing the top 2x2 section of what would normally be a
// 3x3 matrix. Used to transform 2 component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"


namespace libmaths
{

	//********************************************************************************
	// A structure to represent a 2x2 matrix.
	//********************************************************************************

	class Matrix22of33
	{

	private:

#if LIBMATHS_COLUMN_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 2 && U32(column) < 2); return _array[column][row]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 2 && U32(column) < 2); return _array[column][row]; }

#elif LIBMATHS_ROW_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 2 && U32(column) < 2); return _array[row][column]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 2 && U32(column) < 2); return _array[row][column]; }

#else

#error No matrix format defined

#endif

	public:

		// A fixed matrix of all zeroes (0, 0).
		static const Matrix22of33& ZERO(void);

		// A fixed matrix of identity (zeroes with 1, 1 across the diagonal).
		static const Matrix22of33& IDENTITY(void);

		// Default constructor.
		Matrix22of33(void);

		// Construct from 4 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		explicit Matrix22of33(const F32* pValues);

		// Construct from 4 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix22of33(const F32 values[2][2]);

		// Construct from 4 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix22of33(
			F32 m00, F32 m01,
			F32 m10, F32 m11);

		// Construct from a 3x3 matrix.
		// The third column and translation are discarded in the process.
		Matrix22of33(const Matrix33& matrix);

		// Access / alter matrix elements by array index.
		F32& operator [] (const S32 i)
		{ return Element(i >> 1, i & 1); }

		// Access elements by array index.
		F32 operator [] (const S32 i) const
		{ return Element(i >> 1, i & 1); }

		// Access / alter matrix elements by row and column.
		F32& operator () (const S32 row, const S32 column)
		{ return Element(row, column); }

		// Access matrix elements by row and column.
		F32 operator () (const S32 row, const S32 column) const
		{ return Element(row, column); }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void) const
		{ return &_array[0][0]; }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void)
		{ return &_array[0][0]; }

		// Set this matrix to zero.
		Matrix22of33& Zero(void)
		{ return ((*this) = Matrix22of33::ZERO()); }

		// Set this matrix to identity (including translation).
		Matrix22of33& Identity(void)
		{ return ((*this) = Matrix22of33::IDENTITY()); }

		// Get the x axis as a 2D vector.
		Vector2 XAxis(void) const
		{ return Vector2(Element(0, 0), Element(0, 1)); }

		// Set the x axis from a 2D vector.
		void XAxis(const Vector2& axis);

		// Set the x axis from 2 floats.
		void XAxis(F32 x, F32 y)
		{ XAxis(Vector2(x, y)); }

		// Get the y axis as a 2D vector.
		Vector2 YAxis(void) const
		{ return Vector2(Element(1, 0), Element(1, 1)); }

		// Set the y axis from a 2D vector.
		void YAxis(const Vector2& axis);

		// Set the y axis from 2 floats.
		void YAxis(F32 x, F32 y)
		{ YAxis(Vector2(x, y)); }

		// Return indexed row of matrix as a 2D vector.
		Vector2 Row(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 2); return Vector2((*this)(i, 0), (*this)(i, 1)); }

		// Return indexed column of matrix as 2D vector.
		Vector2 Column(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 2); return Vector2((*this)(0, i), (*this)(1, i)); }

		// Build a transformation from an anti-clockwise angle.
		Matrix22of33& FromAngle(F32 angle);

		// Build a transformation from a scale value.
		Matrix22of33& FromScale(F32 scale);

		// Build a transformation from separate x and y scale values.
		Matrix22of33& FromScale(F32 x, F32 y);

		// Post-concatenate a scaling to this matrix.
		void PostCatScale(F32 x, F32 y);

		// Scale this matrix uniformly.
		Matrix22of33& Scale(F32 scale);

		// Scale this matrix non-uniformly.
		Matrix22of33& Scale(F32 x, F32 y);

		// Get a new matrix from this matrix uniformly scaled.
		Matrix22of33 Scaled(F32 scale) const
		{ return Matrix22of33(*this).Scale(scale); }

		// Get a new matrix from this matrix non-uniformly scaled.
		Matrix22of33 Scaled(F32 x, F32 y) const
		{ return Matrix22of33(*this).Scale(x, y); }

		// Transpose this matrix's rows and columns.
		Matrix22of33& Transpose(void);

		// Get a new matrix from this matrix transposed.
		Matrix22of33 Transposed(void) const;

		// Invert this matrix, producing an inverse transformation.
		Matrix22of33& Invert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation.
		Matrix22of33 Inverse(void) const;

		// Invert this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix22of33& OrthonormalInvert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix22of33 OrthonormalInverse(void) const;

		// Orthonormalise this matrix using the Gram-Schmidt algorithm.
		Matrix22of33& Orthonormalise(void);

		// Get the determinant of this matrix.
		F32 Determinant(void) const;

		// Build this matrix as an interpolation between the two given matrices,
		// with possible extra 360 degree spins.
		Matrix22of33& Lerp(F32 t, Matrix22of33& m0, Matrix22of33& m1, U32 numSpins = 0);

		// Apply this matrix to an array of 2 floats.
		void TransformVector2f(const F32* pSource, F32* pDestination) const;

		// Apply this matrix to a 2D vector.
		void TransformVector(const Vector2& source, Vector2& destination) const;

		// Apply this matrix to a 2D vector.
		Vector2 TransformVector(const Vector2& source) const;

		// Apply inverse of this matrix to an array of 2 floats.
		void InverseTransformVector2f(const F32* pSource, F32* pDestination) const;

		// Apply inverse of this matrix to a 2D vector.
		void InverseTransformVector(const Vector2& source, Vector2& destination) const;

		// Apply inverse of this matrix to a 2D vector.
		Vector2 InverseTransformVector(const Vector2& source) const;

		// Compare two matrices for equality.
		Bool operator == (const Matrix22of33& rhs) const;

		// Compare two matrices for inequality.
		Bool operator != (const Matrix22of33& rhs) const
		{ return (!(*this == rhs)); }

		// Get a new matrix that is the negative of this matrix.
		Matrix22of33 operator - (void) const;

		// Get a new matrix from another matrix added to this matrix.
		Matrix22of33 operator + (const Matrix22of33& rhs);

		// Get a new matrix from another matrix subtracted from this matrix.
		Matrix22of33 operator - (const Matrix22of33& rhs);

		// Get a new matrix from this matrix multiplied by (concatenated with) another matrix.
		Matrix22of33 operator * (const Matrix22of33& rhs);

		// Get a new matrix from this matrix scaled by a value.
		Matrix22of33 operator * (F32 scale) const
		{ return Scaled(scale); }

		// Multiply a 2D vector by this matrix.
		Vector2 operator * (const Vector2& rhs) const;

		// Get a new matrix from a matrix scaled by a value.
		friend Matrix22of33 operator * (F32 lhs, Matrix22of33& rhs)
		{ return rhs.Scaled(lhs); }

		// Add another matrix to this matrix.
		Matrix22of33& operator += (const Matrix22of33& rhs);

		// Subtract another matrix from this matrix.
		Matrix22of33& operator -= (const Matrix22of33& rhs);

		// Multiply this matrix by (concatenated with) another matrix.
		Matrix22of33& operator *= (const Matrix22of33& rhs);

		// Multiply (scale) this matrix by a value.
		Matrix22of33& operator *= (F32 scale)
		{ Scale(scale); return *this; }

		// Divide (scale) this matrix by a value.
		Matrix22of33& operator /= (F32 scale)
		{ Scale(1.0f / scale); return *this; }

	private:

		operator F32* ();
		operator F32** ();

		// 4 F32 components.
		F32 _array[2][2];
	};

	// Now include the inline definitions.

#include "matrix22of33.inl"

// namespace libmaths
}

#pragma managed(pop)
