//********************************************************************************
//
// 2x2 matrix class. Representing the top 2x2 section of what would normally be a
// 3x3 matrix. Used to transform 2 component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Defines Affecting this file:
//
// COLUMN_MAJOR_MATRIX_FORMAT
// Use column major storage and operations.
//
// ROW_MAJOR_MATRIX_FORMAT
// Use row major storage and operations (the default)
//********************************************************************************


//********************************************************************************
// A fixed matrix of all zeroes (0, 0).
//********************************************************************************

LIBMATHS_INLINE const Matrix22of33& Matrix22of33::ZERO(void)
{
	static Matrix22of33 matrix(0.0f, 0.0f, 0.0f, 0.0f);

	return matrix;
}


//********************************************************************************
// A fixed matrix of identity (zeroes with 1, 1 across the diagonal).
//********************************************************************************

LIBMATHS_INLINE const Matrix22of33& Matrix22of33::IDENTITY(void)
{
	static Matrix22of33 matrix(1.0f, 0.0f, 0.0f, 1.0f);

	return matrix;
}


//********************************************************************************
// Default constructor.
// N.B. Does not initialise any data members unless DEFAULT_INITIALISERS is defined.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33::Matrix22of33(void)
{
#if defined(DEFAULT_INITIALISERS)
	Identity();
#endif
}


//********************************************************************************
// Construct from 4 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33::Matrix22of33(const F32* pValues)
{
	Element(0, 0) = pValues[0 * 2 + 0]; Element(0, 1) = pValues[0 * 2 + 1];
	Element(1, 0) = pValues[1 * 2 + 0]; Element(1, 1) = pValues[1 * 2 + 1];
}


//********************************************************************************
// Construct from 4 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33::Matrix22of33(const F32 values[2][2])
{
	Element(0, 0) = values[0][0]; Element(0, 1) = values[0][1];
	Element(1, 0) = values[1][0]; Element(1, 1) = values[1][1];
}


//********************************************************************************
// Construct from 4 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33::Matrix22of33(F32 m00, F32 m01,
									   F32 m10, F32 m11)
{
	Element(0, 0) = m00; Element(0, 1) = m01;
	Element(1, 0) = m10; Element(1, 1) = m11;
}


//********************************************************************************
// Set the x axis from a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::XAxis(const Vector2& axis)
{
	Element(0, 0) = axis.X();
	Element(0, 1) = axis.Y();
}


//********************************************************************************
// Set the y axis from a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::YAxis(const Vector2& axis)
{
	Element(1, 0) = axis.X();
	Element(1, 1) = axis.Y();
}


//********************************************************************************
// Post-concatenate a scaling to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::PostCatScale(F32 x, F32 y)
{
	for (S32 i = 0; i < 2; i++)
	{
		Element(i, 0) *= x;
		Element(i, 1) *= y;
	}
}


//********************************************************************************
// Build a transformation from an anti-clockwise angle.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33& Matrix22of33::FromAngle(F32 angle)
{
	F32 c, s;

	CosSin(angle, c, s);

	*this = Matrix22of33(c, s, -s, c);

	return *this;
}


//********************************************************************************
// Transpose this matrix's rows and columns.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33& Matrix22of33::Transpose(void)
{
	return *this = Matrix22of33(Element(0, 0), Element(1, 0), Element(0, 1), Element(1, 1));
}


//********************************************************************************
// Get a new matrix from this matrix transposed.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33 Matrix22of33::Transposed(void) const
{
	return Matrix22of33(Element(0, 0), Element(1, 0), Element(0, 1), Element(1, 1));
}


//********************************************************************************
// Apply this matrix to an array of 2 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::TransformVector2f(const F32* pSource, F32* pDestination) const
{
	F32 ix = pSource[0], iy = pSource[1];

	pDestination[0] = Element(0, 0) * ix + Element(1, 0) * iy;
	pDestination[1] = Element(0, 1) * ix + Element(1, 1) * iy;
}


//********************************************************************************
// Apply this matrix to a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::TransformVector(const Vector2& source, Vector2& destination) const
{
	TransformVector2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 2D vector.
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix22of33::TransformVector(const Vector2& source) const
{
	Vector2 result;
	
	TransformVector(source, result);
	
	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 2 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::InverseTransformVector2f(const F32* pSource, F32* pDestination) const
{
	F32 ix = pSource[0], iy = pSource[1];

	pDestination[0] = Element(0, 0) * ix + Element(0, 1) * iy;
	pDestination[1] = Element(1, 0) * ix + Element(1, 1) * iy;
}


//********************************************************************************
// Apply inverse of this matrix to a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix22of33::InverseTransformVector(const Vector2& source, Vector2& destination) const
{
	InverseTransformVector2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 2D vector.
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix22of33::InverseTransformVector(const Vector2& source) const
{
	Vector2 result;
	
	InverseTransformVector(source, result);
	
	return result;
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix22of33::operator * (const Vector2& rhs) const
{
	return TransformVector(rhs);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 2x2 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector2 Vector2::operator * (const Matrix22of33& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Multiply (transform) this vector by a 2x2 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::operator *= (const Matrix22of33& rhs)
{
	return *this = rhs * (*this);
}


//********************************************************************************
// Compare two matrices for equality.
//********************************************************************************

LIBMATHS_INLINE Bool Matrix22of33::operator == (const Matrix22of33& matrix) const
{
	for (S32 i = 0; i < 2; i++)
	{
		for (S32 j = 0; j < 2; j++)
		{
			if (!Approx(matrix.Element(i, j), Element(i, j), Vector2::Epsilon()))
			{
				return false;
			}
		}
	}

	return true;
}


//********************************************************************************
// Get a new matrix that is the negative of this matrix
//********************************************************************************

LIBMATHS_INLINE Matrix22of33 Matrix22of33::operator - (void) const
{
	return Matrix22of33
		(
		-Element(0, 0), -Element(0, 1),
		-Element(1, 0), -Element(1, 1)
		);
}


//********************************************************************************
// Get a new matrix from another matrix added to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33 Matrix22of33::operator + (const Matrix22of33& rhs)
{
	Matrix22of33 result;

	for (S32 row = 0; row < 2; row++)
	{
		for (S32 col = 0; col < 2; col++)
		{
			result.Element(row, col) = Element(row, col) + rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from another matrix subtracted from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33 Matrix22of33::operator - (const Matrix22of33& rhs)
{
	Matrix22of33 result;

	for (S32 row = 0; row < 2; row++)
	{
		for (S32 col = 0; col < 2; col++)
		{
			result.Element(row, col) = Element(row, col) - rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from this matrix multiplied by (concatenated with) another
// matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33 Matrix22of33::operator * (const Matrix22of33& rhs)
{
	Matrix22of33 result(*this);

	result *= rhs;

	return result;
}


//********************************************************************************
// Add another matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33& Matrix22of33::operator += (const Matrix22of33& rhs)
{
	for (S32 row = 0; row < 2; row++)
	{
		for (S32 col = 0; col < 2; col++)
		{
			Element(row, col) += rhs.Element(row, col);
		}
	}

	return *this;
}


//********************************************************************************
// Subtract another matrix from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33& Matrix22of33::operator -= (const Matrix22of33& rhs)
{
	for (S32 row = 0; row < 2; row++)
	{
		for (S32 col = 0; col < 2; col++)
		{
			Element(row, col) -= rhs.Element(row, col);
		}
	}

	return *this;
}


//********************************************************************************
// Multiply this matrix by (concatenated with) another matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix22of33& Matrix22of33::operator *= (const Matrix22of33& rhs)
{
	Matrix22of33 result;
	F32 t = 0.0f;

	for (S32 i = 0; i < 2; i++)
	{
		for (S32 j = 0; j < 2; j++)
		{
			t = 0.0f;

			for (S32 k = 0; k < 2; k++)
			{
				t += Element(i, k) * rhs.Element(k, j);
			}

			result.Element(i, j) = t;
		}
	}

	return (*this = result);

}
