//********************************************************************************
//
// 3x3 matrix class. Used for homogenous 2D transformations with 2 and 3
// component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"


namespace libmaths
{

	class Matrix22of33;


	//********************************************************************************
	// 3x3 matrix class.
	// Used to transform 2 and 3 component vectors.
	//********************************************************************************

	class Matrix33
	{

	private:

#if LIBMATHS_COLUMN_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[column][row]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[column][row]; }

#elif LIBMATHS_ROW_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[row][column]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[row][column]; }

#else

#error No matrix format defined

#endif

	public:

		// A fixed matrix of all zeroes (0, 0, 0).
		static const Matrix33& ZERO(void);

		// A fixed matrix of identity (zeroes with 1, 1, 1 across the diagonal).
		static const Matrix33& IDENTITY(void);

		// Default constructor.
		Matrix33(void);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		explicit Matrix33(const F32* pValues);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix33(const F32 values[3][3]);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix33(
			F32 m00, F32 m01, F32 m02,
			F32 m10, F32 m11, F32 m12,
			F32 m20, F32 m21, F32 m22);

		// Construct from a 2x2 matrix.
		// The third column is initialised to (0, 0, 1)
		// The translation is initialised to (0, 0)
		Matrix33(const Matrix22of33& matrix);

		// Access / alter matrix elements by array index.
		F32& operator [] (const S32 i)
		{ return Element(i / 3, i % 3); }

		// Access elements by array index.
		F32 operator [] (const S32 i) const
		{ return Element(i / 3, i % 3); }

		// Access / alter matrix elements by row and column.
		F32& operator () (const S32 row, const S32 column)
		{ return Element(row, column); }

		// Access matrix elements by row and column.
		F32 operator () (const S32 row, const S32 column) const
		{ return Element(row, column); }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void) const
		{ return &_array[0][0]; }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void)
		{ return &_array[0][0]; }

		// Set this matrix to zero.
		Matrix33& Zero(void)
		{ return ((*this) = Matrix33::ZERO()); }

		// Set this matrix to identity (including translation).
		Matrix33& Identity(void)
		{ return ((*this) = Matrix33::IDENTITY()); }

		// Get the x axis as a 2D vector.
		Vector2 XAxis(void) const
		{ return Vector2(Element(0, 0), Element(0, 1)); }

		// Set the x axis from a 2D vector.
		void XAxis(const Vector2& axis);

		// Set the x axis from 2 floats.
		void XAxis(F32 x, F32 y)
		{ XAxis(Vector2(x, y)); }

		// Get the y axis as a 2D vector.
		Vector2 YAxis(void) const
		{ return Vector2(Element(1, 0), Element(1, 1)); }

		// Set the y axis from a 2D vector.
		void YAxis(const Vector2& axis);

		// Set the y axis from 2 floats.
		void YAxis(F32 x, F32 y)
		{ YAxis(Vector2(x, y)); }

		// Get the translation as a 2D vector.
		Vector2 Translation(void) const
		{ return Vector2(Element(2, 0), Element(2, 1)); }

		// Set the translation from a 2D vector.
		void Translation(const Vector2& translation);

		// Set the translation from 2 floats.
		void Translation(F32 x, F32 y)
		{ Translation(Vector2(x, y)); }

		// Return indexed row of matrix as a 3D vector.
		Vector3x Row(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 3); return Vector3x((*this)(i, 0), (*this)(i, 1), (*this)(i, 2)); }

		// Return indexed column of matrix as 3D vector.
		Vector3x Column(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 3); return Vector3x((*this)(0, i), (*this)(1, i), (*this)(2, i)); }

		// Build a transformation from an anti-clockwise angle.
		Matrix33& FromAngle(F32 angle);

		// Build a transformation from a scale value.
		Matrix33& FromScale(F32 scale);

		// Build a transformation from separate x, y and z scale values.
		Matrix33& FromScale(F32 x, F32 y, F32 z = 1.0f);

		// Post-concatenate a scaling to this matrix.
		void PostCatScale(F32 x, F32 y, F32 z = 1.0f);

		// Post-concatenate a translation to this matrix.
		void PostCatTranslate(F32 x, F32 y);

		// Scale this matrix uniformly.
		Matrix33& Scale(F32 scale);

		// Scale this matrix non-uniformly.
		Matrix33& Scale(F32 x, F32 y, F32 z);

		// Get a new matrix from this matrix uniformly scaled.
		Matrix33 Scaled(F32 scale) const
		{ return Matrix33(*this).Scale(scale); }

		// Get a new matrix from this matrix non-uniformly scaled.
		Matrix33 Scaled(F32 x, F32 y, F32 z) const
		{ return Matrix33(*this).Scale(x, y, z); }

		// Transpose this matrix's rows and columns.
		Matrix33& Transpose(void);

		// Get a new matrix from this matrix transposed.
		Matrix33 Transposed(void) const;

		// Invert this matrix, producing an inverse transformation.
		Matrix33& Invert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation.
		Matrix33 Inverse(void) const;

		// Invert this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix33& OrthonormalInvert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix33 OrthonormalInverse(void) const;

		// Orthonormalise this matrix using the Gram-Schmidt algorithm.
		// NOTE: Only the upper 3x3 sub-matrix is affected.
		Matrix33& Orthonormalise(void);

		// Get the determinant of this matrix.
		F32 Determinant(void) const;

		// Build this matrix as an interpolation between the two given matrices,
		// with possible extra 360 degree spins.
		Matrix33& Lerp(F32 t, Matrix33& m0, Matrix33& m1, U32 numSpins = 0);

		// Apply this matrix to an array of 2 floats.
		void TransformPoint2f(const F32* pSource, F32* pDestination) const;

		// Apply this matrix to 2D point (including translation).
		void TransformPoint(const Vector2& source, Vector2& destination) const;

		// Apply this matrix to a 2D point (including translation).
		Vector2 TransformPoint(const Vector2& source) const;

		// Apply this matrix to an array of 3 floats.
		void Transform3f(const F32* pSource, F32* pDestination) const;

		// Apply this matrix to a homogenous 3D vector.
		void Transform(const Vector3x& source, Vector3x& destination) const;

		// Apply this matrix to a homogenous 3D vector.
		Vector3x Transform(const Vector3x& source) const;

		// Apply this matrix to a homogenous 3D vector.
		void Transform(const Vector3& source, Vector3& destination) const;

		// Apply this matrix to a homogenous 3D vector.
		Vector3 Transform(const Vector3& source) const;

		// Apply this matrix to an array of 2 floats (excluding translation).
		void TransformVector2f(const F32* pSource, F32* pDestination) const;

		// Apply this matrix to a 2D vector (excluding translation).
		void TransformVector(const Vector2& source, Vector2& destination) const;

		// Apply this matrix to a 2D vector (excluding translation).
		Vector2 TransformVector(const Vector2& source) const;

		// Apply inverse of this matrix to an array of 2 floats (including translation).
		void InverseTransformPoint2f(const F32* pSource, F32* pDestination) const;

		// Apply inverse of this matrix to a 2D point (including translation).
		void InverseTransformPoint(const Vector2& source, Vector2& destination) const;

		// Apply inverse of this matrix to a 2D point (including translation).
		Vector2 InverseTransformPoint(const Vector2& source) const;

		// Apply inverse of this matrix to an array of 3 floats.
		void InverseTransform3f(const F32* pSource, F32* pDestination) const;

		// Apply inverse of this matrix to a homogenous 3D vector.
		void InverseTransform(const Vector3x& source, Vector3x& destination) const;

		// Apply inverse of this matrix to a homogenous 3D vector.
		Vector3x InverseTransform(const Vector3x& source) const;

		// Apply inverse of this matrix to a homogenous 3D vector.
		void InverseTransform(const Vector3& source, Vector3& destination) const;

		// Apply inverse of this matrix to a homogenous 3D vector.
		Vector3 InverseTransform(const Vector3& source) const;

		// Apply inverse of this matrix to an array of 2 floats (excluding translation).
		void InverseTransformVector2f(const F32* pSource, F32* pDestination) const;

		// Apply inverse of this matrix to a 2D vector (excluding translation).
		void InverseTransformVector(const Vector2& source, Vector2& destination) const;

		// Apply inverse of this matrix to a 2D vector (excluding translation).
		Vector2 InverseTransformVector(const Vector2& source) const;

		// Compare two matrices for equality.
		Bool operator == (const Matrix33& rhs) const;

		// Compare two matrices for inequality.
		Bool operator != (const Matrix33& rhs) const
		{ return (!(*this == rhs)); }

		// Get a new matrix that is the negative of this matrix.
		Matrix33 operator - (void) const;

		// Get a new matrix from another matrix added to this matrix.
		Matrix33 operator + (const Matrix33& rhs);

		// Get a new matrix from another matrix subtracted from this matrix.
		Matrix33 operator - (const Matrix33& rhs);

		// Get a new matrix from this matrix multiplied by (concatenated with) another matrix.
		Matrix33 operator * (const Matrix22of33& rhs);

		// Get a new matrix from this matrix multiplied by (concatenated with) another matrix.
		Matrix33 operator * (const Matrix33& rhs);

		// Get a new matrix from this matrix scaled by a value.
		Matrix33 operator * (F32 scale) const
		{ return Scaled(scale); }

		// Multiply a 2D vector by this matrix.
		Vector2 operator * (const Vector2& vector) const;

		// Multiply a 3D vector by this matrix.
		Vector3x operator * (const Vector3x& vector) const;

		// Multiply a 3D vector by this matrix.
		Vector3 operator * (const Vector3& vector) const;

		// Get a new matrix from a matrix scaled by a value.
		friend Matrix33 operator * (F32 lhs, Matrix33& rhs)
		{ return rhs.Scaled(lhs); }

		// Add another matrix to this matrix.
		Matrix33& operator += (const Matrix33& rhs);

		// Subtract another matrix from this matrix.
		Matrix33& operator -= (const Matrix33& rhs);

		// Multiply this matrix by (concatenated with) another matrix.
		Matrix33& operator *= (const Matrix22of33& rhs);

		// Multiply this matrix by (concatenated with) another matrix.
		Matrix33& operator *= (const Matrix33& rhs);

		// Multiply (scale) this matrix by a value.
		Matrix33& operator *= (F32 scale)
		{ Scale(scale); return *this; }

		// Divide (scale) this matrix by a value.
		Matrix33& operator /= (F32 scale)
		{ Scale(1.0f / scale); return *this; }

	private:

		operator F32* (void);
		operator F32** (void);

		// 9 F32 components.
		F32 _array[3][3];
	};

	// Now include the inline definitions.

#include "matrix33.inl"

// namespace libmaths
}

#pragma managed(pop)
