//********************************************************************************
//
// 3x3 matrix class. Used for homogenous 2D transformations with 2 and 3
// component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Defines Affecting this file:
//
// COLUMN_MAJOR_MATRIX_FORMAT
// Use column major storage and operations.
//
// ROW_MAJOR_MATRIX_FORMAT
// Use row major storage and operations (the default)
//********************************************************************************


//********************************************************************************
// A fixed matrix of all zeroes (0, 0, 0).
//********************************************************************************

LIBMATHS_INLINE const Matrix33& Matrix33::ZERO(void)
{
	static Matrix33 matrix(
		0.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f);

	return matrix;
}


//********************************************************************************
// A fixed matrix of identity (zeroes with 1, 1, 1 across the diagonal).
//********************************************************************************

LIBMATHS_INLINE const Matrix33& Matrix33::IDENTITY(void)
{
	static Matrix33 matrix(
		1.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 1.0f);

	return matrix;
}


//********************************************************************************
// Default constructor.
// N.B. Does not initialise any data members unless DEFAULT_INITIALISERS is defined.
//********************************************************************************

LIBMATHS_INLINE Matrix33::Matrix33(void)
{
#if defined(DEFAULT_INITIALISERS)
	Identity();
#endif
}


//********************************************************************************
// Construct from 9 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix33::Matrix33(const F32* pValues)
{
	Element(0, 0) = pValues[0 * 3 + 0]; Element(0, 1) = pValues[0 * 3 + 1]; Element(0, 2) = pValues[0 * 3 + 2];
	Element(1, 0) = pValues[1 * 3 + 0]; Element(1, 1) = pValues[1 * 3 + 1]; Element(1, 2) = pValues[1 * 3 + 2];
	Element(2, 0) = pValues[2 * 3 + 0]; Element(2, 1) = pValues[2 * 3 + 1]; Element(2, 2) = pValues[2 * 3 + 2];
}


//********************************************************************************
// Construct from 9 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix33::Matrix33(const F32 values[3][3])
{
	Element(0, 0) = values[0][0]; Element(0, 1) = values[0][1]; Element(0, 2) = values[0][2];
	Element(1, 0) = values[1][0]; Element(1, 1) = values[1][1]; Element(1, 2) = values[1][2];
	Element(2, 0) = values[2][0]; Element(2, 1) = values[2][1]; Element(2, 2) = values[2][2];
}


//********************************************************************************
// Construct from 9 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix33::Matrix33(	F32 m00, F32 m01, F32 m02,
										F32 m10, F32 m11, F32 m12,
										F32 m20, F32 m21, F32 m22)
{
	Element(0, 0) = m00; Element(0, 1) = m01; Element(0, 2) = m02;
	Element(1, 0) = m10; Element(1, 1) = m11; Element(1, 2) = m12;
	Element(2, 0) = m20; Element(2, 1) = m21; Element(2, 2) = m22;
}


//********************************************************************************
// Set the x axis from a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::XAxis(const Vector2& axis)
{
	Element(0, 0) = axis.X();
	Element(0, 1) = axis.Y();
}


//********************************************************************************
// Set the y axis from a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::YAxis(const Vector2& axis)
{
	Element(1, 0) = axis.X();
	Element(1, 1) = axis.Y();
}


//********************************************************************************
// Set the translation from a 2D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::Translation(const Vector2& translation)
{
	Element(2, 0) = translation.X();
	Element(2, 1) = translation.Y();
}


//********************************************************************************
// Post-concatenate a scaling to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::PostCatScale(F32 x, F32 y, F32 z)
{
	for (S32 i = 0; i < 3; i++)
	{
		Element(i, 0) *= x;
		Element(i, 1) *= y;
		Element(i, 2) *= z;
	}
}


//********************************************************************************
// Post-concatenate a translation to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::PostCatTranslate(F32 x, F32 y)
{
	for (S32 i = 0; i < 3; i++)
	{
		Element(i, 0) += Element(i, 2) * x;
		Element(i, 1) += Element(i, 2) * y;
	}
}


//********************************************************************************
// Build a transformation from an anti-clockwise angle.
//********************************************************************************

LIBMATHS_INLINE Matrix33& Matrix33::FromAngle(F32 angle)
{
	F32 c, s;

	CosSin(angle, c, s);

	*this = Matrix33(c, s, 0.0f, -s, c, 0.0f, 0.0f, 0.0f, 1.0f);

	return *this;
}


//********************************************************************************
// Transpose this matrix's rows and columns.
//********************************************************************************

LIBMATHS_INLINE Matrix33& Matrix33::Transpose(void)
{
	return *this = Matrix33
		(
		Element(0, 0), Element(1, 0), Element(2, 0),
		Element(0, 1), Element(1, 1), Element(2, 1),
		Element(0, 2), Element(1, 2), Element(2, 2)
		);
}


//********************************************************************************
// Get a new matrix from this matrix transposed.
//********************************************************************************

LIBMATHS_INLINE Matrix33 Matrix33::Transposed(void) const
{
	return Matrix33
		(
		Element(0, 0), Element(1, 0), Element(2, 0),
		Element(0, 1), Element(1, 1), Element(2, 1),
		Element(0, 2), Element(1, 2), Element(2, 2)
		);
}


//********************************************************************************
// Apply this matrix to an array of 2 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::TransformPoint2f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1];

	pDestination[0] = Element(0, 0) * x + Element(1, 0) * y + Element(2, 0);
	pDestination[1] = Element(0, 1) * x + Element(1, 1) * y + Element(2, 1);
}


//********************************************************************************
// Apply this matrix to 2D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::TransformPoint(const Vector2& source, Vector2& destination) const
{
	TransformPoint2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 2D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix33::TransformPoint(const Vector2& source) const
{
	Vector2 result;

	TransformPoint(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 2 floats (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::TransformVector2f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1];

	pDestination[0] = Element(0, 0) * x + Element(1, 0) * y;
	pDestination[1] = Element(0, 1) * x + Element(1, 1) * y;
}


//********************************************************************************
// Apply this matrix to a 2D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::TransformVector(const Vector2& source, Vector2& destination) const
{
	TransformVector2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 2D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix33::TransformVector(const Vector2& source) const
{
	Vector2 result;

	TransformVector(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 3 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::Transform3f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1], z = pSource[2];

	pDestination[0] = Element(0, 0) * x + Element(1, 0) * y + Element(2, 0) * z;
	pDestination[1] = Element(0, 1) * x + Element(1, 1) * y + Element(2, 1) * z;
	pDestination[2] = Element(0, 2) * x + Element(1, 2) * y + Element(2, 2) * z;
}


//********************************************************************************
// Apply this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::Transform(const Vector3x& source, Vector3x& destination) const
{
	Transform3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix33::Transform(const Vector3x& source) const
{
	Vector3x result;

	Transform(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::Transform(const Vector3& source, Vector3& destination) const
{
	Transform3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix33::Transform(const Vector3& source) const
{
	Vector3 result;
	
	Transform(source, result);
	
	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 2 floats (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransformPoint2f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0] - Element(2, 0), y = pSource[1] - Element(2, 1);

	pDestination[0] = Element(0, 0) * x + Element(0, 1) * y;
	pDestination[1] = Element(1, 0) * x + Element(1, 1) * y;
}


//********************************************************************************
// Apply inverse of this matrix to a 2D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransformPoint(const Vector2& source, Vector2& destination) const
{
	InverseTransformPoint2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 2D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix33::InverseTransformPoint(const Vector2& source) const
{
	Vector2 result;

	InverseTransformPoint(source, result);

	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 2 floats (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransformVector2f(const F32* pSource, F32* pDestination) const
{
	F32 ix = pSource[0], iy = pSource[1];

	pDestination[0] = Element(0, 0) * ix + Element(0, 1) * iy;
	pDestination[1] = Element(1, 0) * ix + Element(1, 1) * iy;
}


//********************************************************************************
// Apply inverse of this matrix to a 2D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransformVector(const Vector2& source, Vector2& destination) const
{
	InverseTransformVector2f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 2D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix33::InverseTransformVector(const Vector2& source) const
{
	Vector2 result;

	InverseTransformVector(source, result);

	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 3 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransform3f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1], z = pSource[2];

	pDestination[0] = Element(0, 0) * x + Element(0, 1) * y + Element(0, 2) * z;
	pDestination[1] = Element(1, 0) * x + Element(1, 1) * y + Element(1, 2) * z;
	pDestination[2] = Element(2, 0) * x + Element(2, 1) * y + Element(2, 2) * z;
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransform(const Vector3x& source, Vector3x& destination) const
{
	InverseTransform3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix33::InverseTransform(const Vector3x& source) const
{
	Vector3x result;
	
	InverseTransform(source, result);
	
	return result;
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix33::InverseTransform(const Vector3& source, Vector3& destination) const
{
	InverseTransform3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 3D vector.
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix33::InverseTransform(const Vector3& source) const
{
	Vector3 result;

	InverseTransform(source, result);

	return result;
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector2 Matrix33::operator * (const Vector2& rhs) const
{
	return TransformPoint(rhs);
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix33::operator * (const Vector3x& rhs) const
{
	return Transform(rhs);
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix33::operator * (const Vector3& rhs) const
{
	return Transform(rhs);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector2 Vector2::operator * (const Matrix33& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3x Vector3x::operator * (const Matrix33& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3 Vector3::operator * (const Matrix33& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Multiply (transform) this vector by a 3x3 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::operator *= (const Matrix33& rhs)
{
	return *this = rhs * (*this);
}


//********************************************************************************
// Compare two matrices for equality.
//********************************************************************************

LIBMATHS_INLINE Bool Matrix33::operator == (const Matrix33& matrix) const
{
	for (S32 i = 0; i < 3; i++)
	{
		for (S32 j = 0; j < 3; j++)
		{
			if (!Approx(matrix.Element(i, j), Element(i, j), Vector3x::Epsilon()))
			{
				return false;
			}
		}
	}

	return true;
}


//********************************************************************************
// Get a new matrix that is the negative of this matrix
//********************************************************************************

LIBMATHS_INLINE Matrix33 Matrix33::operator - (void) const
{
	return Matrix33
		(
		-Element(0, 0), -Element(0, 1), -Element(0, 2),
		-Element(1, 0), -Element(1, 1), -Element(1, 2),
		-Element(2, 0), -Element(2, 1), -Element(2, 2)
		);
}


//********************************************************************************
// Get a new matrix from another matrix added to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33 Matrix33::operator + (const Matrix33& rhs)
{
	Matrix33 result;

	for (S32 row = 0; row < 3; row++)
	{
		for (S32 col = 0; col < 3; col++)
		{
			result.Element(row, col) = Element(row, col) + rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from another matrix subtracted from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33 Matrix33::operator - (const Matrix33& rhs)
{
	Matrix33 result;

	for (S32 row = 0; row < 3; row++)
	{
		for (S32 col = 0; col < 3; col++)
		{
			result.Element(row, col) = Element(row, col) - rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from this matrix multiplied by (concatenated with) another
// matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33 Matrix33::operator * (const Matrix33& rhs)
{
	Matrix33 result(*this);

	result *= rhs;

	return result;
}


//********************************************************************************
// Add another matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33& Matrix33::operator += (const Matrix33& rhs)
{
	for (S32 row = 0; row < 3; row++)
	{
		for (S32 col = 0; col < 3; col++)
		{
			Element(row, col) += rhs.Element(row, col);
		}
	}

	return *this;
}


//********************************************************************************
// Subtract another matrix from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33& Matrix33::operator -= (const Matrix33& rhs)
{
	for (S32 row = 0; row < 3; row++)
	{
		for (S32 col = 0; col < 3; col++)
		{
			Element(row, col) -= rhs.Element(row, col);
		}
	}

	return *this;
}


#if LIBMATHS_WIN32_MSVC

//********************************************************************************
// Macros for doing dot products with 3 elements.
//********************************************************************************

#undef ELEMENT_OFFSET
#define ELEMENT_OFFSET(i, j) (((i * 3) + j) << 2)

#undef LOAD_MULTIPLY
#define LOAD_MULTIPLY(k, row, column) \
	__asm fld dword ptr [eax + ELEMENT_OFFSET(row, k)] \
	__asm fmul dword ptr [ebx + ELEMENT_OFFSET(k, column)]

#undef DOT_PRODUCT_3
#define DOT_PRODUCT_3(row, column) \
	LOAD_MULTIPLY(0, row, column) \
	LOAD_MULTIPLY(1, row, column) \
	LOAD_MULTIPLY(2, row, column) \
	__asm faddp st(1), st(0) \
	__asm faddp st(1), st(0) \
	__asm fstp dword ptr [ecx + ELEMENT_OFFSET(row, column)]

#endif // LIBMATHS_WIN32_MSVC


//********************************************************************************
// Multiply this matrix by (concatenated with) another matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix33& Matrix33::operator *= (const Matrix33& rhs)
{

#if LIBMATHS_WIN32_MSVC

	Matrix33 result;
	Matrix33* pResult = &result;

	__asm mov eax, dword ptr this
	__asm mov ebx, dword ptr rhs
	__asm mov ecx, dword ptr pResult

	DOT_PRODUCT_3(0, 0) DOT_PRODUCT_3(0, 1) DOT_PRODUCT_3(0, 2)
	DOT_PRODUCT_3(1, 0) DOT_PRODUCT_3(1, 1) DOT_PRODUCT_3(1, 2)
	DOT_PRODUCT_3(2, 0) DOT_PRODUCT_3(2, 1) DOT_PRODUCT_3(2, 2)

	return (*this = result);

#endif // LIBMATHS_WIN32_MSVC

#if LIBMATHS_GENERIC

	Matrix33 result;
	F32 t = 0.0f;

	for (S32 i = 0; i < 3; i++)
	{
		for (S32 j = 0; j < 3; j++)
		{
			t = 0.0f;

			for (S32 k = 0; k < 3; k++)
			{
				t += Element(i, k) * rhs.Element(k, j);
			}

			result.Element(i, j) = t;
		}
	}

	return (*this = result);

#endif // LIBMATHS_GENERIC

}

#undef DOT_PRODUCT_3
#undef LOAD_MULTIPLY
#undef ELEMENT_OFFSET
