//********************************************************************************
//
// 3x3 matrix class. Representing the top 3x3 section of what would normally be a
// 4x4 matrix. Used to transform 3 component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "quaternion.h"


namespace libmaths
{

	//********************************************************************************
	// 3x3 matrix class.
	// Used to transform 2 and 3 component vectors.
	//********************************************************************************

	class Matrix33of44
	{

	private:

#if LIBMATHS_COLUMN_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[column][row]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[column][row]; }

#elif LIBMATHS_ROW_MAJOR_MATRICES

		F32 Element(S32 row, S32 column) const
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[row][column]; }

		F32& Element(S32 row, S32 column)
		{ LIBMATHS_ASSERT(U32(row) < 3 && U32(column) < 3); return _array[row][column]; }

#else

#error No matrix format defined

#endif

	public:

		// A fixed matrix of all zeroes (0, 0, 0).
		static const Matrix33of44& ZERO(void);

		// A fixed matrix of identity (zeroes with 1, 1, 1 across the diagonal).
		static const Matrix33of44& IDENTITY(void);

		// Default constructor.
		Matrix33of44(void);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		explicit Matrix33of44(const F32* pValues);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix33of44(const F32 values[3][3]);

		// Construct from 9 floats.
		// The data should be in row-major format. Conversion to other formats is handled
		// transparently.
		Matrix33of44(
			F32 m00, F32 m01, F32 m02,
			F32 m10, F32 m11, F32 m12,
			F32 m20, F32 m21, F32 m22);

		// Construct from a 4x4 matrix.
		// The fourth column and translation are discarded in the process.
		Matrix33of44(const Matrix44& matrix);

		// Construct from a quaternion.
		explicit Matrix33of44(const Quaternion& rotation)
		{ FromQuaternion(rotation); }

		// Assignment from a quaternion.
		Matrix33of44& operator = (const Quaternion& rhs)
		{ FromQuaternion(rhs); return (*this); }

		// Access / alter matrix elements by array index.
		F32& operator [] (const S32 i)
		{ return Element(i / 3, i % 3); }

		// Access elements by array index.
		F32 operator [] (const S32 i) const
		{ return Element(i / 3, i % 3); }

		// Access / alter matrix elements by row and column.
		F32& operator () (const S32 row, const S32 column)
		{ return Element(row, column); }

		// Access matrix elements by row and column.
		F32 operator () (const S32 row, const S32 column) const
		{ return Element(row, column); }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void) const
		{ return &_array[0][0]; }

		// Access matrix elements as a constant array of floats.
		operator const F32* (void)
		{ return &_array[0][0]; }

		// Set this matrix to zero.
		Matrix33of44& Zero(void)
		{ return ((*this) = Matrix33of44::ZERO()); }

		// Set this matrix to identity (including translation).
		Matrix33of44& Identity(void)
		{ return ((*this) = Matrix33of44::IDENTITY()); }

		// Get the x axis as a 3D vector.
		Vector3x XAxis(void) const
		{ return Vector3x(Element(0, 0), Element(0, 1), Element(0, 2)); }

		// Set the x axis from a 3D vector.
		void XAxis(const Vector3x& axis);

		// Set the x axis from 3 floats.
		void XAxis(F32 x, F32 y, F32 z)
		{ XAxis(Vector3x(x, y, z)); }

		// Get the y axis as a 3D vector.
		Vector3x YAxis(void) const
		{ return Vector3x(Element(1, 0), Element(1, 1), Element(1, 2)); }

		// Set the y axis from a 3D vector.
		void YAxis(const Vector3x& axis);

		// Set the y axis from 3 floats.
		void YAxis(F32 x, F32 y, F32 z)
		{ YAxis(Vector3x(x, y, z)); }

		// Get the z axis as a 3D vector.
		Vector3x ZAxis(void) const
		{ return Vector3x(Element(2, 0), Element(2, 1), Element(2, 2)); }

		// Set the z axis from a 3D vector.
		void ZAxis(const Vector3x& axis);

		// Set the z axis from 3 floats.
		void ZAxis(F32 x, F32 y, F32 z)
		{ ZAxis(Vector3x(x, y, z)); }

		// Return indexed row of matrix as a 3D vector.
		Vector3x Row(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 3); return Vector3x((*this)(i, 0), (*this)(i, 1), (*this)(i, 2)); }

		// Return indexed column of matrix as 3D vector.
		Vector3x Column(S32 i) const
		{ LIBMATHS_ASSERT(U32(i) < 3); return Vector3x((*this)(0, i), (*this)(1, i), (*this)(2, i)); }

		// Build a transformation from a quaternion.
		Matrix33of44& FromQuaternion(const Quaternion& rotation);

		// Convert this matrix into a quaternion representing the same rotation.
		Quaternion ToQuaternion(void) const;

		// Build a transformation representing a rotation of angle radians anti-clockwise
		// about the axis defined by axis. Angles of zero will produce no transform.
		Matrix33of44& FromAxisAngle(const Vector3x& axis, F32 angle);

		// Decompose the rotation part of this matrix into a rotation of radians
		// anti-clockwise about an axis.
		void ToAxisAngle(Vector3x& axis, F32& angle) const;

		// Build a transformation from a direction vector with a specified "up" vector.
		// NOTE: This assumes right-handed orientation.
		Matrix33of44& FromDirectionRH(const Vector3x& direction, const Vector3x& up = Vector3x::ZAXIS());

		// Build a transformation from a direction vector with a specified "up" vector.
		// NOTE: This assumes left-handed orientation.
		Matrix33of44& FromDirectionLH(const Vector3x& direction, const Vector3x& up = Vector3x::YAXIS());

		// Build a transformation from an anti-clockwise angle around the x axis.
		Matrix33of44& FromXAngle(F32 angle);

		// Build a transformation from an anti-clockwise angle around the y axis.
		Matrix33of44& FromYAngle(F32 angle);

		// Build a transformation from an anti-clockwise angle around the z axis.
		Matrix33of44& FromZAngle(F32 angle);

		// Build a transformation about X then Y then Z by the given anti-clockwise angles.
		Matrix33of44& FromXYZAngles(const Vector3x& angles);

		// Build a transformation about X then Z then Y by the given anti-clockwise angles.
		Matrix33of44& FromXZYAngles(const Vector3x& angles);

		// Build a transformation about Y then X then Z by the given anti-clockwise angles.
		Matrix33of44& FromYXZAngles(const Vector3x& angles);

		// Build a transformation about Y then Z then X by the given anti-clockwise angles.
		Matrix33of44& FromYZXAngles(const Vector3x& angles);

		// Build a transformation about Z then X then Y by the given anti-clockwise angles.
		Matrix33of44& FromZXYAngles(const Vector3x& angles);

		// Build a transformation about Z then Y then X by the given anti-clockwise angles.
		Matrix33of44& FromZYXAngles(const Vector3x& angles);

		// Build a transformation about Z then Y then X by the given anti-clockwise angles.
		Matrix33of44& FromAngles(const Vector3x& angles)
		{ return FromZYXAngles(angles); }

		// Create a matrix from pitch, yaw and roll angles.
		// NOTE: This assumes right-handed orientation.
		Matrix33of44& FromPitchYawRollRH(F32 pitch, F32 yaw, F32 roll)
		{ FromXYZAngles(Vector3x(pitch, roll, yaw)); return *this; }

		// Create a matrix from pitch, yaw and roll angles.
		// NOTE: This assumes left-handed orientation.
		Matrix33of44& FromPitchYawRollLH(F32 pitch, F32 yaw, F32 roll)
		{ FromXYZAngles(Vector3x(pitch, yaw, roll)); return *this; }

		// Build a transformation from a scale value.
		Matrix33of44& FromScale(F32 scale);

		// Build a transformation from separate x, y and z scale values.
		Matrix33of44& FromScale(F32 x, F32 y, F32 z);

		// Post-concatenate a x axis rotation matrix to this matrix.
		void PostCatX(F32 angle);

		// Post-concatenate a y axis rotation matrix to this matrix.
		void PostCatY(F32 angle);

		// Post-concatenate a z axis rotation matrix to this matrix.
		void PostCatZ(F32 angle);

		// Post-concatenate a scaling to this matrix.
		void PostCatScale(F32 x, F32 y, F32 z);

		// Scale this matrix uniformly.
		Matrix33of44& Scale(F32 scale);

		// Scale this matrix non-uniformly.
		Matrix33of44& Scale(F32 x, F32 y, F32 z);

		// Get a new matrix from this matrix uniformly scaled.
		Matrix33of44 Scaled(F32 scale) const
		{ return Matrix33of44(*this).Scale(scale); }

		// Get a new matrix from this matrix non-uniformly scaled.
		Matrix33of44 Scaled(F32 x, F32 y, F32 z) const
		{ return Matrix33of44(*this).Scale(x, y, z); }

		// Transpose this matrix's rows and columns.
		Matrix33of44& Transpose(void);

		// Get a new matrix from this matrix transposed.
		Matrix33of44 Transposed(void) const;

		// Invert this matrix, producing an inverse transformation.
		Matrix33of44& Invert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation.
		Matrix33of44 Inverse(void) const;

		// Invert this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix33of44& OrthonormalInvert(void);

		// Get a new matrix from the inverse of this matrix, producing an inverse transformation (assumes this matrix is an orthonormal transformation).
		Matrix33of44 OrthonormalInverse(void) const;

		// Orthonormalise this matrix using the Gram-Schmidt algorithm.
		// NOTE: Only the upper 3x3 sub-matrix is affected.
		Matrix33of44& Orthonormalise(void);

		// Get the determinant of this matrix.
		F32 Determinant(void) const;

		// Build this matrix as an interpolation between the two given matrices,
		// with possible extra 360 degree spins.
		Matrix33of44& Lerp(F32 t, Matrix33of44& m0, Matrix33of44& m1, U32 numSpins = 0);

		// Apply this matrix to an array of 3 floats.
		void TransformVector3f(const F32* pSource, F32* pDestination) const;

		// Apply this matrix to a 3D vector.
		void TransformVector(const Vector3x& source, Vector3x& destination) const;

		// Apply this matrix to a 3D vector.
		Vector3x TransformVector(const Vector3x& source) const;

		// Apply this matrix to a 3D vector.
		void TransformVector(const Vector3& source, Vector3& destination) const;

		// Apply this matrix to a 3D vector.
		Vector3 TransformVector(const Vector3& source) const;

		// Apply this matrix to an array of 3 floats and scale the result.
		void SkinVector3f(const F32* pSource, F32* pDestination, F32* pScale) const;

		// Apply this matrix to a 3D vector and scale the result.
		void SkinVector(const Vector3x& source, Vector3x& destination, F32 scale) const;

		// Apply this matrix to a 3D vector and scale the result.
		Vector3x SkinVector(const Vector3x& source, F32 scale) const;

		// Apply this matrix to a 3D vector and scale the result.
		void SkinVector(const Vector3& source, Vector3& destination, F32 scale) const;

		// Apply this matrix to a 3D vector scale the result.
		Vector3 SkinVector(const Vector3& source, F32 scale) const;

		// Apply inverse of this matrix to an array of 3 floats.
		void InverseTransformVector3f(const F32* pSource, F32* pDestination) const;

		// Apply inverse of this matrix to a 3D vector.
		void InverseTransformVector(const Vector3x& source, Vector3x& destination) const;

		// Apply inverse of this matrix to a 3D vector.
		Vector3x InverseTransformVector(const Vector3x& source) const;

		// Apply inverse of this matrix to a 3D vector.
		void InverseTransformVector(const Vector3& source, Vector3& destination) const;

		// Apply inverse of this matrix to a 3D vector.
		Vector3 InverseTransformVector(const Vector3& source) const;

		// Compare two matrices for equality.
		Bool operator == (const Matrix33of44& rhs) const;

		// Compare two matrices for inequality.
		Bool operator != (const Matrix33of44& rhs) const
		{ return (!(*this == rhs)); }

		// Get a new matrix that is the negative of this matrix.
		Matrix33of44 operator - (void) const;

		// Get a new matrix from another matrix added to this matrix.
		Matrix33of44 operator + (const Matrix33of44& rhs);

		// Get a new matrix from another matrix subtracted from this matrix.
		Matrix33of44 operator - (const Matrix33of44& rhs);

		// Get a new matrix from this matrix multiplied by (concatenated with) another matrix.
		Matrix33of44 operator * (const Matrix33of44& rhs);

		// Get a new matrix from this matrix scaled by a value.
		Matrix33of44 operator * (F32 scale) const
		{ return Scaled(scale); }

		// Multiply a 3D vector by this matrix.
		Vector3x operator * (const Vector3x& vector) const;

		// Multiply a 3D vector by this matrix.
		Vector3 operator * (const Vector3& vector) const;

		// Get a new matrix from a matrix scaled by a value.
		friend Matrix33of44 operator * (F32 lhs, Matrix33of44& rhs)
		{ return rhs.Scaled(lhs); }

		// Add another matrix to this matrix.
		Matrix33of44& operator += (const Matrix33of44& rhs);

		// Subtract another matrix from this matrix.
		Matrix33of44& operator -= (const Matrix33of44& rhs);

		// Multiply this matrix by (concatenated with) another matrix.
		Matrix33of44& operator *= (const Matrix33of44& rhs);

		// Multiply (scale) this matrix by a value.
		Matrix33of44& operator *= (F32 scale)
		{ Scale(scale); return *this; }

		// Divide (scale) this matrix by a value.
		Matrix33of44& operator /= (F32 scale)
		{ Scale(1.0f / scale); return *this; }

	private:

		operator F32* (void);
		operator F32** (void);

		// 9 F32 components.
		F32 _array[3][3];
	};

	// Now include the inline definitions.

#include "matrix33of44.inl"

// namespace libmaths
}

#pragma managed(pop)
