//********************************************************************************
//
// 4x4 matrix class. Used for homogenous 3D transformations with 3 and 4
// component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Defines Affecting this file:
//
// DEFAULT_INITIALISERS:
// To set default values to components when default constructor is used.
//
// LIBMATHS_COLUMN_MAJOR_MATRICES
// Use column major storage and operations.
//
// LIBMATHS_ROW_MAJOR_MATRICES
// Use row major storage and operations (the default).
//********************************************************************************


//********************************************************************************
// A fixed matrix of all zeroes (0, 0, 0, 0).
//********************************************************************************

LIBMATHS_INLINE const Matrix44& Matrix44::ZERO(void)
{
	static Matrix44 matrix(
		0.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f);

	return matrix;
}


//********************************************************************************
// A fixed matrix of identity (zeroes with 1, 1, 1, 1 across the diagonal).
//********************************************************************************

LIBMATHS_INLINE const Matrix44& Matrix44::IDENTITY(void)
{
	static Matrix44 matrix(
		1.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 1.0f);

	return matrix;
}


//********************************************************************************
// Default constructor.
// N.B. Does not initialise any data members unless DEFAULT_INITIALISERS is defined.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44(void)
{
#if defined(DEFAULT_INITIALISERS)
	Identity();
#endif
}


//********************************************************************************
// Construct from 16 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44(const F32* pValues)
{
	Element(0, 0) = pValues[0 * 4 + 0]; Element(0, 1) = pValues[0 * 4 + 1]; Element(0, 2) = pValues[0 * 4 + 2]; Element(0, 3) = pValues[0 * 4 + 3];
	Element(1, 0) = pValues[1 * 4 + 0]; Element(1, 1) = pValues[1 * 4 + 1]; Element(1, 2) = pValues[1 * 4 + 2]; Element(1, 3) = pValues[1 * 4 + 3];
	Element(2, 0) = pValues[2 * 4 + 0]; Element(2, 1) = pValues[2 * 4 + 1]; Element(2, 2) = pValues[2 * 4 + 2]; Element(2, 3) = pValues[2 * 4 + 3];
	Element(3, 0) = pValues[3 * 4 + 0]; Element(3, 1) = pValues[3 * 4 + 1]; Element(3, 2) = pValues[3 * 4 + 2]; Element(3, 3) = pValues[3 * 4 + 3];
}


//********************************************************************************
// Construct from 16 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44(const F32 values[4][4])
{
	Element(0, 0) = values[0][0]; Element(0, 1) = values[0][1]; Element(0, 2) = values[0][2]; Element(0, 3) = values[0][3];
	Element(1, 0) = values[1][0]; Element(1, 1) = values[1][1]; Element(1, 2) = values[1][2]; Element(1, 3) = values[1][3];
	Element(2, 0) = values[2][0]; Element(2, 1) = values[2][1]; Element(2, 2) = values[2][2]; Element(2, 3) = values[2][3];
	Element(3, 0) = values[3][0]; Element(3, 1) = values[3][1]; Element(3, 2) = values[3][2]; Element(3, 3) = values[3][3];
}


//********************************************************************************
// Construct from 16 floats.
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44( F32 m00, F32 m01, F32 m02, F32 m03,
										F32 m10, F32 m11, F32 m12, F32 m13,
										F32 m20, F32 m21, F32 m22, F32 m23,
										F32 m30, F32 m31, F32 m32, F32 m33)
{
	Element(0, 0) = m00; Element(0, 1) = m01; Element(0, 2) = m02; Element(0, 3) = m03;
	Element(1, 0) = m10; Element(1, 1) = m11; Element(1, 2) = m12; Element(1, 3) = m13;
	Element(2, 0) = m20; Element(2, 1) = m21; Element(2, 2) = m22; Element(2, 3) = m23;
	Element(3, 0) = m30; Element(3, 1) = m31; Element(3, 2) = m32; Element(3, 3) = m33;
}


//********************************************************************************
// Construct from 12 floats.
// The first 9 values represent the upper left 3x3 rotation matrix and the final
// 3 represent the lower left 3x1 translation vector.
// The fourth column is initialised to (0, 0, 0, 1)
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44( F32 m00, F32 m01, F32 m02,
										F32 m10, F32 m11, F32 m12,
										F32 m20, F32 m21, F32 m22,
										F32 m30, F32 m31, F32 m32)
{
	Element(0, 0) = m00; Element(0, 1) = m01; Element(0, 2) = m02; Element(0, 3) = 0.0f;
	Element(1, 0) = m10; Element(1, 1) = m11; Element(1, 2) = m12; Element(1, 3) = 0.0f;
	Element(2, 0) = m20; Element(2, 1) = m21; Element(2, 2) = m22; Element(2, 3) = 0.0f;
	Element(3, 0) = m30; Element(3, 1) = m31; Element(3, 2) = m32; Element(3, 3) = 1.0f;
}


//********************************************************************************
// Construct from 9 floats.
// The 9 values represent the upper left 3x3 rotation matrix.
// The fourth column is initialised to (0, 0, 0, 1)
// The translation is initialised to (0, 0, 0)
// The data should be in row-major format. Conversion to other formats is handled
// transparently.
//********************************************************************************

LIBMATHS_INLINE Matrix44::Matrix44( F32 m00, F32 m01, F32 m02,
										F32 m10, F32 m11, F32 m12,
										F32 m20, F32 m21, F32 m22)
{
	Element(0, 0) = m00; Element(0, 1) = m01; Element(0, 2) = m02; Element(0, 3) = 0.0f;
	Element(1, 0) = m10; Element(1, 1) = m11; Element(1, 2) = m12; Element(1, 3) = 0.0f;
	Element(2, 0) = m20; Element(2, 1) = m21; Element(2, 2) = m22; Element(2, 3) = 0.0f;
	Element(3, 0) = 0.f; Element(3, 1) = 0.f; Element(3, 2) = 0.f; Element(3, 3) = 1.0f;
}


//********************************************************************************
// Set this matrix to identity (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::IdentityRotation(void)
{
	XAxis(1.0f, 0.0f, 0.0f);
	YAxis(0.0f, 1.0f, 0.0f);
	ZAxis(0.0f, 0.0f, 1.0f);
}


//********************************************************************************
// Set the x axis from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::XAxis(const Vector3x& axis)
{
	Element(0, 0) = axis.X();
	Element(0, 1) = axis.Y();
	Element(0, 2) = axis.Z();
}


//********************************************************************************
// Set the y axis from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::YAxis(const Vector3x& axis)
{
	Element(1, 0) = axis.X();
	Element(1, 1) = axis.Y();
	Element(1, 2) = axis.Z();
}


//********************************************************************************
// Set the z axis from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::ZAxis(const Vector3x& axis)
{
	Element(2, 0) = axis.X();
	Element(2, 1) = axis.Y();
	Element(2, 2) = axis.Z();
}


//********************************************************************************
// Set the w axis from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::Perspective(const Vector3x& axis)
{
	Element(0, 3) = axis.X();
	Element(1, 3) = axis.Y();
	Element(2, 3) = axis.Z();
}


//********************************************************************************
// Set the translation from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::Translation(const Vector3x& translation)
{
	Element(3, 0) = translation.X();
	Element(3, 1) = translation.Y();
	Element(3, 2) = translation.Z();
}


//********************************************************************************
// Post-concatenate a x axis rotation matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::PostCatX(F32 angle)
{
	F32 c = Cos(angle);
	F32 s = Sin(angle);

	for (S32 i = 0; i < 4; i++)
	{
		F32 t = Element(i, 1);
		Element(i, 1) = t * c - Element(i, 2) * s;
		Element(i, 2) = t * s + Element(i, 2) * c;
	}
}


//********************************************************************************
// Post-concatenate a y axis rotation matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::PostCatY(F32 angle)
{
	F32 c = Cos(angle);
	F32 s = Sin(angle);

	for (S32 i = 0; i < 4; i++)
	{
		F32 t = Element(i, 0);
		Element(i, 0) = t * c + Element(i, 2) * s;
		Element(i, 2) = Element(i, 2) * c - t * s;
	}
}


//********************************************************************************
// Post-concatenate a z axis rotation matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::PostCatZ(F32 angle)
{
	F32 c = Cos(angle);
	F32 s = Sin(angle);

	for (S32 i = 0; i < 4; i++)
	{
		F32 t = Element(i, 0);
		Element(i, 0) = t * c - Element(i, 1) * s;
		Element(i, 1) = t * s + Element(i, 1) * c;
	}
}


//********************************************************************************
// Post-concatenate a scaling to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::PostCatScale(F32 x, F32 y, F32 z, F32 w)
{
	for (S32 i = 0; i < 4; i++)
	{
		Element(i, 0) *= x;
		Element(i, 1) *= y;
		Element(i, 2) *= z;
		Element(i, 3) *= w;
	}
}


//********************************************************************************
// Post-concatenate a translation to this matrix.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::PostCatTranslate(F32 x, F32 y, F32 z)
{
	for (S32 i = 0; i < 4; i++)
	{
		Element(i, 0) += Element(i, 3) * x;
		Element(i, 1) += Element(i, 3) * y;
		Element(i, 2) += Element(i, 3) * z;
	}
}


//********************************************************************************
// Transpose this matrix's rows and columns.
//********************************************************************************

LIBMATHS_INLINE Matrix44& Matrix44::Transpose(void)
{
	return *this = Matrix44
		(
		Element(0, 0), Element(1, 0), Element(2, 0), Element(3, 0),
		Element(0, 1), Element(1, 1), Element(2, 1), Element(3, 1),
		Element(0, 2), Element(1, 2), Element(2, 2), Element(3, 2),
		Element(0, 3), Element(1, 3), Element(2, 3), Element(3, 3)
		);
}


//********************************************************************************
// Get a new matrix from this matrix transposed.
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::Transposed(void) const
{
	return Matrix44
		(
		Element(0, 0), Element(1, 0), Element(2, 0), Element(3, 0),
		Element(0, 1), Element(1, 1), Element(2, 1), Element(3, 1),
		Element(0, 2), Element(1, 2), Element(2, 2), Element(3, 2),
		Element(0, 3), Element(1, 3), Element(2, 3), Element(3, 3)
		);
}


//********************************************************************************
// Transpose this matrix's rows and columns, the rotational part only.
//********************************************************************************

LIBMATHS_INLINE Matrix44& Matrix44::TransposeRotation(void)
{
	return *this = Matrix44
		(
		Element(0, 0), Element(1, 0), Element(2, 0), Element(0, 3),
		Element(0, 1), Element(1, 1), Element(2, 1), Element(1, 3),
		Element(0, 2), Element(1, 2), Element(2, 2), Element(2, 3),
		Element(3, 0), Element(3, 1), Element(3, 2), Element(3, 3)
		);
}


//********************************************************************************
// Get a new matrix from this matrix transposed, the rotational part only.
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::TransposedRotation(void) const
{
	return Matrix44
		(
		Element(0, 0), Element(1, 0), Element(2, 0), Element(0, 3),
		Element(0, 1), Element(1, 1), Element(2, 1), Element(1, 3),
		Element(0, 2), Element(1, 2), Element(2, 2), Element(2, 3),
		Element(3, 0), Element(3, 1), Element(3, 2), Element(3, 3)
		);
}


//********************************************************************************
// Apply this matrix to an array of 3 floats (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformPoint3f(const F32* pSource, F32* pDestination) const
{

#if LIBMATHS_GENERIC

	// Original non FPU code copy input so input and output can be the same pointer.

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz + Element(3, 0);
	pDestination[1] = Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz + Element(3, 1);
	pDestination[2] = Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz + Element(3, 2);

#else // LIBMATHS_GENERIC

#if LIBMATHS_WIN32_MSVC

#if LIBMATHS_COLUMN_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+36]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+24]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fadd dword ptr [eax+12]
		fstp dword ptr [ecx]
		fadd dword ptr [eax+28]
		fstp dword ptr [ecx+4]
		fadd dword ptr [eax+44]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_COLUMN_MAJOR_MATRICES

#if LIBMATHS_ROW_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+24]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+36]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fadd dword ptr [eax+48]
		fstp dword ptr [ecx]
		fadd dword ptr [eax+52]
		fstp dword ptr [ecx+4]
		fadd dword ptr [eax+56]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_ROW_MAJOR_MATRICES

#else // LIBMATHS_WIN32_MSVC

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz + Element(3, 0);
	pDestination[1] = Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz + Element(3, 1);
	pDestination[2] = Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz + Element(3, 2);

#endif // LIBMATHS_WIN32_MSVC

#endif // LIBMATHS_GENERIC

}


//********************************************************************************
// Apply this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformPoint(const Vector3x& source, Vector3x& destination) const
{
	TransformPoint3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::TransformPoint(const Vector3x& source) const
{
	Vector3x result;
	
	TransformPoint(source, result);
	
	return result;
}


//********************************************************************************
// Apply this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformPoint(const Vector3& source, Vector3& destination) const
{
	TransformPoint3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::TransformPoint(const Vector3& source) const
{
	Vector3 result;

	TransformPoint(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 4 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::Transform4f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1], z = pSource[2], w = pSource[3];

	pDestination[0] = Element(0, 0) * x + Element(1, 0) * y + Element(2, 0) * z + Element(3, 0) * w;
	pDestination[1] = Element(0, 1) * x + Element(1, 1) * y + Element(2, 1) * z + Element(3, 1) * w;
	pDestination[2] = Element(0, 2) * x + Element(1, 2) * y + Element(2, 2) * z + Element(3, 2) * w;
	pDestination[3] = Element(0, 3) * x + Element(1, 3) * y + Element(2, 3) * z + Element(3, 3) * w;
}


//********************************************************************************
// Apply this matrix to a homogenous 4D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::Transform(const Vector4& source, Vector4& destination) const
{
	Transform4f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a homogenous 4D vector.
//********************************************************************************

LIBMATHS_INLINE Vector4 Matrix44::Transform(const Vector4& source) const
{
	Vector4 result;

	Transform(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 3 floats (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformVector3f(const F32* pSource, F32* pDestination) const
{

#if LIBMATHS_GENERIC

	// Original non FPU code copy input so input and output can be the same pointer.

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz;
	pDestination[1] = Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz;
	pDestination[2] = Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz;

#else // LIBMATHS_GENERIC

#if LIBMATHS_WIN32_MSVC
	
#if LIBMATHS_COLUMN_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+36]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+24]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fstp dword ptr [ecx]
		fstp dword ptr [ecx+4]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_COLUMN_MAJOR_MATRICES

#if LIBMATHS_ROW_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+24]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+36]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fstp dword ptr [ecx]
		fstp dword ptr [ecx+4]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_ROW_MAJOR_MATRICES

#else // LIBMATHS_WIN32_MSVC

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz;
	pDestination[1] = Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz;
	pDestination[2] = Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz;

#endif // LIBMATHS_WIN32_MSVC

#endif // LIBMATHS_GENERIC

}


//********************************************************************************
// Apply this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformVector(const Vector3x& source, Vector3x& destination) const
{
	TransformVector3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::TransformVector(const Vector3x& source) const
{
	Vector3x result;
	
	TransformVector(source, result);
	
	return result;
}


//********************************************************************************
// Apply this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::TransformVector(const Vector3& source, Vector3& destination) const
{
	TransformVector3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::TransformVector(const Vector3& source) const
{
	Vector3 result;

	TransformVector(source, result);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 3 floats and scale the result (including
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinPoint3f(const F32* pSource, F32* pDestination, F32* pScale) const
{

#if LIBMATHS_GENERIC

	// Original non FPU code copy input so input and output can be the same pointer.

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] += (Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz + Element(3, 0)) * *pScale;
	pDestination[1] += (Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz + Element(3, 1)) * *pScale;
	pDestination[2] += (Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz + Element(3, 2)) * *pScale;

#else // LIBMATHS_GENERIC

#if LIBMATHS_WIN32_MSVC
	
#if LIBMATHS_COLUMN_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination
		mov edx, dword ptr pScale

		fld dword ptr [ebx]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+36]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+24]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST

		fadd dword ptr [eax+12]
		fmul dword ptr [edx]
		fstp dword ptr [ecx]

		fadd dword ptr [eax+28]
		fmul dword ptr [edx]
		fstp dword ptr [ecx+4]

		fadd dword ptr [eax+44]
		fmul dword ptr [edx]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_COLUMN_MAJOR_MATRICES

#if LIBMATHS_ROW_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination
		mov edx, dword ptr pScale

		fld dword ptr [ebx]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+24]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+36]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST

		fadd dword ptr [eax+48]
		fmul dword ptr [edx]
		fadd dword ptr [ecx]
		fstp dword ptr [ecx]

		fadd dword ptr [eax+52]
		fmul dword ptr [edx]
		fadd dword ptr [ecx+4]
		fstp dword ptr [ecx+4]

		fadd dword ptr [eax+56]
		fmul dword ptr [edx]
		fadd dword ptr [ecx+8]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_ROW_MAJOR_MATRICES

#else // LIBMATHS_WIN32_MSVC

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] += (Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz + Element(3, 0)) * *pScale;
	pDestination[1] += (Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz + Element(3, 1)) * *pScale;
	pDestination[2] += (Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz + Element(3, 2)) * *pScale;

#endif // LIBMATHS_WIN32_MSVC

#endif // LIBMATHS_GENERIC

}


//********************************************************************************
// Apply this matrix to a 3D point and scale the result (including
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinPoint(const Vector3x& source, Vector3x& destination, F32 scale) const
{
	SkinPoint3f((F32*)&source, (F32*)&destination, &scale);
}


//********************************************************************************
// Apply this matrix to a 3D point and scale the result (including
// translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::SkinPoint(const Vector3x& source, F32 scale) const
{
	Vector3x result;
	
	SkinPoint(source, result, scale);
	
	return result;
}


//********************************************************************************
// Apply this matrix to a 3D point and scale the result (including
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinPoint(const Vector3& source, Vector3& destination, F32 scale) const
{
	SkinPoint3f((F32*)&source, (F32*)&destination, &scale);
}


//********************************************************************************
// Apply this matrix to a 3D point and scale the result (including
// translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::SkinPoint(const Vector3& source, F32 scale) const
{
	Vector3 result;

	SkinPoint(source, result, scale);

	return result;
}


//********************************************************************************
// Apply this matrix to an array of 3 floats and scale the result (excluding
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinVector3f(const F32* pSource, F32* pDestination, F32* pScale) const
{

#if LIBMATHS_GENERIC

	// Original non FPU code copy input so input and output can be the same pointer.

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] += (Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz) * *pScale;
	pDestination[1] += (Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz) * *pScale;
	pDestination[2] += (Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz) * *pScale;

#else // LIBMATHS_GENERIC

#if LIBMATHS_WIN32_MSVC
	
#if LIBMATHS_COLUMN_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination
		mov edx, dword ptr pScale

		fld dword ptr [ebx]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+36]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+24]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fmul dword ptr [edx]
		fadd dword ptr [ecx]
		fstp dword ptr [ecx]

		fmul dword ptr [edx]
		fadd dword ptr [ecx+4]
		fstp dword ptr [ecx+4]

		fmul dword ptr [edx]
		fadd dword ptr [ecx+8]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_COLUMN_MAJOR_MATRICES

#if LIBMATHS_ROW_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination
		mov edx, dword ptr pScale

		fld dword ptr [ebx]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+24]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+36]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fmul dword ptr [edx]
		fadd dword ptr [ecx]
		fstp dword ptr [ecx]

		fmul dword ptr [edx]
		fadd dword ptr [ecx+4]
		fstp dword ptr [ecx+4]

		fmul dword ptr [edx]
		fadd dword ptr [ecx+8]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_ROW_MAJOR_MATRICES

#else // LIBMATHS_WIN32_MSVC

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] += (Element(0, 0) * ix + Element(1, 0) * iy + Element(2, 0) * iz) * *pScale;
	pDestination[1] += (Element(0, 1) * ix + Element(1, 1) * iy + Element(2, 1) * iz) * *pScale;
	pDestination[2] += (Element(0, 2) * ix + Element(1, 2) * iy + Element(2, 2) * iz) * *pScale;

#endif // LIBMATHS_WIN32_MSVC

#endif // LIBMATHS_GENERIC

}


//********************************************************************************
// Apply this matrix to a 3D vector and scale the result (excluding
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinVector(const Vector3x& source, Vector3x& destination, F32 scale) const
{
	SkinVector3f((F32*)&source, (F32*)&destination, &scale);
}


//********************************************************************************
// Apply this matrix to a 3D vector and scale the result (excluding
// translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::SkinVector(const Vector3x& source, F32 scale) const
{
	Vector3x result;
	
	SkinVector(source, result, scale);
	
	return result;
}


//********************************************************************************
// Apply this matrix to a 3D vector and scale the result (excluding
// translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::SkinVector(const Vector3& source, Vector3& destination, F32 scale) const
{
	SkinVector3f((F32*)&source, (F32*)&destination, &scale);
}


//********************************************************************************
// Apply this matrix to a 3D vector and scale the result (excluding
// translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::SkinVector(const Vector3& source, F32 scale) const
{
	Vector3 result;

	SkinVector(source, result, scale);

	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 3 floats (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformPoint3f(const F32* pSource, F32* pDestination) const
{
	// TODO: Write FPU code for this.

	F32 ix = pSource[0] - Element(3,0), iy = pSource[1] - Element(3,1), iz = pSource[2] - Element(3,2);

	pDestination[0] = Element(0,0) * ix + Element(0,1) * iy + Element(0,2) * iz;
	pDestination[1] = Element(1,0) * ix + Element(1,1) * iy + Element(1,2) * iz;
	pDestination[2] = Element(2,0) * ix + Element(2,1) * iy + Element(2,2) * iz;
}


//********************************************************************************
// Apply inverse of this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformPoint(const Vector3x& source, Vector3x& destination) const
{
	InverseTransformPoint3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::InverseTransformPoint(const Vector3x& source) const
{
	Vector3x result;
	
	InverseTransformPoint(source, result);
	
	return result;
}


//********************************************************************************
// Apply inverse of this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformPoint(const Vector3& source, Vector3& destination) const
{
	InverseTransformPoint3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 3D point (including translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::InverseTransformPoint(const Vector3& source) const
{
	Vector3 result;

	InverseTransformPoint(source, result);

	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 4 floats.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransform4f(const F32* pSource, F32* pDestination) const
{
	F32 x = pSource[0], y = pSource[1], z = pSource[2], w = pSource[3];

	pDestination[0] = Element(0, 0) * x + Element(0, 1) * y + Element(0, 2) * z + Element(0, 3) * w;
	pDestination[1] = Element(1, 0) * x + Element(1, 1) * y + Element(1, 2) * z + Element(1, 3) * w;
	pDestination[2] = Element(2, 0) * x + Element(2, 1) * y + Element(2, 2) * z + Element(2, 3) * w;
	pDestination[3] = Element(3, 0) * x + Element(3, 1) * y + Element(3, 2) * z + Element(3, 3) * w;
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 4D vector.
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransform(const Vector4& source, Vector4& destination) const
{
	InverseTransform4f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a homogenous 4D vector.
//********************************************************************************

LIBMATHS_INLINE Vector4 Matrix44::InverseTransform(const Vector4& source) const
{
	Vector4 result;

	InverseTransform(source, result);

	return result;
}


//********************************************************************************
// Apply inverse of this matrix to an array of 3 floats (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformVector3f(const F32* pSource, F32* pDestination) const
{

#if LIBMATHS_GENERIC

	// Original non FPU code copy input so input and output can be the same pointer.

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(0, 1) * iy + Element(0, 2) * iz;
	pDestination[1] = Element(1, 0) * ix + Element(1, 1) * iy + Element(1, 2) * iz;
	pDestination[2] = Element(2, 0) * ix + Element(2, 1) * iy + Element(2, 2) * iz;

#else // LIBMATHS_GENERIC

#if LIBMATHS_WIN32_MSVC
	
#if LIBMATHS_COLUMN_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+24]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+36]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fstp dword ptr [ecx]
		fstp dword ptr [ecx+4]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_COLUMN_MAJOR_MATRICES

#if LIBMATHS_ROW_MAJOR_MATRICES

	__asm
	{
		mov eax, dword ptr this
		mov ebx, dword ptr pSource
		mov ecx, dword ptr pDestination

		fld dword ptr [ebx]
		fmul dword ptr [eax+32]
		fld dword ptr [ebx]
		fmul dword ptr [eax+16]
		fld dword ptr [ebx]
		fmul dword ptr [eax]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+4]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+20]
		fld dword ptr [ebx+4]
		fmul dword ptr [eax+36]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+8]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+24]
		fld dword ptr [ebx+8]
		fmul dword ptr [eax+40]
		fxch ST(2)
		faddp ST(3),ST
		faddp ST(3),ST
		faddp ST(3),ST
		fstp dword ptr [ecx]
		fstp dword ptr [ecx+4]
		fstp dword ptr [ecx+8]
	};

#endif // LIBMATHS_ROW_MAJOR_MATRICES

#else // LIBMATHS_WIN32_MSVC

	F32 ix = pSource[0], iy = pSource[1], iz = pSource[2];

	pDestination[0] = Element(0, 0) * ix + Element(0, 1) * iy + Element(0, 2) * iz;
	pDestination[1] = Element(1, 0) * ix + Element(1, 1) * iy + Element(1, 2) * iz;
	pDestination[2] = Element(2, 0) * ix + Element(2, 1) * iy + Element(2, 2) * iz;

#endif // LIBMATHS_WIN32_MSVC

#endif // LIBMATHS_GENERIC

}


//********************************************************************************
// Apply inverse of this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformVector(const Vector3x& source, Vector3x& destination) const
{
	InverseTransformVector3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::InverseTransformVector(const Vector3x& source) const
{
	Vector3x result;
	
	InverseTransformVector(source, result);
	
	return result;
}


//********************************************************************************
// Apply inverse of this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE void Matrix44::InverseTransformVector(const Vector3& source, Vector3& destination) const
{
	InverseTransformVector3f((F32*)&source, (F32*)&destination);
}


//********************************************************************************
// Apply inverse of this matrix to a 3D vector (excluding translation).
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::InverseTransformVector(const Vector3& source) const
{
	Vector3 result;

	InverseTransformVector(source, result);

	return result;
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector3x Matrix44::operator * (const Vector3x& vector) const
{
	return TransformPoint(vector);
}


//********************************************************************************
// Multiply the given vector by the matrix
//********************************************************************************

LIBMATHS_INLINE Vector3 Matrix44::operator * (const Vector3& vector) const
{
	return TransformPoint(vector);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3x Vector3x::operator * (const Matrix44& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3 Vector3::operator * (const Matrix44& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Get a new vector from this vector multiplied (transformed) by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector4 Vector4::operator * (const Matrix44& rhs) const
{
	return rhs * (*this);
}


//********************************************************************************
// Multiply (transform) this vector by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3x& Vector3x::operator *= (const Matrix44& rhs)
{
	return *this = rhs * (*this);
}


//********************************************************************************
// Multiply (transform) this vector by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector3& Vector3::operator *= (const Matrix44& rhs)
{
	return *this = rhs * (*this);
}


//********************************************************************************
// Multiply (transform) this vector by a 4x4 matrix.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::operator *= (const Matrix44& rhs)
{
	return *this = rhs * (*this);
}


//********************************************************************************
// Compare two matrices for equality.
//********************************************************************************

LIBMATHS_INLINE Bool Matrix44::operator == (const Matrix44& matrix) const
{
	for (S32 i = 0; i < 4; i++)
	{
		for (S32 j = 0; j < 4; j++)
		{
			if (!Approx(matrix.Element(i, j), Element(i, j), Vector4::Epsilon()))
			{
				return false;
			}
		}
	}

	return true;
}


//********************************************************************************
// Get a new matrix that is the negative of this matrix
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::operator - (void) const
{
	return Matrix44
		(
		-Element(0, 0), -Element(0, 1), -Element(0, 2), -Element(0, 3),
		-Element(1, 0), -Element(1, 1), -Element(1, 2), -Element(1, 3),
		-Element(2, 0), -Element(2, 1), -Element(2, 2), -Element(2, 3),
		-Element(3, 0), -Element(3, 1), -Element(3, 2), -Element(3, 3)
		);
}


//********************************************************************************
// Get a new matrix from another matrix added to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::operator + (const Matrix44& rhs)
{
	Matrix44 result;

	for (S32 row = 0; row < 4; row++)
	{
		for (S32 col = 0; col < 4; col++)
		{
			result.Element(row, col) = Element(row, col) + rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from another matrix subtracted from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::operator - (const Matrix44& rhs)
{
	Matrix44 result;

	for (S32 row = 0; row < 4; row++)
	{
		for (S32 col = 0; col < 4; col++)
		{
			result.Element(row, col) = Element(row, col) - rhs.Element(row, col);
		}
	}

	return result;
}


//********************************************************************************
// Get a new matrix from this matrix multiplied by (concatenated with) another
// matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44 Matrix44::operator * (const Matrix44& rhs)
{
	Matrix44 result(*this);

	result *= rhs;

	return result;
}


//********************************************************************************
// Add another matrix to this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44& Matrix44::operator += (const Matrix44& rhs)
{
	for (S32 row = 0; row < 4; row++)
	{
		for (S32 col = 0; col < 4; col++)
		{
			Element(row, col) += rhs.Element(row, col);
		}
	}

	return *this;
}


//********************************************************************************
// Subtract another matrix from this matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44& Matrix44::operator -= (const Matrix44& rhs)
{
	for (S32 row = 0; row < 4; row++)
	{
		for (S32 col = 0; col < 4; col++)
		{
			Element(row, col) -= rhs.Element(row, col);
		}
	}

	return *this;
}


#if LIBMATHS_WIN32_MSVC

//********************************************************************************
// Macros for doing dot products with 4 elements
//********************************************************************************

#undef ELEMENT_OFFSET
#define ELEMENT_OFFSET(i, j) (((i << 2) + j) << 2)

#undef LOAD_MULTIPLY
#define LOAD_MULTIPLY(k, row, column) \
	__asm fld dword ptr [eax + ELEMENT_OFFSET(row, k)] \
	__asm fmul dword ptr [ebx + ELEMENT_OFFSET(k, column)]

#undef DOT_PRODUCT_4
#define DOT_PRODUCT_4(row, column) \
	LOAD_MULTIPLY(0, row, column) \
	LOAD_MULTIPLY(1, row, column) \
	LOAD_MULTIPLY(2, row, column) \
	LOAD_MULTIPLY(3, row, column) \
	__asm faddp st(1), st(0) \
	__asm fxch st(2) \
	__asm faddp st(1), st(0) \
	__asm faddp st(1), st(0) \
	__asm fstp dword ptr [ecx + ELEMENT_OFFSET(row, column)]

#endif // LIBMATHS_WIN32_MSVC


//********************************************************************************
// Multiply this matrix by (concatenated with) another matrix.
//********************************************************************************

LIBMATHS_INLINE Matrix44& Matrix44::operator *= (const Matrix44& rhs)
{

#if LIBMATHS_WIN32_MSVC

	Matrix44 result;
	Matrix44* pResult = &result;

	__asm mov eax, dword ptr this
	__asm mov ebx, dword ptr rhs
	__asm mov ecx, dword ptr pResult

	DOT_PRODUCT_4(0, 0) DOT_PRODUCT_4(0, 1) DOT_PRODUCT_4(0, 2) DOT_PRODUCT_4(0, 3)
	DOT_PRODUCT_4(1, 0) DOT_PRODUCT_4(1, 1) DOT_PRODUCT_4(1, 2) DOT_PRODUCT_4(1, 3)
	DOT_PRODUCT_4(2, 0) DOT_PRODUCT_4(2, 1) DOT_PRODUCT_4(2, 2) DOT_PRODUCT_4(2, 3)
	DOT_PRODUCT_4(3, 0) DOT_PRODUCT_4(3, 1) DOT_PRODUCT_4(3, 2) DOT_PRODUCT_4(3, 3)

	return (*this = result);

#else // LIBMATHS_WIN32_MSVC

	Matrix44 result;
	F32 t = 0.0f;

	for (S32 i = 0; i < 4; i++)
	{
		for (S32 j = 0; j < 4; j++)
		{
			t = 0.0f;

			for (S32 k = 0; k < 4; k++)
			{
				t += Element(i, k) * rhs.Element(k, j);
			}

			result.Element(i, j) = t;
		}
	}

	return (*this = result);

#endif // LIBMATHS_WIN32_MSVC

}

#undef DOT_PRODUCT_4
#undef LOAD_MULTIPLY
#undef ELEMENT_OFFSET
