//********************************************************************************
//
// Amplitude translation classes returning an amplitude for a parametric value.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "basictypes.h"


namespace libmaths
{

	//********************************************************************************
	// Class to describe a basic amplitude translator from which other implementation
	// classes derive.
	//********************************************************************************

	class ParametricConverter
	{

	public:

		// Construct a ParametricConverter.
		ParametricConverter(Bool looped, F32 time) :
			_looped(looped),
			_time(time),
			_invTime(1.0f / time)
		{ }

		// Destroy a ParametricConverter.
		virtual ~ParametricConverter(void)
		{ }

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value) = 0;

		// Is a value in range?
		Bool IsInRange(F32 value)
		{ return ((_looped) ? true : (value < _time)); }

	protected:

		// Normalise an input value to between 0 and 1, or > 1 if not looped.
		F32 Normalise(F32 value);

		// The time
		F32 Time(void)
		{ return _time; }

		// The inverse of the time
		F32 InvTime(void)
		{ return _invTime; }

	private:

		// Is the translator looped?
		Bool _looped;

		// The time
		F32 _time;

		// The inverse of the time
		F32 _invTime;
	};


	//********************************************************************************
	// A simple linear, attack, sustain, decay amplitude translator.
	//********************************************************************************

	class GeAttackSustainDecayAT : public ParametricConverter
	{

	public:

		// Construct a GeAttackSustainDecayAT.
		GeAttackSustainDecayAT(Bool looped, F32 attack, F32 sustain, F32 decay, F32 amplitude);

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value);

	protected:

		F32 m_Amplitude;
		F32 m_Attack;
		F32 m_InvAttack;
		F32 m_InvDecay;
		F32 m_DecayThreshold;
	};


	//********************************************************************************
	// A simple squared curve, attack, sustain, decay amplitude translator.
	//********************************************************************************

	class GeAttackSustainDecaySqrAT : public GeAttackSustainDecayAT
	{

	public:

		// Construct a GeAttackSustainDecaySqrAT.
		GeAttackSustainDecaySqrAT(Bool looped, F32 attack, F32 sustain, F32 decay, F32 amplitude);

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value);
	};


	//********************************************************************************
	// A simple ease-in cosine-based transition from 0 to 1.
	//********************************************************************************

	class SineEaseInParametricConverter : public ParametricConverter
	{

	public:

		// Construct a SineEaseInParametricConverter.
		SineEaseInParametricConverter(Bool looped, F32 time)
			: ParametricConverter(looped, time)
		{ }

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value);
	};


	//********************************************************************************
	// A simple ease-out cosine-based transition from 0 to 1.
	//********************************************************************************

	class SineEaseOutParametricConverter : public ParametricConverter
	{

	public:

		// Construct a SineEaseOutParametricConverter.
		SineEaseOutParametricConverter(Bool looped, F32 time)
			: ParametricConverter(looped, time)
		{ }

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value);
	};


	//********************************************************************************
	// A simple ease-in / ease-out cosine-based transition from 0 to 1.
	//********************************************************************************

	class SineEaseInOutParametricConverter : public ParametricConverter
	{

	public:

		// Construct a SineEaseInOutParametricConverter.
		SineEaseInOutParametricConverter(Bool looped, F32 time)
			: ParametricConverter(looped, time)
		{ }

		// Translate a value into an amplitude.
		virtual F32 Translate(F32 value);
	};

	// namespace libmaths
}

#pragma managed(pop)
