//********************************************************************************
//
// This class allows planes to be defined and tested for collision with rays
// across the game world.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "ray3.h"
#include "linesegment3.h"


namespace libmaths
{

	//********************************************************************************
	// A three dimensional plane.
	//
	// Supports intersection testing with rays and may optionally use back facing
	// plane culling.
	//********************************************************************************

	class Plane
	{

	public:

		// Construct uninitialised plane.
		Plane(void)
		{ }

		// Construct a plane with the given normal and origin.
		Plane(const Vector3x& origin, const Vector3x& normal) :
			_origin(origin), _normal(normal)
		{ }

		// Construct a plane with the given normal and origin.
		Plane(const Vector3& origin, const Vector3& normal) :
			_origin(origin), _normal(normal)
		{ }

		// Construct a plane passing through the three given points.
		Plane(const Vector3x& v0, const Vector3x& v1, const Vector3x& v2)
		{
			_origin = v0;
			_normal = Vector3x::Normal(v0, v1, v2);
		}

		// Construct a plane passing through the three given points.
		Plane(const Vector3& v0, const Vector3& v1, const Vector3& v2)
		{
			_origin = v0;
			_normal = Vector3x(Vector3::Normal(v0, v1, v2));
		}

		// Access the origin of the plane
		const Vector3x& Origin(void) const
		{ return _origin; }

		// Set the origin of the plane
		void Origin(const Vector3x& origin)
		{ _origin = origin; }

		// Access the normal of the plane
		const Vector3x& Normal(void) const
		{ return _normal; }

		// Set the normal of the plane
		void Normal(const Vector3x& normal)
		{ _normal = normal; }

		// Get D, the plane equation, for the plane.
		F32 D(void) const
		{ return -(_normal * _origin); }

		// Return the perpendicular distance of the given point from the plane.
		// (Negative if behind the plane)
		F32 Distance(const Vector3x& point) const
		{ return (point - Origin()) * Normal(); }

		// Check if the given point lies on or in front of the plane.
		Bool Contains(const Vector3x& point) const
		{ return (Distance(point) >= 0.0f); }

		// Find the point on the plane closest to the given point.
		Vector3x ClosestPoint(const Vector3x& point) const
		{ return point - (Normal() * Distance(point)); }

		// Check if the ray intersects this plane.
		Bool Intersects(const Ray3& ray, Bool twoSided = false) const;

		// Check if the ray intersects this plane.
		Bool Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal, Bool twoSided = false) const;

		// Check if the line intersects this plane.
		Bool Intersects(const LineSegment3& line, Bool twoSided = false) const
		{
			Vector3x point, normal;

			if (Intersects(Ray3(line), point, normal, twoSided))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this plane.
		Bool Intersects(const LineSegment3& line, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided = false) const
		{
			if (Intersects(Ray3(line), point, normal, twoSided))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Flip the plane normal and adjust the distance constant accordingly.
		void Flip(void)
		{ _normal *= -1.0f; }

	private:

		// Plane origin
		Vector3x _origin;

		// Plane normal
		Vector3x _normal;
	};

// namespace libmaths
}

#pragma managed(pop)
