//********************************************************************************
//
// Polar coordinates representing a 2D vector.
// Consists of a radius from the origin and an azimuth angle from the +X axis.
//
// http://en.wikipedia.org/wiki/Polar_coordinates
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"


namespace libmaths
{

	//********************************************************************************
	// Polar coordinates representing a 2D vector.
	// Consists of a radius from the origin and an azimuth angle from the +X axis.
	//********************************************************************************

	class Polar2
	{

	public:

		// Get the epsilon value for equality comparisons.
		static const F32 Epsilon(void)
		{ return 1.0e-5f; }

		// Default constructor.
		Polar2(void) : _radius(0.0f), _azimuth(0.0f)
		{ }

		// Construct from azimuth.
		Polar2(F32 azimuth) : _radius(1.0f), _azimuth(azimuth)
		{ }

		// Construct from azimuth and radius.
		Polar2(F32 azimuth, F32 radius) : _radius(radius), _azimuth(azimuth)
		{ }

		// Assignment from a Vector2, in right-hand coordinate system.
		Polar2& operator = (const Vector2& rhs);

		// Get the radius.
		F32 Radius(void) const
		{ return _radius; }

		// Set the radius.
		void Radius(F32 radius)
		{ _radius = radius; }

		// Get the azimuth.
		F32 Azimuth(void) const
		{ return _azimuth; }

		// Set the azimuth.
		void Azimuth(F32 azimuth)
		{ _azimuth = azimuth; }

		// Compare two coordinates for equality.
		Bool operator == (const Polar2& rhs) const
		{ return ((Vector2(*this) - Vector2(rhs)).LengthSqr() < Epsilon() * Epsilon()); }

		// Compare two coordinates for inequality.
		Bool operator != (const Polar2& rhs) const
		{ return !((*this) == rhs); }

		// Get a new coordinate from another coordinate added to this coordinate.
		Polar2 operator + (const Polar2& rhs)
		{ return Polar2(_azimuth + rhs._azimuth, _radius + rhs._radius); }

		// Get a new coordinate from another coordinate subtracted this coordinate.
		Polar2 operator - (const Polar2& rhs)
		{ return Polar2(_azimuth - rhs._azimuth, _radius - rhs._radius); }

		// Get a new coordinate from this coordinate with a radius scaled by a value.
		Polar2 operator * (const F32 rhs)
		{ return Polar2(_azimuth, _radius * rhs); }

		// Get a new coordinate from this coordinate with a radius scaled by a value.
		Polar2 operator / (const F32 rhs)
		{ return *this * (1.0f / rhs); }

		// Add another coordinate to this coordinate.
		Polar2& operator += (const Polar2& rhs)
		{ _azimuth += rhs._azimuth; _radius += rhs._radius; return *this; }

		// Subtract another coordinate from this coordinate.
		Polar2& operator -= (const Polar2& rhs)
		{ _azimuth -= rhs._azimuth; _radius -= rhs._radius; return *this; }

		// Multiply (scale) this coordinate by a value.
		Polar2& operator *= (const F32 rhs)
		{ _radius *= rhs; return *this; }

		// Divide (scale) this coordinate by a value.
		Polar2& operator /= (const F32 rhs)
		{ return *this *= (1.0f / rhs); }

		// Multiply (scale) this coordinate by a value (pre-multiply).
		friend Polar2 operator * (const F32 lhs, const Polar2& rhs)
		{ return Polar2(lhs * rhs._azimuth, lhs * rhs._radius); }

		// Multiply (scale) this coordinate by a value.
		Polar2& Scale(const F32 scale)
		{ *this *= scale; return *this; }

		// Get a new coordinate from this coordinate scaled by a value.
		Polar2 Scaled(const F32 scale)
		{ return Polar2(*this).Scale(scale); }

		// Get the squared length of this coordinate.
		F32 LengthSqr(void) const
		{ return _radius * _radius; }

		// Get the length of this coordinate.
		F32 Length(void) const
		{ return _radius; }

		// Normalise this coordinate returning its length.
		F32 Normalise(void);

		// Get a new coordinate normalised from this coordinate.
		Polar2 Normalised(void) const
		{ return Polar2(_azimuth, 1.0f); }

	private:

		F32 _radius;
		F32 _azimuth;
	};

	// Now include the inline definitions.

#include "polar2.inl"

// namespace libmaths
}

#pragma managed(pop)
