//********************************************************************************
//
// Quaternions are a space efficient representation for 3 dimensional rotations.
// They can be converted to and from regular rotation matrices for ease of use.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "math.h"
#include "vector3.h"
#include "fixedpoint.h"


namespace libmaths
{
	//********************************************************************************
	// Quaternion class.
	//********************************************************************************

	class Quaternion
	{

	public:

		// Get the epsilon value for equality comparisons.
		static const F32 Epsilon(void)
		{ return 1.0e-5f; }

		// A fixed quaternion of all zeroes.
		static const Quaternion& ZERO(void);

		// A fixed quaternion of identity.
		static const Quaternion& IDENTITY(void);

		// Default constructor.
		Quaternion(void);

		// Copy constructor.
		Quaternion(const Quaternion& rhs) : _x(rhs._x), _y(rhs._y), _z(rhs._z), _w(rhs._w)
		{ }

		// Construct from components.
		Quaternion(F32 x, F32 y, F32 z, F32 w) : _x(x), _y(y), _z(z), _w(w)
		{ }

		// Construct from pointer to components
		explicit Quaternion(const F32* pValues) : _x(pValues[0]), _y(pValues[1]), _z(pValues[2]), _w(pValues[3])
		{ }

		// Construct from a vector and scalar.
		Quaternion(const Vector3x& vector, F32 scalar) : _x(vector.X()), _y(vector.Y()), _z(vector.Z()), _w(scalar)
		{ }

		// Construct from a 3x3 matrix.
		explicit Quaternion(const Matrix33of44& matrix);

		// Construct from a 4x4 matrix.
		// The fourth column and translation are discarded in the process.
		explicit Quaternion(const Matrix44& matrix);

		// Construct from a pointer to 4 16 bit quaternion components in 1.15 fixed point format.
		explicit Quaternion(const Fixed1p15* pValues);

		// Set this quaternion to zero.
		Quaternion& Zero(void)
		{ return (*this = Quaternion::ZERO()); }

		// Set this quaternion to identity.
		Quaternion& Identity(void)
		{ return (*this = Quaternion::IDENTITY()); }

		// Get the x component.
		F32 X(void) const
		{ return _x; }

		// Set the x component.
		void X(const F32 v0)
		{ _x = v0; }

		// Get the y component.
		F32 Y(void) const
		{ return _y; }

		// Set the y component.
		void Y(const F32 v0)
		{ _y = v0; }

		// Get the z component.
		F32 Z(void) const
		{ return _z; }

		// Set the z component.
		void Z(const F32 v0)
		{ _z = v0; }

		// Get the w component.
		F32 W(void) const
		{ return _w; }

		// Set the w component.
		void W(const F32 v0)
		{ _w = v0; }

		// Set the x, y, z and w components,
		void XYZW(const F32 x, const F32 y, const F32 z, const F32 w)
		{ _x = x; _y = y; _z = z; _w = w; }

		// Set this quaternion from a 3x3 matrix.
		Quaternion& FromMatrix(const Matrix33of44& matrix);

		// Convert this quaternion to a 3x3 matrix.
		void ToMatrix(Matrix33of44& matrix) const;

		// Set this quaternion from a 4x4 matrix.
		// The fourth column and translation are discarded in the process.
		Quaternion& FromMatrix(const Matrix44& matrix);

		// Convert this quaternion to a 4x4 matrix (default last row and column).
		void ToMatrix(Matrix44& matrix) const;

		// Build a transformation representing a rotation of angle radians anti-clockwise
		// about the axis defined by axis.
		Quaternion& FromAxisAngle(const Vector3x& axis, const F32 angle);

		// Decompose the quaternion into a rotation of radians anti-clockwise about an axis.
		void ToAxisAngle(Vector3x& axis, F32& angle) const;

		// Build a transformation from a direction vector with a specified "up" vector.
		// NOTE: This assumes right-handed orientation.
		Quaternion& FromDirectionRH(const Vector3x& direction, const Vector3x& up = Vector3x::ZAXIS());

		// Build a transformation from a direction vector with a specified "up" vector.
		// NOTE: This assumes left-handed orientation.
		Quaternion& FromDirectionLH(const Vector3x& direction, const Vector3x& up = Vector3x::YAXIS());

		// Build a transformation that will rotate one vector to another when multiplied with this quaternion.
		Quaternion& Reorient(const Vector3x& from, const Vector3x& to);

		// Invert this quaternion, producing an inverse transformation.
		Quaternion& Invert(void);

		// Get a new quaternion from the inverse of this quaternion, producing an inverse transformation.
		Quaternion Inverse(void) const;

		// Invert this quaternion, producing an inverse transformation.
		// Assumes this is a unit rotation.
		Quaternion UnitInvert(void);

		// Get a new quaternion from the inverse of this quaternion, producing an inverse
		// transformation. Assumes this is a unit rotation.
		Quaternion UnitInverse(void) const;

		// Get the dot product of this quaternion with another quaternion.
		F32 Dot(const Quaternion& rotation) const
		{ return _x * rotation._x + _y * rotation._y + _z * rotation._z + _w * rotation._w; }

		// Get the squared norm of this quaternion.
		F32 NormSqr(void) const
		{ return _x * _x + _y * _y + _z * _z + _w * _w; }

		// Get the norm of this quaternion.
		F32 Norm(void) const;

		// Normalise this quaternion returning its normal.
		F32 Normalise(void);

		// Get a new quaternion normalised from this quaternion.
		Quaternion Normalised(void) const;

		// Get the squared length of the vector part of this quaternion.
		F32 VectorLengthSqr(void) const
		{ return _x * _x + _y * _y + _z * _z; }

		// Get the length of the vector part of this quaternion.
		F32 VectorLength(void) const
		{ return Sqrt(VectorLengthSqr()); }

		// Set this quaternion by spherical linear interpolation between q0 and q1 by an
		// amount t with extra spins. This function takes no account of the possibility
		// that they may lie in opposite hemispheres.
		void Slerp(const F32 t, const Quaternion& q0, const Quaternion& q1, U32 extraspins = 0);

		// Set this quaternion by spherical linear interpolation between q0 and q1 by an
		// amount t with extra spins. This function takes into account the possibility
		// that they lie in opposite hemispheres and will always choose the shortest
		// interpolation path between them.
		void SlerpShortest(const F32 t, const Quaternion& q0, const Quaternion& q1, U32 extraSpins = 0);

		// Spherical quadratic interpolation
		void Squad(const F32 t, const Quaternion& q0, const Quaternion& q1, const Quaternion& q2, const Quaternion& q3);

		// Get the exponential of this quaternion.
		Quaternion Exp(void) const;

		// Get the natural Logarithm of this quaternion.
		Quaternion Log(void) const;

		// Compare two quaternions for equality.
		Bool operator == (const Quaternion& rhs) const
		{ return (Abs(rhs._x - _x) < Epsilon() && Abs(rhs._y - _y) < Epsilon() && Abs(rhs._z - _z) < Epsilon() && Abs(rhs._w - _w) < Epsilon()); }

		// Compare two quaternions for inequality.
		Bool operator != (const Quaternion& rhs) const
		{ return !((*this) == rhs); }

		// Get a new quaternion that is the negative of this quaternion.
		Quaternion operator - (void) const
		{ return Quaternion(-_x, -_y, -_z, -_w); }

		// Multiply (scale) this quaternion by a value.
		Quaternion operator * (F32 scale)
		{ return Quaternion(_x * scale, _y * scale, _z * scale, _w * scale); }

		// Multiply (scale) this quaternion by a value.
		Quaternion operator / (F32 scale)
		{ return *this * (1.0f / scale); }

		// Add another quaternion to this quaternion.
		Quaternion operator + (const Quaternion& rhs)
		{ return Quaternion(_x + rhs._x, _y + rhs._y, _z + rhs._z, _w + rhs._w); }

		// Subtract another quaternion from this quaternion.
		Quaternion operator - (const Quaternion& rhs)
		{ return Quaternion(_x - rhs._x, _y - rhs._y, _z - rhs._z, _w - rhs._w); }

		// Get a new quaternion from this quaternion multiplied by (concatenated with) another quaternion.
		Quaternion operator * (const Quaternion& rhs);

		// Multiply a 3D vector by this quaternion.
		Vector3x operator * (const Vector3x& vector) const;

		// Multiply a 3D vector by this quaternion.
		Vector3 operator * (const Vector3& vector) const;

		// Add another quaternion to this quaternion.
		Quaternion& operator += (const Quaternion& rhs);

		// Subtract another quaternion from this quaternion.
		Quaternion& operator -= (const Quaternion& rhs);

		// Multiply this quaternion by (concatenated with) another quaternion.
		Quaternion& operator *= (const Quaternion& rhs);

		// Multiply (scale) this quaternion by a value.
		Quaternion& operator *= (const F32 scale);

		// Multiply (scale) this quaternion by a value.
		Quaternion& operator /= (const F32 scale)
		{ return *this *= (1.0f / scale); }

		// Get a new quaternion from a quaternion scaled by a value.
		friend Quaternion operator * (F32 lhs, const Quaternion& rhs)
		{ return Quaternion(rhs._x * lhs, rhs._y * lhs, rhs._z * lhs, rhs._w * lhs); }

	private:

		// Access quaternion components as an array (read only).
		const F32 operator [] (const S32 i) const
		{ return reinterpret_cast<const F32*>(&_x)[i]; }

		// Access quaternion components as an array (read-write).
		F32& operator [] (const S32 i)
		{ return reinterpret_cast<F32*>(&_x)[i]; }

		// Quaternion vector x component
		F32 _x;

		// Quaternion vector y component
		F32 _y;

		// Quaternion vector z component
		F32 _z;

		// Quaternion scalar component
		F32 _w;
	};


	// Now include the inline definitions of the structure members

#include "quaternion.inl"

// namespace libmaths
}

#pragma managed(pop)
