//********************************************************************************
//
// Quaternions are a space efficient representation for 3 dimensional rotations.
// They can be converted to and from regular rotation matrices for ease of use.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once



//********************************************************************************
// Defines Affecting this file:
//
// DEFAULT_INITIALISERS:
// To set default values to components when default constructor is used
//********************************************************************************


//********************************************************************************
// A fixed quaternion of all zeroes.
//********************************************************************************

LIBMATHS_INLINE const Quaternion& Quaternion::ZERO(void)
{
	static Quaternion quaternion(Vector3x(0.0f, 0.0f, 0.0f), 0.0f);

	return quaternion;
}


//********************************************************************************
// A fixed quaternion of identity.
//********************************************************************************

LIBMATHS_INLINE const Quaternion& Quaternion::IDENTITY(void)
{
	static Quaternion quaternion(Vector3x(0.0f, 0.0f, 0.0f), 1.0f);

	return quaternion;
}


//********************************************************************************
//Default constructor.
//********************************************************************************

LIBMATHS_INLINE Quaternion::Quaternion(void)
#ifdef DEFAULT_INITIALISERS
: _x(0.0f), _y(0.0f), _z(0.0f), _w(0.0f)
#endif // DEFAULT_INITIALISERS
{
}


//********************************************************************************
// Construct from 64 bit pointer to 4 16 bit quaternion components in 1.15 fixed
// point format.
//********************************************************************************

LIBMATHS_INLINE Quaternion::Quaternion(const Fixed1p15* pValues)
{
	_x = F32(*pValues++);
	_y = F32(*pValues++);
	_z = F32(*pValues++);
	_w = F32(*pValues++);
}


//********************************************************************************
// Get the norm of this quaternion.
//********************************************************************************

LIBMATHS_INLINE F32 Quaternion::Norm(void) const
{
	F32 normSqr = NormSqr();

	LIBMATHS_ASSERT(normSqr != 0.0f);

	return libmaths::Sqrt(normSqr);
}


//********************************************************************************
// Get a new quaternion from this quaternion multiplied by (concatenated with)
// another quaternion.
//********************************************************************************

LIBMATHS_INLINE Quaternion Quaternion::operator * (const Quaternion& rhs)
{
	return Quaternion
		(
		_w * rhs._x + _x * rhs._w + _y * rhs._z - _z * rhs._y,
		_w * rhs._y + _y * rhs._w + _z * rhs._x - _x * rhs._z,
		_w * rhs._z + _z * rhs._w + _x * rhs._y - _y * rhs._x,
		_w * rhs._w - _x * rhs._x - _y * rhs._y - _z * rhs._z
		);
}


//********************************************************************************
// Add another quaternion to this quaternion.
//********************************************************************************

LIBMATHS_INLINE Quaternion& Quaternion::operator += (const Quaternion& rhs)
{
	_x += rhs._x;
	_y += rhs._y;
	_z += rhs._z;
	_w += rhs._w;

	return (*this);
}


//********************************************************************************
// Subtract another quaternion from this quaternion.
//********************************************************************************

LIBMATHS_INLINE Quaternion& Quaternion::operator -= (const Quaternion& rhs)
{
	_x -= rhs._x;
	_y -= rhs._y;
	_z -= rhs._z;
	_w -= rhs._w;

	return (*this);
}


//********************************************************************************
// Multiply this quaternion by (concatenated with) another quaternion.
//********************************************************************************

LIBMATHS_INLINE Quaternion& Quaternion::operator *= (const Quaternion& rhs)
{
	*this = Quaternion
		(
		_w * rhs._x + _x * rhs._w + _y * rhs._z - _z * rhs._y,
		_w * rhs._y + _y * rhs._w + _z * rhs._x - _x * rhs._z,
		_w * rhs._z + _z * rhs._w + _x * rhs._y - _y * rhs._x,
		_w * rhs._w - _x * rhs._x - _y * rhs._y - _z * rhs._z
		);

	return *this;
}


//********************************************************************************
// Multiply (scale) this quaternion by a value.
//********************************************************************************

LIBMATHS_INLINE Quaternion& Quaternion::operator *= (const F32 scale)
{
	_x *= scale;
	_y *= scale;
	_z *= scale;
	_w *= scale;

	return *this;
}
