//********************************************************************************
//
// Fast random number generator in the form of the RandomFast, and a better but
// slower generator in the form of RandomSlow.
//
// From numerical recipes. Example usage:
//
// RandomFast MyRand;
// F32 Number;
//
// 0.0f to 1.0f (exclusive)
// Number = MyRand;
// 0.0f to 10.0f (exclusive)
// Number = MyRand * 10.0f;
// -10.0f to 10.0f (exclusive)
// Number = MyRand ^ 10.0f;
//
// S32 Number;
//
// 0 to MAX_RANDOM (inclusive)
// Number = MyRand;
// 0 to 10 (exclusive)
// Number = MyRand * 10;
// -10 to 10 (exclusive)
// Number = MyRand ^ 10;
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"


namespace libmaths
{

	//********************************************************************************
	// Very fast random number generator class.
	//********************************************************************************

	class RandomFast
	{

	public:

		static const S32 MAX_RANDOM = 0x7fff;

		RandomFast(void) : _root(0)
		{ }
		
		RandomFast(U32 Seed) : _root(Seed)
		{ }

		// Conversion to F32 type, typically a left hand assignment
		// returns a random F32 between 0.0f and 1.0f (1.0f exclusive)
		operator F32(void);

		// Conversion to S32 type, typically a left hand assignment
		// returns a random S32 between 0 and MAX_RANDOM (MAX_RANDOM inclusive)
		operator S32(void)
		{ return FToI(F32(*this) * (MAX_RANDOM + 1)); }

		// RHS Assignment operator, allows the seed to be set
		void operator = (U32 seed)
		{ _root = seed; }

		// return a F32 scaled by an absolute scalar (rhs exclusive)
		F32 operator * (F32 rhs)
		{ return F32(*this) * rhs; }

		// return a S32 scaled by an absolute scalar (rhs exclusive)
		S32 operator * (S32 rhs)
		{ return FToI(F32(*this) * rhs); }

		// return a U32 scaled by an absolute scalar (rhs exclusive)
		U32 operator * (U32 rhs)
		{ return (*this) * S32(rhs); }

		// return a F32 scaled by a signed range (rhs exclusive)
		F32 operator ^ (F32 rhs)
		{ return (F32(*this) * rhs * 2.0f) - rhs; }

		// return a S32 scaled by a signed range (rhs exclusive)
		S32 operator ^ (S32 rhs)
		{ return FToI((F32(*this) * rhs * 2.0f) - rhs); }

		// return a S32 scaled by a signed range (rhs exclusive)
		S32 operator ^ (U32 rhs)
		{ return (*this) ^ S32(rhs); }

	private:

		// The current root for generating random numbers
		U32 _root;
	};


	//********************************************************************************
	// Slower random number generator class, using the Mersenne Twister algorithm.
	//********************************************************************************

	class RandomSlow
	{

	public:

		static const S32 MAX_RANDOM = 0x7fff;

		RandomSlow(void) : mti(N + 1)
		{ *this = 5489; }

		RandomSlow(U32 seed) : mti(N + 1)
		{ *this = seed; }

		// Conversion to F32 type, typically a left hand assignment
		// returns a random F32 between 0.0f and 1.0f (1.0f exclusive)
		operator F32(void)
		{ return Random() * (1.0f / 4294967750.0f); }

		// Conversion to S32 type, typically a left hand assignment
		// returns a random S32 between 0 and MAX_RANDOM (MAX_RANDOM inclusive)
		operator S32(void)
		{ return Random() & MAX_RANDOM; }

		// RHS Assignment operator, allows the seed to be set
		void operator = (U32 seed);

		// return a F32 scaled by an absolute scalar (rhs exclusive)
		F32 operator * (F32 rhs)
		{ return F32(*this) * rhs; }

		// return a S32 scaled by an absolute scalar (rhs exclusive)
		S32 operator * (S32 rhs)
		{ return (Random() % ((rhs < 0) ? -rhs : rhs)); }

		// return a U32 scaled by an absolute scalar (rhs exclusive)
		U32 operator * (U32 rhs)
		{ return Random() % rhs; }

		// return a F32 scaled by a signed range (rhs exclusive)
		F32 operator ^ (F32 rhs)
		{ return (F32(*this) * rhs * 2.0f) - rhs; }

		// return a S32 scaled by a signed range (rhs exclusive)
		S32 operator ^ (S32 rhs)
		{ return (*this ^ U32((rhs < 0) ? -rhs : rhs)); }

		// return a S32 scaled by a signed range (rhs exclusive)
		S32 operator ^ (U32 rhs)
		{ return ((*this) * rhs * 2) - rhs; }

	private:

		U32 Random(void);

		enum GeGenerationConstants
		{
			N = 624,
			M = 397,

			// constant vector a
			MATRIX_A = 0x9908b0df,

			// most significant w-r bits
			UPPER_MASK = 0x80000000,

			// least significant r bits
			LOWER_MASK = 0x7fffffff,

			// Tempering parameters   
			TEMPERING_MASK_B = 0x9d2c5680,
			TEMPERING_MASK_C = 0xefc60000,
			TEMPERING_SHIFT_U = 11,
			TEMPERING_SHIFT_S = 7,
			TEMPERING_SHIFT_T = 15,
			TEMPERING_SHIFT_L = 18
		};

		// The array for the state vector
		U32 mt[N];
		S32 mti;
	};


	//********************************************************************************
	// Inline function definitions, no more interface definitions beyond this point.
	//********************************************************************************

	LIBMATHS_INLINE RandomFast::operator F32(void)
	{
		_root = 1664525L * _root + 1013904223L;

		U32 itemp = 0x3f800000 | (0x007fffff & _root);

		return (*(F32 *)&itemp) - 1.0f;
	}

	// RHS Assignment operator, allows the seed to be set
	LIBMATHS_INLINE void RandomSlow::operator = (U32 seed)
	{
		mt[0] = seed & 0xffffffff;

		for (mti = 1; mti < N; mti++)
		{
			mt[mti] = (69069 * mt[mti - 1]) & 0xffffffff;
		}
	}

	LIBMATHS_INLINE U32 RandomSlow::Random(void)
	{
		U32 result;
		static U32 mag01[2] = { 0x0, (U32)MATRIX_A };

		// Generate N words at one time.

		if (mti >= N)
		{
			S32 kk;

			if (mti == N + 1)
			{
				*this = 5489;
			}

			for (kk = 0; kk < N - M; kk++)
			{
				result = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
				mt[kk] = mt[kk + M] ^ (result >> 1) ^ mag01[result & 0x1];
			}

			for (; kk < N - 1; kk++)
			{
				result = (mt[kk] & UPPER_MASK) | (mt[kk + 1] & LOWER_MASK);
				mt[kk] = mt[kk + (M - N)] ^ (result >> 1) ^ mag01[result & 0x1];
			}

			result = (mt[N - 1] & UPPER_MASK) | (mt[0] & LOWER_MASK);
			mt[N - 1] = mt[M - 1] ^ (result >> 1) ^ mag01[result & 0x1];

			mti = 0;
		}

		result = mt[mti++];

		result ^= (result >> TEMPERING_SHIFT_U);
		result ^= (result << TEMPERING_SHIFT_S) & TEMPERING_MASK_B;
		result ^= (result << TEMPERING_SHIFT_T) & TEMPERING_MASK_C;
		result ^= (result >> TEMPERING_SHIFT_L);

		return result;
	}

// namespace libmaths
}

#pragma managed(pop)
