//********************************************************************************
//
// A class used to define rays in 2D space, often used for detecting intersections
// with other geometric types such as circles and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "linesegment2.h"


namespace libmaths
{

	//********************************************************************************
	// Class for a 2-dimensional ray. It has a start position and a unit length
	// direction, effectively giving it infinite length. If you need a ray of finite
	// length then use a LineSegment2 class instead.
	//********************************************************************************

	class Ray2
	{

	public:

		// Construct a ray
		Ray2(void) :
			_origin(Vector2::ZERO()), _direction(Vector2::ZERO())
		{ }

		// Construct a ray with a start point and a direction (should always be normalised)
		Ray2(const Vector2& start, const Vector2& direction) :
			_origin(start), _direction(direction)
		{ LIBMATHS_ASSERT(Abs(direction.Length() - 1.0f) < Epsilon); }

		// Construct a ray
		Ray2(const LineSegment2& line) :
			_origin(line.Origin()), _direction(line.Direction().Normalised())
		{ }

		// Access the start position of the ray
		const Vector2& Origin(void) const
		{ return _origin; }

		// Set the start position of the ray
		void Origin(const Vector2& origin)
		{ _origin = origin; }

		// Access the direction of the ray
		const Vector2& Direction(void) const
		{ return _direction; }

		// Set the direction of the ray
		void Direction(const Vector2& direction)
		{ _direction = direction; }

		// Does the line intersect a rectangle?
		Bool Intersects(AxisAlignedRectangle& rectangle) const;

		// Does the line intersect a rectangle?
		Bool Intersects(AxisAlignedRectangle& rectangle, Vector2& point, Vector2& normal, F32& t) const;

		// Does the line intersect a sphere?
		Bool Intersects(Circle& circle) const;

		// Does the line intersect a sphere?
		Bool Intersects(Circle& circle, Vector2& point, Vector2& normal, F32& t) const;

		// Does the line intersect a triangle?
		Bool Intersects(Triangle2& triangle) const;

		// Does the line intersect a triangle?
		Bool Intersects(Triangle2& triangle, Vector2& point, Vector2& normal, F32& t) const;

	private:

		// The start position of the ray
		Vector2 _origin;

		// The direction of the ray
		Vector2 _direction;
	};

// namespace libmaths
}

#pragma managed(pop)
