//********************************************************************************
//
// A class used to define rays in 3D space, often used for detecting intersections
// with other geometric types such as spheres and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "linesegment3.h"


namespace libmaths
{

	class AxisAlignedBox;
	class Capsule;
	class Sphere;
	class Triangle3;
	class Plane;


	//********************************************************************************
	// Class for a 3-dimensional ray. It has a start position and a unit length
	// direction, effectively giving it infinite length. If you need a ray of finite
	// length then use a LineSegment3 class instead.
	//********************************************************************************

	class Ray3
	{

	public:

		// Construct a ray
		Ray3(void) :
			_origin(Vector3x::ZERO()), _direction(Vector3x::ZERO())
		{ }

		// Construct a ray with a start point and a direction (should always be normalised)
		Ray3(const Vector3x& start, const Vector3x& direction) :
				_origin(start), _direction(direction)
		{ LIBMATHS_ASSERT(Abs(direction.Length() - 1.0f) < Epsilon); }

		// Construct a ray
		Ray3(const LineSegment3& line) :
			_origin(line.Origin()), _direction(line.Direction().Normalised())
		{ }

		// Access the start position of the ray
		const Vector3x& Origin(void) const
		{ return _origin; }

		// Set the start position of the ray
		void Origin(const Vector3x& origin)
		{ _origin = origin; }

		// Access the direction of the ray
		const Vector3x& Direction(void) const
		{ return _direction; }

		// Set the direction of the ray
		void Direction(const Vector3x& direction)
		{ _direction = direction; }

		// Get the distance of the ray from a point in space
		F32 Distance(const Vector3x& point, F32& pointOnRay, Bool clamp = true) const;

		// Get the squared distance of the ray from a point in space
		F32 DistanceSqr(const Vector3x& point, F32& pointOnRay, Bool clamp = true) const;

		// Does the ray intersect a box?
		Bool Intersects(const AxisAlignedBox& box) const;

		// Does the ray intersect a box?
		Bool Intersects(const AxisAlignedBox& box, Vector3x& point, Vector3x& normal) const;

		// Does the ray intersect a capsule?
		Bool Intersects(const Capsule& capsule) const;

		// Does the ray intersect a capsule?
		Bool Intersects(const Capsule& capsule, Vector3x& point, Vector3x& normal) const;

		// Does the ray intersect a sphere?
		Bool Intersects(const Sphere& sphere) const;

		// Does the ray intersect a sphere?
		Bool Intersects(const Sphere& sphere, Vector3x& point, Vector3x& normal) const;

		// Does the ray intersect a triangle?
		Bool Intersects(const Triangle3& triangle, Bool twoSided = false) const;

		// Does the ray intersect a triangle?
		Bool Intersects(const Triangle3& triangle, Vector3x& point, Vector3x& normal, Bool twoSided = false) const;

		// Does the ray intersect a plane?
		Bool Intersects(const Plane& plane, Bool twoSided = false) const;

		// Does the ray intersect a plane?
		Bool Intersects(const Plane& plane, Vector3x& point, Vector3x& normal, Bool twoSided = false) const;

	private:

		// The start position of the ray
		Vector3x _origin;

		// The direction of the ray
		Vector3x _direction;
	};

// namespace libmaths
}

#pragma managed(pop)
