//********************************************************************************
//
// 2D bounding boxes for collision checks.
//
// These can be tested for collisions with other rectangles, points, circles or
// rays through the game world.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"
#include "circle.h"
#include "triangle2.h"


namespace libmaths
{

	//********************************************************************************
	// 2 dimensional bounding rectangle (Vector2 components).
	//
	// 2 dimensional bounding rectangle formed from a centre point and an extent
	// representing the half width of the box in each axis.
	//
	// These can be tested for collisions with other boxes, points, circles or rays
	// through the game world.
	//
	// The box centre and extent are stored in Vector2 format.
	//********************************************************************************

	class AxisAlignedRectangle
	{

	public:

		// Construct a rectangle
		AxisAlignedRectangle(void) :
			_centre(0, 0), _extent(-Epsilon, -Epsilon)
		{ }

		// Construct a rectangle with the given centre and extent.
		AxisAlignedRectangle(const Vector2& centre, const Vector2& extent) :
			_centre(centre), _extent(extent)
		{ }

		// Construct a rectangle around a circle.
		AxisAlignedRectangle(const Circle& circle)
		{
			F32 e = circle.Radius();

			_centre = circle.Point();
			_extent = Vector2(e, e);
		}

		// Construct a rectangle with the given centre and uniform extent r in each axis.
		AxisAlignedRectangle(const Vector2& centre, F32 extent) :
			_centre(centre), _extent(extent, extent)
		{ }

		// Destroy a rectangle.
		~AxisAlignedRectangle(void)
		{ }

		// Set the rectangle centre.
		void Centre(const Vector2& centre)
		{ _centre = centre; }

		// Set the rectangle extent.
		void Extent(const Vector2& extent)
		{ _extent = extent; }

		// Return the rectangle centre.
		const Vector2& Centre(void) const
		{ return _centre; }

		// Return the rectangle extent.
		const Vector2& Extent(void) const
		{ return _extent; }

		// Sets the minimum and maximum corner points of the rectangle.
		void MinMax(const Vector2& min, const Vector2& max)
		{
			_extent = (max - min) * 0.5f;
			_centre = min + _extent;
		}

		// Return the minimum corner point of the rectangle.
		Vector2 Minimum(void) const
		{ return _centre - _extent; }

		// Return the maximum corner point of the rectangle.
		Vector2 Maximum(void) const
		{ return _centre + _extent; }

		// Return the component of the minimum corner point of the rectangle in the given axis.
		F32 Minimum(const S32 axis) const
		{ return _centre[axis] - _extent[axis]; }

		// Return the component of the maximum corner point of the rectangle in the given axis.
		F32 Maximum(const S32 axis) const
		{ return _centre[axis] + _extent[axis]; }

		// Set the rectangle to empty. i.e. enclosing no volume.
		void Empty(void)
		{ _centre = Vector2(0.0f, 0.0f); _extent = Vector2(-Epsilon, -Epsilon); }

		// Build a rectangle surrounding the two given bounding boxes.
		void Surround(const AxisAlignedRectangle& rectangle0, const AxisAlignedRectangle& rectangle1)
		{
			Vector2 min, max;

			min.MinOf(rectangle0.Minimum(), rectangle1.Minimum());
			max.MaxOf(rectangle0.Maximum(), rectangle1.Maximum());

			MinMax(min, max);
		}

		// Extend the rectangle to additionally enclose the given point.
		void Add(const Vector2& point)
		{
			Vector2 min, max;

			min.MinOf(Minimum(), point);
			max.MaxOf(Maximum(), point);

			MinMax(min, max);
		}

		// Extend the rectangle to additionally enclose the given rectangle.
		void Add(const AxisAlignedRectangle& rectangle)
		{ Surround(*this, rectangle); }

		// Extend the rectangle to additionally enclose the given triangle.
		void Add(const Triangle2& triangle)
		{
			Add(triangle[0]);
			Add(triangle[1]);
			Add(triangle[2]);
		}

		// Extend the rectangle to additionally enclose the given circle.
		void Add(const Circle& circle)
		{ Surround(*this, AxisAlignedRectangle(circle)); }

		// Return the maximum extent of the rectangle in all axes.
		F32 Size(void) const
		{ return Max(_extent[0], _extent[1]); }

		// Return the axis of the rectangle with the maximum extent.
		S32 MaxAxis(void) const
		{ return (Abs(_extent[0]) < Abs(_extent[1])) ? 1 : 0; }

		// Check if the rectangle intersects the given rectangle.
		Bool Intersects(const AxisAlignedRectangle& rectangle) const
		{
			const Vector2 centre(Abs(_centre - rectangle._centre));
			const Vector2 extent(_extent + rectangle._extent);

			if (centre.X() > extent.X() ||
				centre.Y() > extent.Y())
			{
				return false;
			}

			return true;
		}

		// Check if the rectangle intersects (contains) the given point.
		Bool Intersects(const Vector2& point) const
		{
			Vector2 centre(Abs(_centre - point));

			if (centre.X() > _extent.X() ||
				centre.Y() > _extent.Y())
			{
				return false;
			}

			return true;
		}

		// Check if the rectangle intersects the given circle.
		Bool Intersects(const Circle& circle) const
		{
			Vector2 centre(Abs(_centre - circle.Point()));
			Vector2 extent(_extent + circle.Extent());

			if (centre.X() > extent.X() ||
				centre.Y() > extent.Y())
			{
				return false;
			}

			return true;
		}

		// Check if the rectangle contains a point in space.
		Bool Contains(const Vector2& point) const
		{
			Vector2 topLeft(Minimum());
			Vector2 bottomRight(Maximum());

			return (point.X() >= topLeft.X() && point.X() <= bottomRight.X() &&
				point.Y() >= topLeft.Y() && point.Y() <= bottomRight.Y());
		}

		// Check if the line intersects this rectangle.
		// Does not allow intersections from within the rectangle.
		Bool Intersects(const Ray2& ray) const
		{
			Vector2 point, normal;

			return Intersects(ray, point, normal);
		}

		// Check if the line intersects this rectangle.
		// Does not allow intersections from within the rectangle.
		Bool Intersects(const Ray2& ray, Vector2& point, Vector2& normal) const;

		// Check if the line intersects this rectangle.
		// Does not allow intersections from within the rectangle.
		Bool Intersects(const LineSegment2& line) const;

		// Check if the line intersects this rectangle.
		// Does not allow intersections from within the rectangle.
		Bool Intersects(const LineSegment2& line, Vector2& point, Vector2& normal, F32& t) const
		{
			if (Intersects(Ray2(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the area of the rectangle.
		F32 Area(void)
		{ return (_extent.x * 2) * (_extent.y * 2); }

	private:

		// Centre of the rectangle.
		Vector2 _centre;

		// Extent of the rectangle in each axis.
		Vector2 _extent;
	};

// namespace libmaths
}

#pragma managed(pop)
