//********************************************************************************
//
// 3D bounding spheres for collision checks.
//
// Bounding spheres can detect overlaps with other spheres, points and bounding
// boxes.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"
#include "ray3.h"
#include "linesegment3.h"
#include "matrix44.h"


namespace libmaths
{

	class AxisAlignedBox;

	//********************************************************************************
	// A bounding sphere.
	//
	// Supports intersection checks, bounding of point sets, boxes and other spheres.
	//********************************************************************************

	class Sphere
	{

	public:

		// Construct an empty bounding sphere (enclosing no volume)
		Sphere(void) :
			_sphere(0, 0, 0)
		{ _sphere.W(-Epsilon); }

		// Construct bounding sphere with the given centre and radius.
		Sphere(const F32 x, const F32 y, const F32 z, const F32 radius) :
			_sphere(x, y, z)
		{ _sphere.W(radius); }

		// Construct bounding sphere with the given centre and radius.
		Sphere(const Vector3x& point, const F32 radius) :
			_sphere(point)
		{ _sphere.W(radius); }

		// Construct bounding sphere surrounding the given bounding box.
		Sphere(const AxisAlignedBox& b);

		// Destroy bounding sphere.
		~Sphere(void)
		{ }

		// Empty the bounding sphere such that it contains no volume.
		void Empty(void)
		{ _sphere.XYZ(0, 0, 0); _sphere.W(-Epsilon); }

		// Is the bounding sphere empty, such that it contains no volume?
		Bool IsEmpty(void) const
		{ return (_sphere.W() <= 0.0f); }

		// Build bounding sphere enclosing the two given bounding spheres.
		void Surround(const Sphere& sphere0, const Sphere& sphere1);

		// Build bounding sphere enclosing the given bounding sphere and box.
		void Surround(const Sphere& sphere, const AxisAlignedBox& box);

		// Extend the bounding sphere to additionally enclose the given bounding sphere.
		void Add(const Sphere& sphere);

		// Extend the bounding sphere to additionally enclose the given bounding box.
		void Add(const AxisAlignedBox& box);

		// Return the radius of the bounding sphere.
		const F32 Radius(void) const
		{ return _sphere.W(); }

		// Set the radius of the bounding sphere.
		void Radius(F32 radius)
		{ _sphere.W(radius); }

		// Return the radius squared of the bounding sphere.
		const F32 RadiusSqr(void) const
		{ return _sphere.W() * _sphere.W(); }

		// Return the centre of the bounding sphere.
		const Vector3x& Point(void) const
		{ return _sphere; }

		// Set the centre of the bounding sphere.
		void Point(const Vector3x& point)
		{ _sphere.XYZ(point.X(), point.Y(), point.Z()); }

		// Move a bounding sphere in space, adding an offset to its current position.
		void Translate(const Vector3x& offset)
		{ _sphere += offset; }

		// Return a vector representing the uniform extent of the bounding sphere in all axes.
		const Vector3x Extent(void) const
		{ return Vector3x(Radius(), Radius(), Radius()); }

		// Get the extent of the bounding sphere.
		void Extent(Vector3x& min, Vector3x& max) const
		{
			Vector3x r(Extent());

			min = Point() - r;
			max = Point() + r;
		}

		// Check if the sphere contains a point in space.
		Bool Contains(const Vector3x& point) const
		{ return ((point - Point()).LengthSqr() <= RadiusSqr()); }

		// Check if the bounding sphere intersects the given bounding sphere.
		Bool Intersects(const Sphere& sphere) const
		{
			Vector3x diff = _sphere - sphere._sphere;
			F32 radii = Radius() + sphere.Radius();

			return (diff.LengthSqr() < radii * radii);
		}

		// Check if the ray intersects this sphere.
		// Does not allow intersections from within the sphere.
		Bool Intersects(const Ray3& ray) const;

		// Check if the ray intersects this sphere.
		// Does not allow intersections from within the sphere.
		Bool Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const;

		// Check if the line intersects this sphere.
		// Does not allow intersections from within the sphere.
		Bool Intersects(const LineSegment3& line) const
		{
			Vector3x point, normal;

			if (Intersects(Ray3(line), point, normal))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this sphere.
		// Does not allow intersections from within the sphere.
		Bool Intersects(const LineSegment3& line, Vector3x& point, Vector3x& normal, F32& t) const
		{
			if (Intersects(Ray3(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the volume of the sphere.
		F32 Volume(void)
		{ return PI * Radius() * Radius() * Radius() * (4.0f / 3.0f); }

		// Get a new sphere from this sphere multiplied (transformed) by a 4x4 matrix.
		Sphere operator * (const Matrix44& rhs) const
		{
			Sphere sphere;
			
			sphere.Point(Point() * rhs);

			Vector3x maxLength(rhs.XAxis().LengthSqr(), rhs.YAxis().LengthSqr(), rhs.ZAxis().LengthSqr());

			sphere.Radius(Radius() * Sqrt(Max3(maxLength.x, maxLength.y, maxLength.z)));

			return sphere;
		}

	private:

		// The centre / radius of the bounding sphere.
		Vector3x _sphere;
	};

// namespace libmaths
}

#pragma managed(pop)
