//********************************************************************************
//
// Spherical coordinates representing a 3D vector.
//
// http://en.wikipedia.org/wiki/Spherical_coordinates
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"


namespace libmaths
{

	//********************************************************************************
	// Spherical coordinates representing a 3D vector.
	// Consists of a radius from the origin, a inclination angle from the +Z axis and
	// an azimuth angle from the +X axis.
	//********************************************************************************

	class Spherical3
	{

	public:

		// Get the epsilon value for equality comparisons.
		static const F32 Epsilon(void)
		{ return 1.0e-5f; }

		// Get the radius.
		F32 Radius(void) const
		{ return _radius; }

		// Set the radius.
		void Radius(F32 radius)
		{ _radius = radius; }

		// Get the azimuth.
		F32 Azimuth(void) const
		{ return _azimuth; }

		// Set the azimuth.
		void Azimuth(F32 azimuth)
		{ _azimuth = azimuth; }

		// Get the inclination.
		F32 Inclination(void) const
		{ return _inclination; }

		// Set the inclination.
		void Inclination(F32 inclination)
		{ _inclination = inclination; }

		// Get a new coordinate from another coordinate added to this coordinate.
		Spherical3 operator + (const Spherical3& rhs)
		{ return Spherical3(_azimuth + rhs._azimuth, _inclination + rhs._inclination, _radius + rhs._radius); }

		// Get a new coordinate from another coordinate subtracted this coordinate.
		Spherical3 operator - (const Spherical3& rhs)
		{ return Spherical3(_azimuth - rhs._azimuth, _inclination - rhs._inclination, _radius - rhs._radius); }

		// Get a new coordinate from this coordinate with a radius scaled by a value.
		Spherical3 operator * (const F32 rhs)
		{ return Spherical3(_inclination, _azimuth, _radius * rhs); }

		// Get a new coordinate from this coordinate with a radius scaled by a value.
		Spherical3 operator / (const F32 rhs)
		{ return *this * (1.0f / rhs); }

		// Add another coordinate to this coordinate.
		Spherical3& operator += (const Spherical3& rhs)
		{ _azimuth += rhs._azimuth; _inclination += rhs._inclination; _radius += rhs._radius; return *this; }

		// Subtract another coordinate from this coordinate.
		Spherical3& operator -= (const Spherical3& rhs)
		{ _azimuth -= rhs._azimuth; _inclination -= rhs._inclination; _radius -= rhs._radius; return *this; }

		// Multiply (scale) this coordinate by a value.
		Spherical3& operator *= (const F32 rhs)
		{ _radius *= rhs; return *this; }

		// Divide (scale) this coordinate by a value.
		Spherical3& operator /= (const F32 rhs)
		{ return *this *= (1.0f / rhs); }

		// Multiply (scale) this coordinate by a value (pre-multiply).
		friend Spherical3 operator * (const F32 lhs, const Spherical3& rhs)
		{ return Spherical3(lhs * rhs._azimuth, lhs * rhs._inclination, lhs * rhs._radius); }

		// Multiply (scale) this coordinate by a value.
		Spherical3& Scale(const F32 scale)
		{ *this *= scale; return *this; }

		// Get a new coordinate from this coordinate scaled by a value.
		Spherical3 Scaled(const F32 scale)
		{ return Spherical3(*this).Scale(scale); }

		// Get the squared length of this coordinate.
		F32 LengthSqr(void) const
		{ return _radius * _radius; }

		// Get the length of this coordinate.
		F32 Length(void) const
		{ return _radius; }

		// Normalise this coordinate returning its length.
		F32 Normalise(void);

		// Get a new coordinate normalised from this coordinate.
		Spherical3 Normalised(void) const
		{ return Spherical3(_azimuth, _inclination, 1.0f); }

	protected:

		// Default constructor.
		Spherical3(void) : _radius(0.0f), _inclination(0.0f), _azimuth(0.0f)
		{ }

		// Construct from azimuth and inclination.
		Spherical3(F32 azimuth, F32 inclination) : _radius(1.0f), _azimuth(azimuth), _inclination(inclination)
		{ }

		// Construct from azimuth, inclination and radius.
		Spherical3(F32 azimuth, F32 inclination, F32 radius) : _radius(radius), _azimuth(azimuth), _inclination(inclination)
		{ }

		F32 _radius;
		F32 _azimuth;
		F32 _inclination;
	};

	// Now include the inline definitions.

#include "spherical3.inl"

// namespace libmaths
}

#pragma managed(pop)
