//********************************************************************************
//
// Defines simple 2D triangles. Intended mainly for collision and ray intersection
// testing.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector2.h"
#include "ray2.h"


namespace libmaths
{

	//********************************************************************************
	// Two dimensional triangle class for collision detection purposes.
	//********************************************************************************

	class Triangle2
	{

	public:

		// Construct an uninitialised triangle.
		Triangle2(void)
		{ }

		// Destroy triangle.
		~Triangle2(void)
		{ }

		// Construct a triangle with the given corner points.
		Triangle2(const Vector2& a, const Vector2& b, const Vector2& c)
		{
			_points[0] = a;
			_points[1] = b;
			_points[2] = c;
		}

		// Build a triangle from the given points
		void Set(const Vector2& a, const Vector2& b, const Vector2& c)
		{
			_points[0] = a;
			_points[1] = b;
			_points[2] = c;
		}

		// Access the i'th point of the triangle.
		// index may be 0, 1 or 2.
		// This is the subscript operator associated with square brackets, e.g. point[index]
		const Vector2& operator [] (S32 index) const
		{ LIBMATHS_ASSERT(index >= 0 && index < 3); return _points[index]; }

		// Retrieve a point in the triangle.
		Vector2 Point(U32 point) const
		{ LIBMATHS_ASSERT(point >= 0 && point < 3); return _points[point]; }

		// Retrieve an edge in the triangle.
		LineSegment2 Edge(U32 edge) const
		{ LIBMATHS_ASSERT(edge >= 0 && edge < 3); return LineSegment2(_points[edge], _points[(edge + 1) % 3] - _points[edge]); }

		// Check if the given point is inside the triangle.
		Bool Contains(const Vector2& point) const;

		// Check if the ray intersects this triangle.
		Bool Intersects(const Ray2& ray) const;

		// Check if the ray intersects this triangle.
		Bool Intersects(const Ray2& ray, Vector2& point, Vector2& normal) const;

		// Check if the line intersects this triangle.
		Bool Intersects(const LineSegment2& line) const
		{
			Vector2 point, normal;

			if (Intersects(Ray2(line), point, normal))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this triangle.
		Bool Intersects(const LineSegment2& line, Vector2& point, Vector2& normal, F32& t) const
		{
			if (Intersects(Ray2(line), point, normal))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the distance to and the point on the triangle closest to the given point.
		F32 Distance(const Vector2& point, Vector2& intersection) const;

		// Return the squared distance to and the point on the triangle closest to the given point.
		F32 DistanceSqr(const Vector2& point, Vector2& intersection) const;

		// Reverse the triangles winding order.
		void Flip(void)
		{ Swap(_points[1], _points[2]); }

	private:

		// Three corner points of the triangle.
		Vector2 _points[3];
	};

// namespace libmaths
}

#pragma managed(pop)
