//********************************************************************************
//
// Defines simple 3D triangles. Intended mainly for collision and ray intersection
// testing.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "plane.h"
#include "ray3.h"


namespace libmaths
{

	//********************************************************************************
	// Three dimensional triangle class for collision detection purposes.
	//********************************************************************************

	class Triangle3
	{

	public:

		// Construct an uninitialised triangle.
		Triangle3(void)
		{ }

		// Destroy triangle.
		~Triangle3(void)
		{ }

		// Construct a triangle with the given corner points.
		Triangle3(const Vector3x& a, const Vector3x& b, const Vector3x& c) :
			_plane(a, b, c)
		{
			_points[0] = a;
			_points[1] = b;
			_points[2] = c;
		}

		// Construct a triangle with the given corner points.
		Triangle3(const Vector3& a, const Vector3& b, const Vector3& c) :
			_plane(a, b, c)
		{
			_points[0] = Vector3x(a);
			_points[1] = Vector3x(b);
			_points[2] = Vector3x(c);
		}

		// Build a triangle from the given points
		void Set(const Vector3x& a, const Vector3x& b, const Vector3x& c)
		{
			_plane = libmaths::Plane(a, b, c);

			_points[0] = a;
			_points[1] = b;
			_points[2] = c;
		}

		// Build a triangle from the given points
		void Set(const Vector3& a, const Vector3& b, const Vector3& c)
		{
			_plane = libmaths::Plane(a, b, c);

			_points[0] = a;
			_points[1] = b;
			_points[2] = c;
		}

		// Access the i'th point of the triangle.
		// index may be 0, 1 or 2.
		// This is the subscript operator associated with square brackets, e.g. point[index]
		const Vector3x& operator[](S32 index) const
		{ LIBMATHS_ASSERT(index >= 0 && index < 3); return _points[index]; }

		// Retrieve a point in the triangle.
		Vector3x Point(U32 point) const
		{ LIBMATHS_ASSERT(point >= 0 && point < 3); return _points[point]; }

		// Retrieve an edge in the triangle.
		LineSegment3 Edge(U32 edge) const
		{ LIBMATHS_ASSERT(edge >= 0 && edge < 3); return LineSegment3(_points[edge], _points[(edge + 1) % 3] - _points[edge]); }

		// Check if the given point is inside the triangle.
		Bool Contains(const Vector3x& point) const;

		// Check if the ray intersects this triangle.
		Bool Intersects(const Ray3& ray, Bool twoSided = false) const;

		// Check if the ray intersects this triangle.
		Bool Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal, Bool twoSided = false) const;

		// Check if the line intersects this triangle.
		Bool Intersects(const LineSegment3& line, Bool twoSided = false) const
		{
			Vector3x point, normal;

			if (Intersects(Ray3(line), point, normal, twoSided))
			{
				return ((point - line.Origin()).LengthSqr() < line.Direction().LengthSqr());
			}

			return false;
		}

		// Check if the line intersects this triangle.
		Bool Intersects(const LineSegment3& line, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided = false) const
		{
			if (Intersects(Ray3(line), point, normal, twoSided))
			{
				t = (point - line.Origin()).Length() / line.Direction().Length();

				return (t < 1.0f);
			}

			return false;
		}

		// Return the area of the triangle.
		F32 Area(void)
		{ return (_points[1] - _points[0]).Cross(_points[2] - _points[0]).Length() * 0.5f; }

		// Return the distance to and the point on the triangle closest to the given point.
		F32 Distance(const Vector3x& point, Vector3x& intersection) const;

		// Return the squared distance to and the point on the triangle closest to the given point.
		F32 DistanceSqr(const Vector3x& point, Vector3x& intersection) const;

		// Return a reference to this triangle's plane.
		const Plane& Plane(void) const
		{ return _plane; }

		// Reverse the triangles winding order.
		void Flip(void)
		{ Swap(_points[1], _points[2]); _plane.Flip(); }

	private:

		// Plane containing the triangle.
		libmaths::Plane _plane;

		// Three corner points of the triangle.
		Vector3x _points[3];
	};

// namespace libmaths
}

#pragma managed(pop)
