//********************************************************************************
//
// 2D vector class.
// Often used for 2D work in screen coordinates and also for texture coordinates.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"


namespace libmaths
{

	//********************************************************************************
	// Forward declarations
	//********************************************************************************

	class Matrix22of33;
	class Matrix33;
	class Polar2;


	//********************************************************************************
	// 2D vector class.
	// Often used for 2D work in screen coordinates and also for texture coordinates.
	//********************************************************************************

	class Vector2
	{

	public:

		// Get the epsilon value for equality comparisons.
		static const F32 Epsilon(void)
		{ return libmaths::Epsilon; }

		// A fixed vector of all zeroes (0, 0).
		static Vector2 ZERO(void)
		{ return Vector2(0.0f, 0.0f); }

		// A fixed vector of all ones (1, 1).
		static Vector2 UNITY(void)
		{ return Vector2(1.0f, 1.0f); }

		// A fixed vector representing the positive x axis (1, 0).
		static Vector2 XAXIS(void)
		{ return Vector2(1.0f, 0.0f); }

		// A fixed vector representing the positive y axis (0, 1).
		static Vector2 YAXIS(void)
		{ return Vector2(0.0f, 1.0f); }

		// Default constructor.
		Vector2(void) : x(0.0f), y(0.0f)
		{ }

		// Copy constructor.
		Vector2(const Vector2& rhs) : x(rhs.x), y(rhs.y)
		{ }

		// Construct from 2 components.
		Vector2(F32 v0, F32 v1) : x(v0), y(v1)
		{ }

		// Construct from a polar coordinate, resulting in a right-handed vector.
		Vector2(const Polar2& rhs)
		{ *this = rhs; }

		// Assignment from a polar coordinate, resulting in a right-handed vector.
		Vector2& operator = (const Polar2& rhs);

		// Get the x component.
		F32 X(void) const
		{ return x; }

		// Set the x component.
		void X(F32 v0)
		{ x = v0; }

		// Get the y component.
		F32 Y(void) const
		{ return y; }

		// Set the y component.
		void Y(F32 v0)
		{ y = v0; }

		// Set the x and y components.
		void XY(const F32 v0, const F32 v1)
		{ x = v0, y = v1; }

		// Compare two vectors for equality.
		Bool operator == (const Vector2& rhs) const
		{ return (Abs(rhs.x - x) < Epsilon() && Abs(rhs.y - y) < Epsilon()); }

		// Get a new vector from the absolute values of another vectors components.
		friend Vector2 Abs(const Vector2& rhs)
		{ return Vector2(Abs(rhs.x), Abs(rhs.y)); }

		// Get the dot product between this vector and another.
		F32 Dot(const Vector2& rhs) const
		{ return *this * rhs; }

		// Set from minimum vector components into this vector from this vector and another.
		void MinOf(const Vector2& rhs)
		{ MinOf(*this, rhs); }

		// Set from maximum vector components into this vector from this vector and another.
		void MaxOf(const Vector2& rhs)
		{ MaxOf(*this, rhs); }

		// Set from minimum vector components into this vector from two other vectors.
		Vector2& MinOf(const Vector2& v0, const Vector2& v1);

		// Set from maximum vector components into this vector from two other vectors.
		Vector2& MaxOf(const Vector2& v0, const Vector2& v1);

		// Multiply (scale) this vector by a value.
		Vector2& Scale(const F32 scale);

		// Multiply (scale) this vector by another vector.
		Vector2& Scale(const Vector2& scale);

		// Get a new vector from this vector scaled by a value.
		Vector2 Scaled(const F32 scale) const
		{ return Vector2(*this).Scale(scale); }

		// Get a new vector from this vector scaled by another vector.
		Vector2 Scaled(const Vector2& scale) const
		{ return Vector2(*this).Scale(scale); }

		// Get the squared length of this vector.
		F32 LengthSqr(void) const
		{ return (x * x) + (y * y); }

		// Get the length of this vector.
		F32 Length(void) const
		{ return Sqrt(LengthSqr()); }

		// Get the fast length of this vector. (Up to 0.2% under, never over).
		F32 LengthFast(void) const
		{ return SqrtFast(LengthSqr()); }

		// Get the super fast length of this vector. (Up to 6% over, never under)
		F32 LengthSuperFast(void) const
		{ return SqrtSuperFast(LengthSqr()); }

		// Normalise this vector returning its length.
		F32 Normalise(void);

		// Normalise this vector fast returning its length. (Up to 0.2% under, never over).
		F32 NormaliseFast(void);

		// Normalise this vector super fast returning its length. (Up to 6% over, never under)
		F32 NormaliseSuperFast(void);

		// Get a new vector normalised from this vector.
		Vector2 Normalised(void) const;

		// Set the 2D point in this vector from a 3D point on its dominant plane.
		void GetDominantPlanePoint(S32 plane, const Vector3x& position);

		// Compare two vectors for inequality.
		Bool operator != (const Vector2& rhs) const
		{ return !(*this == rhs); }

		// Get a new vector that is the negative of this vector.
		Vector2 operator - (void) const
		{ return Vector2(-x, -y); }

		// Get a new vector from this vector scaled by a value.
		Vector2 operator * (F32 rhs) const
		{ return Vector2(x * rhs, y * rhs); }

		// Get a new vector from this vector scaled by a value.
		Vector2 operator / (F32 rhs) const
		{ return *this * (1.0f / rhs); }

		// Get the dot product between this vector and another.
		F32 operator * (const Vector2& rhs) const
		{ return (x * rhs.x) + (y * rhs.y); }

		// Get a new vector from this vector multiplied (transformed) by a 2x2 matrix.
		Vector2 operator * (const Matrix22of33& rhs) const;

		// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
		Vector2 operator * (const Matrix33& rhs) const;

		// Get a new vector from another vector added to this vector.
		Vector2 operator + (const Vector2& rhs) const
		{ return Vector2(x + rhs.x, y + rhs.y); }

		// Get a new vector from another vector subtracted from this vector.
		Vector2 operator - (const Vector2& rhs) const
		{ return Vector2(x - rhs.x, y - rhs.y); }

		// Add another vector to this vector.
		Vector2& operator += (const Vector2& rhs)
		{ x += rhs.x; y += rhs.y; return *this; }

		// Subtract another vector from this vector.
		Vector2& operator -= (const Vector2& rhs)
		{ x -= rhs.x; y -= rhs.y; return *this; }

		// Multiply (scale) this vector by a value.
		Vector2& operator *= (const F32 rhs)
		{ x *= rhs; y *= rhs; return *this; }

		// Multiply (scale) this vector by another vector.
		Vector2& operator *= (const Vector2& rhs)
		{ x *= rhs.x; y *= rhs.y; return *this; }

		// Multiply (transform) this vector by a 2x2 matrix.
		Vector2& operator *= (const Matrix22of33& rhs);

		// Multiply (transform) this vector by a 3x3 matrix.
		Vector2& operator *= (const Matrix33& rhs);

		// Divide (scale) this vector by a value.
		Vector2& operator /= (const F32 rhs)
		{ return *this /= (1.0f / rhs); }

		// Multiply (scale) this vector by a value (pre-multiply).
		friend Vector2 operator * (const F32 lhs, const Vector2& rhs)
		{ return Vector2(lhs * rhs.x, lhs * rhs.y); }

		// Access vector components as an array (read only).
		const F32 operator[] (const S32 i) const;

		// Access vector components as an array (read-write).
		F32& operator[] (const S32 i);

		// The vector components.
		F32 x;
		F32 y;
	};

	// Now include the inline definitions.

#include "vector2.inl"

// namespace libmaths
}

#pragma managed(pop)
