//********************************************************************************
//
// 2D vector class.
// Often used for 2D work in screen coordinates and also for texture coordinates.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Set from minimum vector components into this vector from two other vectors.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::MinOf(const Vector2& v0, const Vector2& v1)
{
	x = Min(v0.x, v1.x);
	y = Min(v0.y, v1.y);

	return (*this);
}


//********************************************************************************
// Set from maximum vector components into this vector from two other vectors.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::MaxOf(const Vector2& v0, const Vector2& v1)
{
	x = Max(v0.x, v1.x);
	y = Max(v0.y, v1.y);

	return (*this);
}


//********************************************************************************
// Multiply (scale) this vector by a value.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::Scale(const F32 scale)
{
	x *= scale;
	y *= scale;

	return *this;
}


//********************************************************************************
// Multiply (scale) this vector by another vector.
//********************************************************************************

LIBMATHS_INLINE Vector2& Vector2::Scale(const Vector2& scale)
{
	x *= scale.x;
	y *= scale.y;

	return *this;
}


//********************************************************************************
// Normalise this vector returning its length.
//********************************************************************************

LIBMATHS_INLINE F32 Vector2::Normalise(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrt(t);

		x *= t;
		y *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector fast returning its length. (Up to 0.2% under, never
// over).
//********************************************************************************

LIBMATHS_INLINE F32 Vector2::NormaliseFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtFast(t);

		x *= t;
		y *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector super fast returning its length. (Up to 6% over, never
// under)
//********************************************************************************

LIBMATHS_INLINE F32 Vector2::NormaliseSuperFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtSuperFast(t);

		x *= t;
		y *= t;
	}

	return t;
}


//********************************************************************************
// Get a new vector normalised from this vector.
//********************************************************************************

LIBMATHS_INLINE Vector2 Vector2::Normalised() const
{
	Vector2 vector(*this);

	vector.Normalise();

	return vector;
}


//********************************************************************************
// Set the 2D point in this vector from a 3D point on its dominant plane.
//********************************************************************************

LIBMATHS_INLINE void Vector2::GetDominantPlanePoint(S32 plane, const Vector3x& position)
{
	const S32* pOrder = GetDominantPlaneOrder(plane);

	x = position[pOrder[0]];
	y = position[pOrder[1]];
}


//********************************************************************************
// Access vector components as an array (read only).
//********************************************************************************

LIBMATHS_INLINE const F32 Vector2::operator[] (const S32 i) const
{
	LIBMATHS_ASSERT(i >= 0 && i < 2);
	return reinterpret_cast<const F32*>(&x)[i];
}


//********************************************************************************
// Access vector components as an array (read-write).
//********************************************************************************

LIBMATHS_INLINE F32& Vector2::operator[] (const S32 i)
{
	LIBMATHS_ASSERT(i >= 0 && i < 2);
	return reinterpret_cast<F32*>(&x)[i];
}

#pragma managed(pop)
