//********************************************************************************
//
// 3D vector class.
// Most commonly used for 3D work - in 3 and 4 component flavours (4th component
// is ignored but is present for performance reasons).
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "libmaths.h"


namespace libmaths
{

	//********************************************************************************
	// Forward declarations
	//********************************************************************************

	class Matrix33;
	class Matrix33of44;
	class Matrix44;
	class Quaternion;
	class Vector3x;
	class Spherical3LH;
	class Spherical3RH;


	//********************************************************************************
	// 3D vector class with 3 components.
	//********************************************************************************

	class Vector3
	{

	public:

		// Default constructor.
		Vector3(void) : x(0.0f), y(0.0f), z(0.0f)
		{ }

		// Copy constructor.
		Vector3(const Vector3& rhs) : x(rhs.x), y(rhs.y), z(rhs.z)
		{ }

		// Construct from 3 components.
		Vector3(F32 v0, F32 v1, F32 v2) : x(v0), y(v1), z(v2)
		{ }

#define VECTOR3_TYPE Vector3
#define OTHER_VECTOR3_TYPE Vector3x
#include "vector3common.h"
#undef OTHER_VECTOR3_TYPE
#undef VECTOR3_TYPE

		// The vector components.
		F32 x;
		F32 y;
		F32 z;
	};


	//********************************************************************************
	// 3D vector class with 4 components.
	// Optimised for modern architectures by including a 4th element and requiring 128
	// bit alignment for the storage. The 4-th component does not take part in any
	// internal calculations and is available to store additional data.
	//********************************************************************************

	class Vector3x
	{

	public:

		// Default constructor.
		Vector3x(void) : x(0.0f), y(0.0f), z(0.0f), w(0.0f)
		{ }

		// Copy constructor.
		Vector3x(const Vector3x& rhs) : x(rhs.x), y(rhs.y), z(rhs.z), w(rhs.w)
		{ }

		// Construct from 3 components (last component defaults to 0).
		Vector3x(F32 v0, F32 v1, F32 v2) : x(v0), y(v1), z(v2), w(0.0f)
		{ }

		// Construct from 4 components.
		Vector3x(F32 v0, F32 v1, F32 v2, F32 v3) : x(v0), y(v1), z(v2), w(v3)
		{ }

#define VECTOR3_TYPE Vector3x
#define OTHER_VECTOR3_TYPE Vector3
#include "vector3common.h"
#undef OTHER_VECTOR3_TYPE
#undef VECTOR3_TYPE

		// Set the w component.
		void W(const F32 f)
		{ w = f; }

		// Set the w component as an integer.
		void WInt(const S32 i)
		{ *(S32 *)&w = i; }

		// Get the w component
		F32 W() const
		{ return w; }

		// Get the w component as an integer.
		S32 WInt() const
		{ return *(S32 *)&w; }

		// The vector components.
		F32 x;
		F32 y;
		F32 z;
		F32 w;
	};

	// Now include the inline definitions.

#include "vector3.inl"

// namespace libmaths
}

#pragma managed(pop)
