//********************************************************************************
//
// 3D vector class.
// Most commonly used for 3D work - in 3 and 4 component flavours (4th component
// is ignored but is present for performance reasons).
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


// Get the epsilon value for equality comparisons.
static const F32 Epsilon(void)
{ return libmaths::Epsilon; }

// A fixed vector of all zeroes (0, 0, 0).
static VECTOR3_TYPE ZERO(void)
{ return VECTOR3_TYPE(0.0f, 0.0f, 0.0f); }

// A fixed vector of all ones (1, 1, 1).
static VECTOR3_TYPE UNITY(void)
{ return VECTOR3_TYPE(1.0f, 1.0f, 1.0f); }

// A fixed vector representing the positive x axis (1, 0, 0).
static VECTOR3_TYPE XAXIS(void)
{ return VECTOR3_TYPE(1.0f, 0.0f, 0.0f); }

// A fixed vector representing the positive y axis (0, 1, 0).
static VECTOR3_TYPE YAXIS(void)
{ return VECTOR3_TYPE(0.0f, 1.0f, 0.0f); }

// A fixed vector representing the positive z axis (0, 0, 1).
static VECTOR3_TYPE ZAXIS(void)
{ return VECTOR3_TYPE(0.0f, 0.0f, 1.0f); }

// Calculate the normal for a vector.
static VECTOR3_TYPE Normal(const VECTOR3_TYPE& point1, const VECTOR3_TYPE& point2, const VECTOR3_TYPE& point3)
{ return VECTOR3_TYPE((point1 - point2) ^ (point3 - point2)).Normalised(); }

// Construct from other vector type.
VECTOR3_TYPE(const OTHER_VECTOR3_TYPE& rhs);

// Construct from a spherical coordinate, resulting in a right-handed vector.
VECTOR3_TYPE(const Spherical3RH& rhs)
{ *this = rhs; }

// Construct from a spherical coordinate, resulting in a left-handed vector.
VECTOR3_TYPE(const Spherical3LH& rhs)
{ *this = rhs; }

// Assignment from another vector.
VECTOR3_TYPE& operator = (const VECTOR3_TYPE& rhs);

// Assignment from an OTHER_VECTOR3_TYPE.
VECTOR3_TYPE& operator = (const OTHER_VECTOR3_TYPE& rhs);

// Assignment from a spherical coordinate, resulting in a right-handed vector.
VECTOR3_TYPE& operator = (const Spherical3RH& rhs);

// Assignment from a spherical coordinate, resulting in a left-handed vector.
VECTOR3_TYPE& operator = (const Spherical3LH& rhs);

// Get the x component.
F32 X(void) const
{ return x; }

// Set the x component.
void X(F32 v0)
{ x = v0; }

// Get the y component.
F32 Y(void) const
{ return y; }

// Set the y component.
void Y(F32 v0)
{ y = v0; }

// Get the z component.
F32 Z(void) const
{ return z; }

// Set the z component.
void Z(F32 v0)
{ z = v0; }

// Set the x, y and z components.
void XYZ(F32 v0, F32 v1, F32 v2)
{ x = v0; y = v1; z = v2; }

// Get a new vector from the absolute values of another vectors components.
friend VECTOR3_TYPE Abs(const VECTOR3_TYPE& rhs)
{ return VECTOR3_TYPE(Abs(rhs.x), Abs(rhs.y), Abs(rhs.z)); }

// Dot product.
F32 Dot(const VECTOR3_TYPE& rhs) const
{ return *this * rhs; }

// Get a new vector that is the cross product between this vector and another.
VECTOR3_TYPE Cross(const VECTOR3_TYPE& rhs) const
{ return *this ^ rhs; }

// Set this vector to the cross product of two other vectors.
VECTOR3_TYPE& Cross(const VECTOR3_TYPE& lhs, const VECTOR3_TYPE& rhs)
{ *this = lhs ^ rhs; return *this; }

// Set from minimum vector components into this vector from this vector and another.
void MinOf(const VECTOR3_TYPE& rhs)
{ MinOf(*this, rhs); }

// Set from maximum vector components into this vector from this vector and another.
void MaxOf(const VECTOR3_TYPE& rhs)
{ MaxOf(*this, rhs); }

// Set from minimum vector components into this vector from two other vectors.
VECTOR3_TYPE& MinOf(const VECTOR3_TYPE& lhs, const VECTOR3_TYPE& rhs);

// Set from maximum vector components into this vector from two other vectors.
VECTOR3_TYPE& MaxOf(const VECTOR3_TYPE& lhs, const VECTOR3_TYPE& rhs);

// Multiply (scale) this vector by a value.
VECTOR3_TYPE& Scale(const F32 scale);

// Multiply (scale) this vector by another vector.
VECTOR3_TYPE& Scale(const VECTOR3_TYPE& scale);

// Get a new vector from this vector scaled by a value.
VECTOR3_TYPE Scaled(const F32 scale) const
{ return VECTOR3_TYPE(*this).Scale(scale); }

// Get a new vector from this vector scaled by another vector.
VECTOR3_TYPE Scaled(const VECTOR3_TYPE& scale) const
{ return VECTOR3_TYPE(*this).Scale(scale); }

// Get the squared length of this vector.
F32 LengthSqr(void) const
{ return (x * x) + (y * y) + (z * z); }

// Get the length of this vector.
F32 Length(void) const
{ return Sqrt(LengthSqr()); }

// Get the fast length of this vector. (Up to 0.2% under, never over).
F32 LengthFast(void) const
{ return SqrtFast(LengthSqr()); }

// Get the super fast length of this vector. (Up to 6% over, never under)
F32 LengthSuperFast(void) const
{ return SqrtSuperFast(LengthSqr()); }

// Normalise this vector returning its length.
F32 Normalise(void);

// Normalise this vector fast returning its length. (Up to 0.2% under, never over).
F32 NormaliseFast(void);

// Normalise this vector super fast returning its length. (Up to 6% over, never under)
F32 NormaliseSuperFast(void);

// Get a new vector normalised from this vector.
VECTOR3_TYPE Normalised(void) const;

// Get the dominant plane for the vector.
S32 GetDominantPlane(void) const;

// Calculate the right and up vectors for a vector.
void RightUpVectors(VECTOR3_TYPE& right, VECTOR3_TYPE& up) const;

// Project this vector onto another vector.
VECTOR3_TYPE& Project(const VECTOR3_TYPE& onto);

// Get a new vector from this vector projected onto another vector.
VECTOR3_TYPE Projected(const VECTOR3_TYPE& onto) const;

// Get the left handed equivalent of a right handed vector.
VECTOR3_TYPE LeftHanded(void) const;

// Get the right handed equivalent of a left handed vector.
VECTOR3_TYPE RightHanded(void) const;

// Compare two vectors for equality.
Bool operator == (const VECTOR3_TYPE& rhs) const
{ return (Abs(rhs.x - x) < Epsilon() && Abs(rhs.y - y) < Epsilon() && Abs(rhs.z - z) < Epsilon()); }

// Compare two vectors for inequality.
Bool operator != (const VECTOR3_TYPE& rhs) const
{ return !((*this) == rhs); }

// Get a new vector that is the negative of this vector.
VECTOR3_TYPE operator - (void) const
{ return VECTOR3_TYPE(-x, -y, -z); }

// Get a new vector from this vector scaled by a value.
VECTOR3_TYPE operator * (F32 rhs) const
{ return VECTOR3_TYPE(x * rhs, y * rhs, z * rhs); }

// Get a new vector from this vector scaled by a value.
VECTOR3_TYPE operator / (F32 rhs) const
{ return *this * (1.0f / rhs); }

// Get the dot product between this vector and another.
F32 operator * (const VECTOR3_TYPE& rhs) const
{ return (x * rhs.x) + (y * rhs.y) + (z * rhs.z); }

// Get a new vector that is the cross product between this vector and another.
VECTOR3_TYPE operator ^ (const VECTOR3_TYPE& rhs) const
{ return VECTOR3_TYPE(y * rhs.z - z * rhs.y, z * rhs.x - x * rhs.z, x * rhs.y - y * rhs.x); }

// Get a new vector from another vector added to this vector.
VECTOR3_TYPE operator + (const VECTOR3_TYPE& rhs) const
{ return VECTOR3_TYPE(x + rhs.x, y + rhs.y, z + rhs.z); }

// Get a new vector from another vector subtracted from this vector.
VECTOR3_TYPE operator - (const VECTOR3_TYPE& rhs) const
{ return VECTOR3_TYPE(x - rhs.x, y - rhs.y, z - rhs.z); }

// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
VECTOR3_TYPE operator * (const Matrix33& rhs) const;

// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
VECTOR3_TYPE operator * (const Matrix33of44& rhs) const;

// Get a new vector from this vector multiplied (transformed) by a 4x4 matrix.
VECTOR3_TYPE operator * (const Matrix44& rhs) const;

// Get a new vector from this vector multiplied (transformed) by a quaternion.
VECTOR3_TYPE operator * (const Quaternion& rhs) const;

// Add another vector to this vector.
VECTOR3_TYPE& operator += (const VECTOR3_TYPE& rhs)
{ x += rhs.x; y += rhs.y; z += rhs.z; return *this; }

// Subtract another vector from this vector.
VECTOR3_TYPE& operator -= (const VECTOR3_TYPE& rhs)
{ x -= rhs.x; y -= rhs.y; z -= rhs.z; return *this; }

// Multiply (scale) this vector by a value.
VECTOR3_TYPE& operator *= (const F32 rhs)
{ x *= rhs; y *= rhs; z *= rhs; return *this; }

// Multiply (transform) this vector by a 3x3 matrix.
VECTOR3_TYPE& operator *= (const Matrix33& rhs);

// Multiply (transform) this vector by a 3x3 matrix.
VECTOR3_TYPE& operator *= (const Matrix33of44& rhs);

// Multiply (transform) this vector by a 4x4 matrix.
VECTOR3_TYPE& operator *= (const Matrix44& rhs);

// Multiply (transform) this vector by a quaternion.
VECTOR3_TYPE& operator *= (const Quaternion& rhs);

// Divide (scale) this vector by a value.
VECTOR3_TYPE& operator /= (const F32 rhs)
{ return *this *= (1.0f / rhs); }

// Set this vector to its cross product with another vector.
VECTOR3_TYPE& operator ^= (const VECTOR3_TYPE& rhs);

// Multiply (scale) this vector by a value (pre-multiply).
friend VECTOR3_TYPE operator * (const F32 lhs, const VECTOR3_TYPE& rhs)
{ return VECTOR3_TYPE(lhs * rhs.x, lhs * rhs.y, lhs * rhs.z); }

// Access vector components as an array (read only).
const F32 operator [] (const S32 i) const;

// Access vector components as an array (read-write).
F32& operator [] (const S32 i);
