//********************************************************************************
//
// 3D vector class.
// Most commonly used for 3D work - in 3 and 4 component flavours (4th component
// is ignored but is present for performance reasons).
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Set this vector's components from the minimum components of two vectors.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::MinOf(const VECTOR3_TYPE& v0, const VECTOR3_TYPE& v1)
{
	x = Min(v0.x, v1.x);
	y = Min(v0.y, v1.y);
	z = Min(v0.z, v1.z);

	return (*this);
}


//********************************************************************************
// Set this vector's components from the maximum components of two vectors.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::MaxOf(const VECTOR3_TYPE& v0, const VECTOR3_TYPE& v1)
{
	x = Max(v0.x, v1.x);
	y = Max(v0.y, v1.y);
	z = Max(v0.z, v1.z);

	return (*this);
}


//********************************************************************************
// Multiply (scale) this vector by a value.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::Scale(const F32 scale)
{
	x *= scale;
	y *= scale;
	z *= scale;

	return *this;
}


//********************************************************************************
// Multiply (scale) this vector by another vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::Scale(const VECTOR3_TYPE& scale)
{
	x *= scale.X();
	y *= scale.Y();
	z *= scale.Z();

	return *this;
}


//********************************************************************************
// Normalise this vector returning its length.
//********************************************************************************

LIBMATHS_INLINE F32 VECTOR3_TYPE::Normalise(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrt(t);

		x *= t;
		y *= t;
		z *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector fast returning its length. (Up to 0.2% under, never
// over).
//********************************************************************************

LIBMATHS_INLINE F32 VECTOR3_TYPE::NormaliseFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtFast(t);

		x *= t;
		y *= t;
		z *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector super fast returning its length. (Up to 6% over, never
// under)
//********************************************************************************

LIBMATHS_INLINE F32 VECTOR3_TYPE::NormaliseSuperFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtSuperFast(t);

		x *= t;
		y *= t;
		z *= t;
	}

	return t;
}


//********************************************************************************
// Get a new vector normalised from this vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE VECTOR3_TYPE::Normalised(void) const
{
	VECTOR3_TYPE vector(*this);

	vector.Normalise();

	return vector;
}



//********************************************************************************
// Get the dominant plane for the vector.
//********************************************************************************

LIBMATHS_INLINE S32 VECTOR3_TYPE::GetDominantPlane(void) const
{
	U32 fx = (*(U32 *)&x) & 0x7fffffff;
	U32 fy = (*(U32 *)&y) & 0x7fffffff;
	U32 fz = (*(U32 *)&z) & 0x7fffffff;

	if ((fx > fy) && (fx > fz))
	{
		return 0;
	}
	else if (fy > fz)
	{
		return 1;
	}
	else
	{
		return 2;
	}
}


//********************************************************************************
// Calculate the right and up vectors for a vector.
//********************************************************************************

LIBMATHS_INLINE void VECTOR3_TYPE::RightUpVectors(VECTOR3_TYPE& right, VECTOR3_TYPE& up) const
{
	VECTOR3_TYPE other(up);

	if (Abs(*this * other) > 0.9999f)
	{
		other.XYZ(up.Y(), up.Z(), up.X());
	}

	right = Cross(other);
	right.Normalise();

	up = right.Cross(*this);
}


//********************************************************************************
// Get a new vector from this vector projected onto another vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE VECTOR3_TYPE::Projected(const VECTOR3_TYPE& onto) const
{
	return ((*this * onto) * (1.0f / onto.LengthSqr())) * onto;
}


//********************************************************************************
// Project this vector onto another vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::Project(const VECTOR3_TYPE& onto)
{
	*this = Projected(onto);

	return *this;
}


//********************************************************************************
// Get the left handed equivalent of a right handed vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE VECTOR3_TYPE::LeftHanded(void) const
{
	return VECTOR3_TYPE(x, z, -y);
}


//********************************************************************************
// Get the right handed equivalent of a left handed vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE VECTOR3_TYPE::RightHanded(void) const
{
	return VECTOR3_TYPE(x, -z, y);
}


//********************************************************************************
// Set this vector to its cross product with another vector.
//********************************************************************************

LIBMATHS_INLINE VECTOR3_TYPE& VECTOR3_TYPE::operator ^= (const VECTOR3_TYPE& rhs)
{
	F32 tx = y * rhs.z - z * rhs.y;
	F32 ty = z * rhs.x - x * rhs.z;
	F32 tz = x * rhs.y - y * rhs.x;

	x = tx;
	y = ty;
	z = tz;

	return *this;
}


//********************************************************************************
// Access vector components as an array (read only).
//********************************************************************************

LIBMATHS_INLINE const F32 VECTOR3_TYPE::operator [] (const S32 i) const
{
	LIBMATHS_ASSERT(i >= 0 && i < 3);
	return reinterpret_cast<const F32*>(&x)[i];
}


//********************************************************************************
// Access vector components as an array (read-write).
//********************************************************************************

LIBMATHS_INLINE F32& VECTOR3_TYPE::operator [] (const S32 i)
{
	LIBMATHS_ASSERT(i >= 0 && i < 3);
	return reinterpret_cast<F32*>(&x)[i];
}
