//********************************************************************************
//
// 4D vector class.
// Used in homogenous 4 component vector work.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "plane.h"


namespace libmaths
{

	//********************************************************************************
	// 4D vector class.
	// Used in homogenous 4 component vector work.
	//********************************************************************************

	class Vector4
	{

	public:

		// Get the epsilon value for equality comparisons.
		static const F32 Epsilon(void)
		{ return libmaths::Epsilon; }

		// A fixed vector of all zeroes (0, 0, 0, 0).
		static Vector4 ZERO(void)
		{ return Vector4(0.0f, 0.0f, 0.0f, 0.0f); }

		// A fixed vector of all ones (1, 1, 1, 1).
		static Vector4 UNITY(void)
		{ return Vector4(1.0f, 1.0f, 1.0f, 1.0f); }

		// A fixed vector representing the positive x axis (1, 0, 0).
		static Vector4 XAXIS(void)
		{ return Vector4(1.0f, 0.0f, 0.0f); }

		// A fixed vector representing the positive y axis (0, 1, 0).
		static Vector4 YAXIS(void)
		{ return Vector4(0.0f, 1.0f, 0.0f); }

		// A fixed vector representing the positive z axis (0, 0, 1).
		static Vector4 ZAXIS(void)
		{ return Vector4(0.0f, 0.0f, 1.0f); }

		// Calculate the normal for a vector.
		static Vector4 Normal(const Vector4& point1, const Vector4& point2, const Vector4& point3)
		{ return Vector4((point1 - point2) ^ (point3 - point2)).Normalised(); }

		// Default constructor.
		Vector4(void) : x(0.0f), y(0.0f), z(0.0f), w(0.0f)
		{ }

		// Copy constructor.
		Vector4(const Vector4& rhs) : x(rhs.x), y(rhs.y), z(rhs.z), w(rhs.w)
		{ }

		// Construct from 3 components (last component defaults to 1).
		Vector4(F32 v0, F32 v1, F32 v2) : x(v0), y(v1), z(v2), w(1.0f)
		{ }

		// Construct from 4 components.
		Vector4(F32 v0, F32 v1, F32 v2, F32 v3) : x(v0), y(v1), z(v2), w(v3)
		{ }

		// Construct from a 3D vector (last component defaults to 1).
		Vector4(const Vector3x& rhs) : x(rhs.x), y(rhs.y), z(rhs.z), w(1.0f)
		{ }

		// Construct from a 3D vector and a scalar.
		Vector4(const Vector3x& rhs, F32 scalar) : x(rhs.x), y(rhs.y), z(rhs.z), w(scalar)
		{ }

		// Construct from a plane.
		Vector4(const Plane& plane) : x(plane.Normal().X()), y(plane.Normal().Y()), z(plane.Normal().Z()), w(-(plane.Normal() * plane.Origin()))
		{ }

		// Assignment from a 3D vector.
		Vector4& operator = (const Vector3x& rhs);

		// Assignment from a plane.
		Vector4& operator = (const Plane& rhs);

		// Get the x component.
		F32 X(void) const
		{ return x; }

		// Set the x component.
		void X(F32 v0)
		{ x = v0; }

		// Get the y component.
		F32 Y(void) const
		{ return y; }

		// Set the y component.
		void Y(F32 v0)
		{ y = v0; }

		// Get the z component.
		F32 Z(void) const
		{ return z; }

		// Set the z component.
		void Z(F32 v0)
		{ z = v0; }

		// Get the w component.
		F32 W(void) const
		{ return w; }

		// Set the w component.
		void W(F32 v0)
		{ w = v0; }

		// Get the w component as an integer.
		S32 WInt(void) const
		{ return reinterpret_cast<const S32*>(&x)[3]; }

		// Set the w component as an integer.
		void WInt(S32 v0)
		{ reinterpret_cast<S32*>(&x)[3] = v0; }

		// Set the x, y and z components.
		void XYZ(F32 v0, F32 v1, F32 v2)
		{ x = v0; y = v1; z = v2; }

		// Set the x, y, z, and w components.
		void XYZW(F32 v0, F32 v1, F32 v2, F32 v3)
		{ x = v0; y = v1; z = v2; w = v3; }

		// Dot product.
		F32 Dot(const Vector4& rhs) const
		{ return *this * rhs; }

		// Get a new vector that is the cross product between this vector and another.
		// NOTE: This functions like a 3D vector with 1 in the w component.
		Vector4 Cross(const Vector4& rhs) const
		{ return *this ^ rhs; }

		// Set this vector to the cross product of two other vectors.
		// NOTE: This functions like a 3D vector with 1 in the w component.
		Vector4& Cross(const Vector4& lhs, const Vector4& rhs)
		{ *this = lhs ^ rhs; return *this; }

		// Set from minimum vector components into this vector from this vector and another.
		void MinOf(const Vector4& rhs)
		{ MinOf(*this, rhs); }

		// Set from maximum vector components into this vector from this vector and another.
		void MaxOf(const Vector4& rhs)
		{ MaxOf(*this, rhs); }

		// Set from minimum vector components into this vector from two other vectors.
		Vector4& MinOf(const Vector4& lhs, const Vector4& rhs);

		// Set from maximum vector components into this vector from two other vectors.
		Vector4& MaxOf(const Vector4& lhs, const Vector4& rhs);

		// Multiply (scale) this vector by a value.
		Vector4& Scale(const F32 scale);

		// Multiply (scale) this vector by another vector.
		Vector4& Scale(const Vector4& scale);

		// Get a new vector from this vector scaled by a value.
		Vector4 Scaled(const F32 scale) const
		{ return Vector4(*this).Scale(scale); }

		// Get a new vector from this vector scaled by another vector.
		Vector4 Scaled(const Vector4& scale) const
		{ return Vector4(*this).Scale(scale); }

		// Get the squared length of this vector.
		F32 LengthSqr(void) const
		{ return (x * x) + (y * y) + (z * z) + (w * w); }

		// Get the length of this vector.
		F32 Length(void) const
		{ return Sqrt(LengthSqr()); }

		// Get the fast length of this vector. (Up to 0.2% under, never over).
		F32 LengthFast(void) const
		{ return SqrtFast(LengthSqr()); }

		// Get the super fast length of this vector. (Up to 6% over, never under)
		F32 LengthSuperFast(void) const
		{ return SqrtSuperFast(LengthSqr()); }

		// Normalise this vector returning its length.
		// NOTE: This also normalises the homogenous w component.
		F32 Normalise(void);

		// Normalise this vector fast returning its length. (Up to 0.2% under, never over).
		// NOTE: This also normalises the homogenous w component.
		F32 NormaliseFast(void);

		// Normalise this vector super fast returning its length. (Up to 6% over, never under)
		// NOTE: This also normalises the homogenous w component.
		F32 NormaliseSuperFast(void);

		// Get a new vector normalised from this vector.
		// NOTE: This also normalises the homogenous w component.
		Vector4 Normalised(void) const;

		// Get the dominant plane for the vector (purposefully ignoring the W component).
		S32 GetDominantPlane(void) const;

		// Compare two vectors for equality.
		Bool operator == (const Vector4& rhs) const
		{ return (Abs(rhs.x - x) < Epsilon() && Abs(rhs.y - y) < Epsilon() && Abs(rhs.z - z) < Epsilon() && Abs(rhs.w - w) < Epsilon()); }

		// Compare two vectors for inequality.
		Bool operator != (const Vector4& rhs) const
		{ return !((*this) == rhs); }

		// Get a new vector that is the negative of this vector.
		Vector4 operator - (void) const
		{ return Vector4(-x, -y, -z, -w); }

		// Get a new vector from this vector scaled by a value.
		Vector4 operator * (F32 rhs) const
		{ return Vector4(x * rhs, y * rhs, z * rhs, w * rhs); }

		// Get a new vector from this vector scaled by a value.
		Vector4 operator / (F32 rhs) const
		{ return *this * (1.0f / rhs); }

		// Get the dot product between this vector and another.
		F32 operator * (const Vector4& rhs) const
		{ return (x * rhs.x) + (y * rhs.y) + (z * rhs.z) + (w * rhs.w); }

		// Get a new vector that is the cross product between this vector and another.
		// NOTE: This functions like a 3D vector with 1 in the w component.
		Vector4 operator ^ (const Vector4& rhs) const
		{ return Vector4(y * rhs.z - z * rhs.y, z * rhs.x - x * rhs.z, x * rhs.y - y * rhs.x, 1.0f); }

		// Get a new vector from another vector added to this vector.
		Vector4 operator + (const Vector4& rhs) const
		{ return Vector4(x + rhs.x, y + rhs.y, z + rhs.z, w + rhs.w); }

		// Get a new vector from another vector subtracted from this vector.
		Vector4 operator - (const Vector4& rhs) const
		{ return Vector4(x - rhs.x, y - rhs.y, z - rhs.z, w - rhs.w); }

		// Get a new vector from this vector multiplied (transformed) by a 3x3 matrix.
		Vector4 operator * (const Matrix33of44& rhs) const;

		// Get a new vector from this vector multiplied (transformed) by a 4x4 matrix.
		Vector4 operator * (const Matrix44& rhs) const;

		// Get a new vector from this vector multiplied (transformed) by a quaternion.
		Vector4 operator * (const Quaternion& rhs) const;

		// Add another vector to this vector.
		Vector4& operator += (const Vector4& rhs)
		{ x += rhs.x; y += rhs.y; z += rhs.z; w += rhs.w; return *this; }

		// Subtract another vector from this vector.
		Vector4& operator -= (const Vector4& rhs)
		{ x -= rhs.x; y -= rhs.y; z -= rhs.z; w -= rhs.w; return *this; }

		// Multiply (scale) this vector by a value.
		Vector4& operator *= (const F32 rhs)
		{ x *= rhs; y *= rhs; z *= rhs; w *= rhs; return *this; }

		// Multiply (transform) this vector by a 3x3 matrix.
		Vector4& operator *= (const Matrix33of44& rhs);

		// Multiply (transform) this vector by a 4x4 matrix.
		Vector4& operator *= (const Matrix44& rhs);

		// Multiply (transform) this vector by a quaternion.
		Vector4& operator *= (const Quaternion& rhs);

		// Divide (scale) this vector by a value.
		Vector4& operator /= (const F32 rhs)
		{ return *this *= (1.0f / rhs); }

		// Set this vector to its cross product with another vector.
		// NOTE: This functions like a 3D vector with 1 in the w component.
		Vector4& operator ^= (const Vector4& rhs);

		// Multiply (scale) this vector by a value (pre-multiply).
		friend Vector4 operator * (const F32 lhs, const Vector4& rhs)
		{ return Vector4(lhs * rhs.x, lhs * rhs.y, lhs * rhs.z, lhs * rhs.w); }

		// Get a new vector from the absolute values of another vectors components.
		friend Vector4 Abs(const Vector4& rhs)
		{ return Vector4(Abs(rhs.x), Abs(rhs.y), Abs(rhs.z), Abs(rhs.w)); }

		// Access vector components as an array (read only).
		const F32 operator[] (const S32 i) const;

		// Access vector components as an array (read-write).
		F32& operator[] (const S32 i);

		// Silent conversion to a 3D vector discarding the w component.
		operator Vector3x(void) const;

		// The vector components.
		F32 x;
		F32 y;
		F32 z;
		F32 w;
	};

	// Now include the inline definitions.

#include "vector4.inl"

// namespace libmaths
}

#pragma managed(pop)
