//********************************************************************************
//
// 4D vector class.
// Used in homogenous 4 component vector work.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once
#pragma managed(push,off)


//********************************************************************************
// Assignment from a 3D vector.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::operator = (const Vector3x& rhs)
{
	x = rhs.X();
	y = rhs.Y();
	z = rhs.Z();
	w = 1.0f;

	return *this;
}


//********************************************************************************
// Assignment from a plane.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::operator = (const Plane& rhs)
{
	*this = rhs.Normal();

	w = -(rhs.Normal() * rhs.Origin());

	return *this;
}


//********************************************************************************
// Set this vector's components from the minimum components of two vectors.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::MinOf(const Vector4& v0, const Vector4& v1)
{
	x = Min(v0.x, v1.x);
	y = Min(v0.y, v1.y);
	z = Min(v0.z, v1.z);
	w = Min(v0.w, v1.w);

	return (*this);
}


//********************************************************************************
// Set this vector's components from the maximum components of two vectors.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::MaxOf(const Vector4& v0, const Vector4& v1)
{
	x = Max(v0.x, v1.x);
	y = Max(v0.y, v1.y);
	z = Max(v0.z, v1.z);
	w = Max(v0.w, v1.w);

	return (*this);
}


//********************************************************************************
// Multiply (scale) this vector by a value.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::Scale(const F32 scale)
{
	x *= scale;
	y *= scale;
	z *= scale;
	w *= scale;

	return *this;
}


//********************************************************************************
// Multiply (scale) this vector by another vector.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::Scale(const Vector4& scale)
{
	x *= scale.X();
	y *= scale.Y();
	z *= scale.Z();
	w *= scale.W();

	return *this;
}


//********************************************************************************
// Normalise this vector returning its length.
// NOTE: This also normalises the homogenous w component.
//********************************************************************************

LIBMATHS_INLINE F32 Vector4::Normalise(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrt(t);

		x *= t;
		y *= t;
		z *= t;
		w *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector fast returning its length. (Up to 0.2% under, never
// over).
// NOTE: This also normalises the homogenous w component.
//********************************************************************************

LIBMATHS_INLINE F32 Vector4::NormaliseFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtFast(t);

		x *= t;
		y *= t;
		z *= t;
		w *= t;
	}

	return t;
}


//********************************************************************************
// Normalise this vector super fast returning its length. (Up to 6% over, never
// under)
// NOTE: This also normalises the homogenous w component.
//********************************************************************************

LIBMATHS_INLINE F32 Vector4::NormaliseSuperFast(void)
{
	F32 t = LengthSqr();

	LIBMATHS_ASSERT(t > FLOAT_NORMALISE_EPSILON);

	if (t > FLOAT_NORMALISE_EPSILON)
	{
		t = InvSqrtSuperFast(t);

		x *= t;
		y *= t;
		z *= t;
		w *= t;
	}

	return t;
}


//********************************************************************************
// Get a new vector normalised from this vector.
// NOTE: This also normalises the homogenous w component.
//********************************************************************************

LIBMATHS_INLINE Vector4 Vector4::Normalised(void) const
{
	Vector4 vector(*this);

	vector.Normalise();

	return vector;
}


//********************************************************************************
// Get the dominant plane for the vector (purposefully ignoring the W component).
//********************************************************************************

LIBMATHS_INLINE S32 Vector4::GetDominantPlane(void) const
{
	U32 fx = (*(U32 *)&x) & 0x7fffffff;
	U32 fy = (*(U32 *)&y) & 0x7fffffff;
	U32 fz = (*(U32 *)&z) & 0x7fffffff;

	if ((fx > fy) && (fx > fz))
	{
		return 0;
	}
	else if (fy > fz)
	{
		return 1;
	}
	else
	{
		return 2;
	}
}


//********************************************************************************
// Set this vector to its cross product with another vector.
// NOTE: This functions like a 3D vector with 1 in the w component.
//********************************************************************************

LIBMATHS_INLINE Vector4& Vector4::operator ^= (const Vector4& rhs)
{
	F32 tx = y * rhs.z - z * rhs.y;
	F32 ty = z * rhs.x - x * rhs.z;
	F32 tz = x * rhs.y - y * rhs.x;

	x = tx;
	y = ty;
	z = tz;
	w = 1.0f;

	return *this;
}


//********************************************************************************
// Access vector components as an array (read only).
//********************************************************************************

LIBMATHS_INLINE const F32 Vector4::operator[] (const S32 i) const
{
	LIBMATHS_ASSERT(i >= 0 && i < 4);
	return reinterpret_cast<const F32*>(&x)[i];
}


//********************************************************************************
// Access vector components as an array (read-write).
//********************************************************************************

LIBMATHS_INLINE F32& Vector4::operator[] (const S32 i)
{
	LIBMATHS_ASSERT(i >= 0 && i < 4);
	return reinterpret_cast<F32*>(&x)[i];
}


//********************************************************************************
// Silent conversion to a 3D vector discarding the w component.
//********************************************************************************

LIBMATHS_INLINE Vector4::operator Vector3x(void) const
{
	return Vector3x(x, y, z);
}

#pragma managed(pop)
