//********************************************************************************
//
// Axis aligned 3D bounding boxes for collision checks.
//
// These can be tested for collisions with other boxes, points, spheres or rays
// through the game world.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "box.h"


namespace libmaths
{

	//********************************************************************************
	// Check if the line intersects this bounding box.
	// Does not allow intersections from within the bounding box.
	//********************************************************************************

	Bool AxisAlignedBox::Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const
	{
		F32 tmin, tmax, tymin, tymax, tzmin, tzmax;

		Vector3x invDirection(1.0f / ray.Direction().X(), 1.0f / ray.Direction().Y(), 1.0f / ray.Direction().Z());
		Vector3x topLeft(Minimum());
		Vector3x bottomRight(Maximum());

		if (invDirection.X() >= 0)
		{
			tmin = (topLeft.X() - ray.Origin().X()) * invDirection.X();
			tmax = (bottomRight.X() - ray.Origin().X()) * invDirection.X();
		}
		else
		{
			tmin = (bottomRight.X() - ray.Origin().X()) * invDirection.X();
			tmax = (topLeft.X() - ray.Origin().X()) * invDirection.X();
		}

		if (invDirection.Y() >= 0)
		{
			tymin = (topLeft.Y() - ray.Origin().Y()) * invDirection.Y();
			tymax = (bottomRight.Y() - ray.Origin().Y()) * invDirection.Y();
		}
		else
		{
			tymin = (bottomRight.Y() - ray.Origin().Y()) * invDirection.Y();
			tymax = (topLeft.Y() - ray.Origin().Y()) * invDirection.Y();
		}

		if ((tmin > tymax) || (tymin > tmax))
			return false;

		if (tymin > tmin)
			tmin = tymin;

		if (tymax < tmax)
			tmax = tymax;

		if (invDirection.Z() >= 0)
		{
			tzmin = (topLeft.Z() - ray.Origin().Z()) * invDirection.Z();
			tzmax = (bottomRight.Z() - ray.Origin().Z()) * invDirection.Z();
		}
		else
		{
			tzmin = (bottomRight.Z() - ray.Origin().Z()) * invDirection.Z();
			tzmax = (topLeft.Z() - ray.Origin().Z()) * invDirection.Z();
		}

		if ((tmin > tzmax) || (tzmin > tmax))
			return false;

		if (tzmin > tmin)
			tmin = tzmin;

		if (tzmax < tmax)
			tmax = tzmax;

		if (tmin <= 0.0f)
			return false;

		point = ray.Origin() + ray.Direction() * tmin;

		// TODO: This is not the surface normal of the side of the box
		// that has been intersected, which is what we ought to really
		// be returning to the caller.

		normal = (point - ray.Origin()).Normalised();

		return true;
	}


	//********************************************************************************
	// Check if the line intersects this bounding box.
	// Does not allow intersections from within the bounding box.
	//********************************************************************************

	Bool AxisAlignedBox::Intersects(const LineSegment3& line) const
	{
		if (Contains(line.Origin()))
		{
			return false;
		}

		Vector3x w(line.Direction() * 0.5f);
		Vector3x c(line.Origin() + w);
		Vector3x m(c-_centre);

		F32 x(Abs(w.X())), y(Abs(w.Y())), z(Abs(w.Z()));

		if (Abs(m.X()) > x + _extent.X())
			return false;

		if (Abs(m.Y()) > y + _extent.Y())
			return false;

		if (Abs(m.Z()) > z + _extent.Z())
			return false;

		if (Abs(m.Y() * w.Z() - m.Z() * w.Y()) > _extent.Y() * z + _extent.Z() * y)
			return false;

		if (Abs(m.X() * w.Z() - m.Z() * w.X()) > _extent.X() * z + _extent.Z() * x)
			return false;

		if (Abs(m.X() * w.Y() - m.Y() * w.X()) > _extent.X() * y + _extent.Y() * x)
			return false;

		return true;
	}


	//********************************************************************************
	// Transform the bounding box with the given matrix.
	//********************************************************************************

	AxisAlignedBox AxisAlignedBox::Transform(const Matrix44& matrix) const
	{
		S32 i, j;
		F32 a, b;

		Vector3x Bmin(0, 0, 0);
		Vector3x Bmax(0, 0, 0);
		Vector3x Amin(Minimum()), Amax(Maximum());

		for (i = 0; i < 3; i++)
		{
			for (j = 0; j < 3; j++)
			{
				a = Amin[j] * matrix(j,i);
				b = Amax[j] * matrix(j,i);

				Bmin[i] += Min(a,b);
				Bmax[i] += Max(a,b);
			}
		}

		AxisAlignedBox box;

		box.MinMax(Vector3x(Bmin) + matrix.Translation(), Vector3x(Bmax) + matrix.Translation());

		return box;
	}


// namespace libmaths
}
