//********************************************************************************
//
// 3D bounding capsules for collision checks.
//
// Bounding capsules can detect overlaps with other capsules, points and
// bounding spheres.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "box.h"
#include "capsule.h"


namespace libmaths
{

	//********************************************************************************
	// Get the distance of the capsule from a point in space.
	//********************************************************************************

	F32 Capsule::Distance(const Vector3x& point, F32& pointOnLine) const
	{
		LineSegment3 line(_endPoints[0], _endPoints[1] - _endPoints[0]);

		F32 distance = line.Distance(point, pointOnLine) - Radius();

		return Max(distance, 0.0f);
	}


	//********************************************************************************
	// Get the squared distance of the capsule from a point in space.
	//********************************************************************************

	F32 Capsule::DistanceSqr(const Vector3x& point, F32& pointOnLine) const
	{
		F32 distance = Distance(point, pointOnLine);

		return distance * distance;
	}


	//********************************************************************************
	// Check if the capsule contains a point in space.
	//********************************************************************************

	Bool Capsule::Contains(const Vector3x& point) const
	{
		F32 pointOnLine;

		return (Distance(point, pointOnLine) == 0.0f);
	}


	//********************************************************************************
	// Does the capsule intersect a capsule?
	//********************************************************************************

	Bool Capsule::Intersects(const Capsule& capsule) const
	{
		Vector3x ip;

		return Intersects(capsule, ip, ip);
	}


	//********************************************************************************
	// Does the capsule intersect a capsule?
	//********************************************************************************

	Bool Capsule::Intersects(const Capsule& capsule, Vector3x& ip0, Vector3x& ip1) const
	{
		F32 t0, t1;

		LineSegment3 line(_endPoints[0], _endPoints[1] - _endPoints[0]);
		LineSegment3 otherLine(capsule._endPoints[0], capsule._endPoints[1] - capsule._endPoints[0]);

		F32 radius = Radius() + capsule.Radius();
		F32 distanceSqr = line.DistanceSqr(otherLine, t0, t1);

		if (distanceSqr < radius * radius)
		{
			ip0 = line.Origin() + (line.Direction() * t0);
			ip1 = otherLine.Origin() + (otherLine.Direction() * t1);

			return true;
		}

		return false;
	}


	//********************************************************************************
	// Check if the ray intersects this capsule.
	// Does not allow intersections from within the capsule.
	//********************************************************************************

	Bool Capsule::Intersects(const Ray3& ray) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}


	//********************************************************************************
	// Check if the ray intersects this capsule.
	// Does not allow intersections from within the capsule.
	//********************************************************************************

	Bool Capsule::Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}

// namespace libmaths
}
