//********************************************************************************
//
// 3D bounding cylinders for collision checks.
//
// Bounding cylinders can detect overlaps with other cylinders, points and
// bounding spheres.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "cylinder.h"
#include "plane.h"


namespace libmaths
{

	//********************************************************************************
	// Get the distance of the cylinder from a point in space.
	//********************************************************************************

	F32 Cylinder::Distance(const Vector3x& point, F32& pointOnLine) const
	{
		LineSegment3 line(_endPoints[0], _endPoints[1] - _endPoints[0]);
		F32 distance = line.Distance(point, pointOnLine);
		S32 endPoint = -1;

		if (pointOnLine < 0.0f)
		{
			endPoint = 0;
		}
		else if (pointOnLine > 1.0f)
		{
			endPoint = 1;
		}
		else
		{
			return Max(distance - Radius(), 0.0f);
		}

		Plane plane(_endPoints[endPoint], (_endPoints[endPoint] - _endPoints[1 - endPoint]).Normalised());
		Vector3x pointOnPlane = plane.ClosestPoint(point);
		Vector3x difference = pointOnPlane - _endPoints[endPoint];
		F32 lengthSqr = difference.LengthSqr();

		if (lengthSqr > RadiusSqr())
		{
			pointOnPlane = _endPoints[endPoint] + (difference * (Radius() / Sqrt(lengthSqr)));
		}

		return (point - pointOnPlane).Length();
	}


	//********************************************************************************
	// Get the squared distance of the cylinder from a point in space.
	//********************************************************************************

	F32 Cylinder::DistanceSqr(const Vector3x& point, F32& pointOnLine) const
	{
		F32 distance = Distance(point, pointOnLine);

		return distance * distance;
	}


	//********************************************************************************
	// Check if the cylinder contains a point in space.
	//********************************************************************************

	Bool Cylinder::Contains(const Vector3x& point) const
	{
		F32 pointOnLine;

		return (Distance(point, pointOnLine) == 0.0f);
	}


	//********************************************************************************
	// Does the cylinder intersect a cylinder?
	//********************************************************************************

	Bool Cylinder::Intersects(const Cylinder& cylinder) const
	{
		Vector3x ip;

		return Intersects(cylinder, ip, ip);
	}


	//********************************************************************************
	// Does the cylinder intersect a cylinder?
	//********************************************************************************

	Bool Cylinder::Intersects(const Cylinder& cylinder, Vector3x& ip0, Vector3x& ip1) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}


	//********************************************************************************
	// Check if the ray intersects this cylinder.
	// Does not allow intersections from within the cylinder.
	//********************************************************************************

	Bool Cylinder::Intersects(const Ray3& ray) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}


	//********************************************************************************
	// Check if the ray intersects this cylinder.
	// Does not allow intersections from within the cylinder.
	//********************************************************************************

	Bool Cylinder::Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}

// namespace libmaths
}
