//********************************************************************************
//
// A class used to define line segments in 2D space, often used for detecting
// intersections with other geometric types such as circles and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "linesegment2.h"
#include "rectangle.h"
#include "circle.h"
#include "triangle2.h"


namespace libmaths
{

	//********************************************************************************
	// Get the distance of the line from a point in space. pointOnLine will return the
	// nearest point on the line in parametric form; 0 .. 1 encompassing the line
	// length, less than zero before the start of the line and greater than one if
	// over the end of the line.
	//********************************************************************************

	F32 LineSegment2::Distance(const Vector2& point, F32& pointOnLine, Bool clamp) const
	{
		return Sqrt(DistanceSqr(point, pointOnLine, clamp));
	}


	//********************************************************************************
	// Get the squared distance of the line from a point in space. pointOnLine will
	// return the nearest point on the line in parametric form; 0 .. 1 encompassing
	// the line length, less than zero before the start of the line and greater than
	// one if over the end of the line.
	//********************************************************************************

	F32 LineSegment2::DistanceSqr(const Vector2& point, F32& pointOnLine, Bool clamp) const
	{
		Vector2 difference = point - _origin;

		// Calculate the scalar for the nearest point on the line to the point that we
		// are comparing.

		Vector2 lineVector = _direction;
		F32 lengthSqr = lineVector.LengthSqr();

		if (lengthSqr < Epsilon)
		{
			pointOnLine = 0.0f;
		}
		else
		{
			pointOnLine = (lineVector * difference) / lengthSqr;

			if (pointOnLine > 0.0f)
			{
				if (pointOnLine < 1.0f)
				{
					// Get the nearest point on the line to the point that we are comparing and
					// return the squared distance between them.

					lineVector *= pointOnLine;
				}

				difference -= lineVector;
			}

			if (clamp)
			{
				pointOnLine = Clamp(pointOnLine, 0.0f, 1.0f);
			}
		}

		return difference.LengthSqr();
	}


	//********************************************************************************
	// Get the distance of the line from a line in space.
	//********************************************************************************

	F32 LineSegment2::Distance(const LineSegment2& line, F32& pointOnLine, F32& pointOnOtherLine) const
	{
		return Sqrt(DistanceSqr(line, pointOnLine, pointOnOtherLine));
	}


	//********************************************************************************
	// Get the squared distance of the line from a line in space.
	// Largely taken from Real Time Collision Detection by Christer Ericson.
	//********************************************************************************

	F32 LineSegment2::DistanceSqr(const LineSegment2& line, F32& pointOnLine, F32& pointOnOtherLine) const
	{
		const Vector2& p1 = Point(0);
		const Vector2& q1 = Point(1);
		const Vector2& p2 = line.Point(0);
		const Vector2& q2 = line.Point(1);

		Vector2 cp0, cp1;

		Vector2 d1 = q1 - p1;
		Vector2 d2 = q2 - p2;
		Vector2 r = p1 - p2;

		F32 a = d1 * d1;
		F32 e = d2 * d2;
		F32 f = d2 * r;

		// Check if either or both segments degenerate into points.

		if (a <= Epsilon && e <= Epsilon)
		{
			// Both segments degenerate into points.

			pointOnLine = pointOnOtherLine = 0.0f;

			cp0 = p1;
			cp1 = p2;
		}
		else
		{
			if (a <= Epsilon)
			{
				// First segment degenerates into a point.

				pointOnLine = 0.0f;
				pointOnOtherLine = f / e;
				pointOnOtherLine = Clamp(pointOnOtherLine, 0.0f, 1.0f);
			}
			else
			{
				F32 c = d1 * r;

				if (e <= Epsilon)
				{
					// Second segment degenerates into a point.

					pointOnOtherLine = 0.0f;
					pointOnLine = Clamp(-c / a, 0.0f, 1.0f);
				}
				else
				{
					// The general non-degenerate case starts here.

					F32 b = d1 * d2;
					F32 denom = a * e - b * b;

					// If segments not parallel, compute closest point on L1 to L2, and
					// clamp to segment S1. Otherwise pick arbitrary pointOnLine (here 0)

					if (denom != 0.0f)
					{
						pointOnLine = Clamp((b * f - c * e) / denom, 0.0f, 1.0f);
					}
					else
					{
						pointOnLine = 0.0f;
					}

					// Compute point on L2 closest to S1(pointOnLine).

					pointOnOtherLine = (b*pointOnLine + f) / e;

					if (pointOnOtherLine < 0.0f)
					{
						pointOnOtherLine = 0.0f;
						pointOnLine = Clamp(-c / a, 0.0f, 1.0f);
					}
					else if (pointOnOtherLine > 1.0f)
					{
						pointOnOtherLine = 1.0f;
						pointOnLine = Clamp((b - c) / a, 0.0f, 1.0f);
					}
				}
			}

			cp0 = p1 + d1 * pointOnLine;
			cp1 = p2 + d2 * pointOnOtherLine;
		}

		return (cp0 - cp1).LengthSqr();
	}


	//********************************************************************************
	// Find the point on the line closest to the given point.
	//********************************************************************************

	Vector2 LineSegment2::ClosestPoint(Vector2 point) const
	{
		F32 pointOnLine;

		DistanceSqr(point, pointOnLine);

		return _origin + (_direction * pointOnLine);
	}


	//********************************************************************************
	// Find the point on the line closest to the given line.
	//********************************************************************************

	Vector2 LineSegment2::ClosestPoint(const LineSegment2& line) const
	{
		F32 pointOnLine;
		F32 pointOnOtherLine;

		DistanceSqr(line, pointOnLine, pointOnOtherLine);

		return _origin + (_direction * pointOnLine);
	}


	//********************************************************************************
	// Does the line intersect a rectangle?
	//********************************************************************************

	Bool LineSegment2::Intersects(AxisAlignedRectangle& rectangle) const
	{
		return rectangle.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a rectangle?
	//********************************************************************************

	Bool LineSegment2::Intersects(AxisAlignedRectangle& rectangle, Vector2& point, Vector2& normal, F32& t) const
	{
		return rectangle.Intersects(*this, point, normal, t);
	}


	//********************************************************************************
	// Does the line intersect a circle?
	//********************************************************************************

	Bool LineSegment2::Intersects(Circle& circle) const
	{
		return circle.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a circle?
	//********************************************************************************

	Bool LineSegment2::Intersects(Circle& circle, Vector2& point, Vector2& normal, F32& t) const
	{
		return circle.Intersects(*this, point, normal, t);
	}


	//********************************************************************************
	// Does the line intersect a triangle?
	//********************************************************************************

	Bool LineSegment2::Intersects(Triangle2& triangle) const
	{
		return triangle.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a triangle?
	//********************************************************************************

	Bool LineSegment2::Intersects(Triangle2& triangle, Vector2& point, Vector2& normal, F32& t) const
	{
		return triangle.Intersects(*this, point, normal, t);
	}

// namespace libmaths
}
