//********************************************************************************
//
// A class used to define line segments in 3D space, often used for detecting
// intersections with other geometric types such as spheres and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "linesegment3.h"
#include "capsule.h"
#include "sphere.h"
#include "triangle3.h"
#include "plane.h"
#include "box.h"


namespace libmaths
{

	//********************************************************************************
	// Get the distance of the line from a point in space. pointOnLine will return the
	// nearest point on the line in parametric form; 0 .. 1 encompassing the line
	// length, less than zero before the start of the line and greater than one if
	// over the end of the line.
	//********************************************************************************

	F32 LineSegment3::Distance(const Vector3x& point, F32& pointOnLine, Bool clamp) const
	{
		return Sqrt(DistanceSqr(point, pointOnLine, clamp));
	}


	//********************************************************************************
	// Get the squared distance of the line from a point in space. pointOnLine will
	// return the nearest point on the line in parametric form; 0 .. 1 encompassing
	// the line length, less than zero before the start of the line and greater than
	// one if over the end of the line.
	//********************************************************************************

	F32 LineSegment3::DistanceSqr(const Vector3x& point, F32& pointOnLine, Bool clamp) const
	{
		Vector3x difference = point - _origin;

		// Calculate the scalar for the nearest point on the line to the point that we
		// are comparing.

		F32 lengthSqr = _direction.LengthSqr();

		if (lengthSqr < Epsilon)
		{
			pointOnLine = 0.0f;
		}
		else
		{
			pointOnLine = (_direction * difference) / lengthSqr;

			if (pointOnLine > 0.0f)
			{
				Vector3x lineVector = _direction;

				if (pointOnLine < 1.0f)
				{
					// Get the nearest point on the line to the point that we are comparing and
					// return the squared distance between them.

					lineVector *= pointOnLine;
				}

				difference -= lineVector;
			}

			if (clamp)
			{
				pointOnLine = Clamp(pointOnLine, 0.0f, 1.0f);
			}
		}

		return difference.LengthSqr();
	}


	//********************************************************************************
	// Get the distance of the line from a line in space.
	//********************************************************************************

	F32 LineSegment3::Distance(const LineSegment3& line, F32& pointOnLine, F32& pointOnOtherLine) const
	{
		return Sqrt(DistanceSqr(line, pointOnLine, pointOnOtherLine));
	}


	//********************************************************************************
	// Get the squared distance of the line from a line in space.
	// Largely taken from Real Time Collision Detection by Christer Ericson.
	//********************************************************************************

	F32 LineSegment3::DistanceSqr(const LineSegment3& line, F32& pointOnLine, F32& pointOnOtherLine) const
	{
		const Vector3x& p1 = Point(0);
		const Vector3x& q1 = Point(1);
		const Vector3x& p2 = line.Point(0);
		const Vector3x& q2 = line.Point(1);

		Vector3x cp0, cp1;

		Vector3x d1 = q1 - p1;
		Vector3x d2 = q2 - p2;
		Vector3x r = p1 - p2;

		F32 a = d1 * d1;
		F32 e = d2 * d2;
		F32 f = d2 * r;

		// Check if either or both segments degenerate into points.

		if (a <= Epsilon && e <= Epsilon)
		{
			// Both segments degenerate into points.

			pointOnLine = pointOnOtherLine = 0.0f;

			cp0 = p1;
			cp1 = p2;
		}
		else
		{
			if (a <= Epsilon)
			{
				// First segment degenerates into a point.

				pointOnLine = 0.0f;
				pointOnOtherLine = f / e;
				pointOnOtherLine = Clamp(pointOnOtherLine, 0.0f, 1.0f);
			}
			else
			{
				F32 c = d1 * r;

				if (e <= Epsilon)
				{
					// Second segment degenerates into a point.

					pointOnOtherLine = 0.0f;
					pointOnLine = Clamp(-c / a, 0.0f, 1.0f);
				}
				else
				{
					// The general non-degenerate case starts here.

					F32 b = d1 * d2;
					F32 denom = a * e - b * b;

					// If segments not parallel, compute closest point on L1 to L2, and
					// clamp to segment S1. Otherwise pick arbitrary pointOnLine (here 0)

					if (denom != 0.0f)
					{
						pointOnLine = Clamp((b * f - c * e) / denom, 0.0f, 1.0f);
					}
					else
					{
						pointOnLine = 0.0f;
					}

					// Compute point on L2 closest to S1(pointOnLine).

					pointOnOtherLine = (b * pointOnLine + f) / e;

					if (pointOnOtherLine < 0.0f)
					{
						pointOnOtherLine = 0.0f;
						pointOnLine = Clamp(-c / a, 0.0f, 1.0f);
					}
					else if (pointOnOtherLine > 1.0f)
					{
						pointOnOtherLine = 1.0f;
						pointOnLine = Clamp((b - c) / a, 0.0f, 1.0f);
					}
				}
			}

			cp0 = p1 + (d1 * pointOnLine);
			cp1 = p2 + (d2 * pointOnOtherLine);
		}

		return (cp0 - cp1).LengthSqr();
	}


	//********************************************************************************
	// Find the point on the line closest to the given point.
	//********************************************************************************

	Vector3x LineSegment3::ClosestPoint(Vector3x point) const
	{
		F32 pointOnLine;

		DistanceSqr(point, pointOnLine);

		return _origin + (_direction * pointOnLine);
	}


	//********************************************************************************
	// Find the point on the line closest to the given line.
	//********************************************************************************

	Vector3x LineSegment3::ClosestPoint(const LineSegment3& line) const
	{
		F32 pointOnLine;
		F32 pointOnOtherLine;

		DistanceSqr(line, pointOnLine, pointOnOtherLine);

		return _origin + (_direction * pointOnLine);
	}


	//********************************************************************************
	// Does the line intersect a box?
	//********************************************************************************

	Bool LineSegment3::Intersects(const AxisAlignedBox& box) const
	{
		return box.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a box?
	//********************************************************************************

	Bool LineSegment3::Intersects(const AxisAlignedBox& box, Vector3x& point, Vector3x& normal, F32& t) const
	{
		return box.Intersects(*this, point, normal, t);
	}


	//********************************************************************************
	// Does the line intersect a capsule?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Capsule& capsule) const
	{
		return capsule.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a capsule?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Capsule& capsule, Vector3x& point, Vector3x& normal, F32& t) const
	{
		return capsule.Intersects(*this, point, normal, t);
	}


	//********************************************************************************
	// Does the line intersect a sphere?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Sphere& sphere) const
	{
		return sphere.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a sphere?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Sphere& sphere, Vector3x& point, Vector3x& normal, F32& t) const
	{
		return sphere.Intersects(*this, point, normal, t);
	}


	//********************************************************************************
	// Does the line intersect a triangle?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Triangle3& triangle, Bool twoSided) const
	{
		return triangle.Intersects(*this, twoSided);
	}


	//********************************************************************************
	// Does the line intersect a triangle?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Triangle3& triangle, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided) const
	{
		return triangle.Intersects(*this, point, normal, t, twoSided);
	}


	//********************************************************************************
	// Does the line intersect a plane?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Plane& plane, Bool twoSided) const
	{
		return plane.Intersects(*this);
	}


	//********************************************************************************
	// Does the line intersect a plane?
	//********************************************************************************

	Bool LineSegment3::Intersects(const Plane& plane, Vector3x& point, Vector3x& normal, F32& t, Bool twoSided) const
	{
		return plane.Intersects(*this, point, normal, t, twoSided);
	}

// namespace libmaths
}