//********************************************************************************
//
// 2x2 matrix class. Representing the top 2x2 section of what would normally be a
// 3x3 matrix. Used to transform 2 component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "matrix22of33.h"
#include "matrix33.h"


namespace libmaths
{

	//********************************************************************************
	// Construct from a 3x3 matrix.
	// The third column and translation are discarded in the process.
	//********************************************************************************

	Matrix22of33::Matrix22of33(const Matrix33& matrix)
	{
		Element(0, 0) = matrix(0, 0);
		Element(0, 1) = matrix(0, 1);

		Element(1, 0) = matrix(1, 0);
		Element(1, 1) = matrix(1, 1);
	}


	//********************************************************************************
	// Build a transformation from a scale value.
	//********************************************************************************

	Matrix22of33& Matrix22of33::FromScale(F32 scale)
	{
		Element(0, 0) = scale;
		Element(0, 1) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = scale;

		return *this;
	}


	//********************************************************************************
	// Build a transformation from separate x and y scale values.
	//********************************************************************************

	Matrix22of33& Matrix22of33::FromScale(F32 x, F32 y)
	{
		Element(0, 0) = x;
		Element(0, 1) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = y;

		return *this;
	}


	//********************************************************************************
	// Scale this matrix uniformly.
	//********************************************************************************

	Matrix22of33& Matrix22of33::Scale(F32 scale)
	{
		for (S32 i = 0; i < 2; i++)
		{
			Element(0, i) *= scale;
			Element(1, i) *= scale;
		}

		return *this;
	}


	//********************************************************************************
	// Scale this matrix non-uniformly.
	//********************************************************************************

	Matrix22of33& Matrix22of33::Scale(F32 x, F32 y)
	{
		for (S32 i = 0; i < 2; i++)
		{
			Element(0, i) *= x;
			Element(1, i) *= y;
		}

		return *this;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation.
	//********************************************************************************

	Matrix22of33& Matrix22of33::Invert(void)
	{
		*this = Inverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation.
	//********************************************************************************

	Matrix22of33 Matrix22of33::Inverse(void) const
	{
		Matrix22of33 result;
		F32 determinant = Determinant();

		if (Abs(determinant) >= Epsilon)
		{
			F32 invDeterminant = 1.0f / determinant;

			result.Element(0, 0) = Element(1, 1) * invDeterminant;
			result.Element(0, 1) = Element(0, 1) * invDeterminant;
			result.Element(1, 0) = Element(1, 0) * invDeterminant;
			result.Element(1, 1) = Element(0, 0) * invDeterminant;
		}
		else
		{
			LIBMATHS_ASSERT(0 && "Singular matrix could not be inverted");
		}

		return result;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation (assumes this matrix
	// is an orthonormal transformation).
	//********************************************************************************

	Matrix22of33& Matrix22of33::OrthonormalInvert(void)
	{
		*this = OrthonormalInverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation (assumes this matrix is an orthonormal transformation).
	//********************************************************************************

	Matrix22of33 Matrix22of33::OrthonormalInverse(void) const
	{
		Matrix22of33 result;
		F32 scale = Element(0, 0) * Element(0, 0) + Element(0, 1) * Element(0, 1);

		scale = 1.0f / scale;

		result.Element(0, 0) = Element(0, 0) * scale;
		result.Element(0, 1) = Element(1, 0) * scale;
		result.Element(1, 0) = Element(0, 1) * scale;
		result.Element(1, 1) = Element(1, 1) * scale;

		return result;
	}


	//********************************************************************************
	// Orthonormalise this matrix using the Gram-Schmidt algorithm.
	// M = [x|y], then orthonormal output matrix is Q = [nx|ny],
	//
	// nx = x/|x|
	// ny = (y-nx(nx*y))/|y-nx(nx*y)|
	//
	// Where |V| indicates length of vector V and A*B indicates dot product of vectors
	// A and B.
	//********************************************************************************

	Matrix22of33& Matrix22of33::Orthonormalise(void)
	{
		Vector2 nx(XAxis().Normalised());
		Vector2 y(YAxis());
		Vector2 ny((y - nx * (nx * y)).Normalised());

		XAxis(nx);
		YAxis(ny);

		return *this;
	}


	//********************************************************************************
	// Return the determinant of this 2x2 matrix.
	//********************************************************************************

	F32 Matrix22of33::Determinant(void) const
	{
		return Element(0, 0) * Element(1, 1) - Element(0, 1) * Element(1, 0);
	}


	//********************************************************************************
	// Build this matrix as an interpolation between the two given matrices,
	// with possible extra 360 degree spins.
	//********************************************************************************

	Matrix22of33& Matrix22of33::Lerp(F32 t, Matrix22of33& m0, Matrix22of33& m1, U32 numSpins)
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented.");

		return *this;
	}

// namespace libmaths
}
