//********************************************************************************
//
// 3x3 matrix class. Used for homogenous 2D transformations with 2 and 3
// component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "matrix22of33.h"
#include "matrix33.h"
#include "matrix44.h"


namespace libmaths
{

	//********************************************************************************
	// Construct from a 2x2 matrix.
	// The third column is initialised to (0, 0, 1)
	// The translation is initialised to (0, 0)
	//********************************************************************************

	Matrix33::Matrix33(const Matrix22of33& matrix)
	{
		Element(0, 0) = matrix(0, 0);
		Element(0, 1) = matrix(0, 1);
		Element(0, 2) = 0.0f;

		Element(1, 0) = matrix(1, 0);
		Element(1, 1) = matrix(1, 1);
		Element(1, 2) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = 1.0f;
	}


	//********************************************************************************
	// Build a transformation from a scale value.
	//********************************************************************************

	Matrix33& Matrix33::FromScale(F32 scale)
	{
		Element(0, 0) = scale;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = scale;
		Element(1, 2) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = 1.0f;

		return *this;
	}


	//********************************************************************************
	// Build a transformation from separate x, y and z scale values.
	//********************************************************************************

	Matrix33& Matrix33::FromScale(F32 x, F32 y, F32 z)
	{
		Element(0, 0) = x;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = y;
		Element(1, 2) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = z;

		return *this;
	}


	//********************************************************************************
	// Scale this matrix uniformly.
	//********************************************************************************

	Matrix33& Matrix33::Scale(F32 scale)
	{
		for (S32 i = 0; i < 2; i++)
		{
			Element(0, i) *= scale;
			Element(1, i) *= scale;
			Element(2, i) *= scale;
		}

		return *this;
	}


	//********************************************************************************
	// Scale this matrix non-uniformly.
	//********************************************************************************

	Matrix33& Matrix33::Scale(F32 x, F32 y, F32 z)
	{
		for (S32 i = 0; i < 2; i++)
		{
			Element(0, i) *= x;
			Element(1, i) *= y;
			Element(2, i) *= z;
		}

		return *this;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation.
	//********************************************************************************

	Matrix33& Matrix33::Invert(void)
	{
		*this = Inverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation.
	//********************************************************************************

	Matrix33 Matrix33::Inverse(void) const
	{
		Matrix33 result(*this);

		F32 cofactor0 = Element(1, 1) * Element(2, 2) - Element(1, 2) * Element(2, 1);
		F32 cofactor3 = Element(0, 2) * Element(2, 1) - Element(0, 1) * Element(2, 2);
		F32 cofactor6 = Element(0, 1) * Element(1, 2) - Element(0, 2) * Element(1, 1);

		F32 determinant = Element(0, 0) * cofactor0 + Element(1, 0) * cofactor3 + Element(2, 0) * cofactor6;

		if (Abs(determinant) >= Epsilon)
		{
			F32 invDeterminant = 1.0f / determinant;

			result(0, 0) = invDeterminant * cofactor0;
			result(0, 1) = invDeterminant * cofactor3;
			result(0, 2) = invDeterminant * cofactor6;

			result(1, 0) = invDeterminant * (Element(1, 2) * Element(2, 0) - Element(1, 0) * Element(2, 2));
			result(1, 1) = invDeterminant * (Element(0, 0) * Element(2, 2) - Element(0, 2) * Element(2, 0));
			result(1, 2) = invDeterminant * (Element(0, 2) * Element(1, 0) - Element(0, 0) * Element(1, 2));

			result(2, 0) = invDeterminant * (Element(1, 0) * Element(2, 1) - Element(1, 1) * Element(2, 0));
			result(2, 1) = invDeterminant * (Element(0, 1) * Element(2, 0) - Element(0, 0) * Element(2, 1));
			result(2, 2) = invDeterminant * (Element(0, 0) * Element(1, 1) - Element(0, 1) * Element(1, 0));
		}
		else
		{
			LIBMATHS_ASSERT(0 && "Singular matrix could not be inverted");
		}

		return result;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation (assumes this matrix
	// is an orthonormal transformation).
	//********************************************************************************

	Matrix33& Matrix33::OrthonormalInvert(void)
	{
		*this = OrthonormalInverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation (assumes this matrix is an orthonormal transformation).
	//********************************************************************************

	Matrix33 Matrix33::OrthonormalInverse(void) const
	{
		F32 scale = Element(0, 0) * Element(0, 0) + Element(0, 1) * Element(0, 1) + Element(0, 2) * Element(0, 2);
		Matrix33 result;

		scale = 1.0f / scale;

		result(0, 0) = Element(0, 0) * scale;
		result(0, 1) = Element(1, 0) * scale;
		result(0, 2) = Element(2, 0) * scale;
		result(1, 0) = Element(0, 1) * scale;
		result(1, 1) = Element(1, 1) * scale;
		result(1, 2) = Element(2, 1) * scale;
		result(2, 0) = Element(0, 2) * scale;
		result(2, 1) = Element(1, 2) * scale;
		result(2, 2) = Element(2, 2) * scale;

		return result;
	}


	//********************************************************************************
	// Orthonormalise this matrix using the Gram-Schmidt algorithm.
	// NOTE: Only the upper 3x3 sub-matrix is affected.
	//
	// M = [x|y|z], then orthonormal output matrix is Q = [nx|ny|nz],
	//
	// nx = x/|x|
	// ny = (y-nx(nx*y))/|y-nx(nx*y)|
	//
	// Where |V| indicates length of vector V and A*B indicates dot product of vectors
	// A and B.
	//********************************************************************************

	Matrix33& Matrix33::Orthonormalise(void)
	{
		Vector2 nx(XAxis().Normalised());
		Vector2 y(YAxis());
		Vector2 ny((y - nx * (nx * y)).Normalised());

		XAxis(nx);
		YAxis(ny);

		return *this;
	}


	//********************************************************************************
	// Get the determinant of this matrix.
	//********************************************************************************

	F32 Matrix33::Determinant(void) const
	{
		return 	Element(0, 0) * (Element(1, 1) * Element(2, 2) - Element(1, 2) * Element(2, 1)) +
				Element(1, 0) * (Element(0, 2) * Element(2, 1) - Element(0, 1) * Element(2, 2)) +
				Element(2, 0) * (Element(0, 1) * Element(1, 2) - Element(0, 2) * Element(1, 1));
	}


	//********************************************************************************
	// Build this matrix as an interpolation between the two given matrices,
	// with possible extra 360 degree spins.
	//********************************************************************************

	Matrix33& Matrix33::Lerp(F32 t, Matrix33& m0, Matrix33& m1, U32 numSpins)
	{
		// TODO: Implement this.

		return *this;
	}

// namespace libmaths
}
