//********************************************************************************
//
// 3x3 matrix class. Representing the top 3x3 section of what would normally be a
// 4x4 matrix. Used to transform 3 component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "matrix33of44.h"
#include "matrix44.h"


namespace libmaths
{

	//********************************************************************************
	// Construct from a 4x4 matrix.
	// The fourth column and translation are discarded in the process.
	//********************************************************************************

	Matrix33of44::Matrix33of44(const Matrix44& matrix)
	{
		Element(0, 0) = matrix(0, 0);
		Element(0, 1) = matrix(0, 1);
		Element(0, 2) = matrix(0, 2);

		Element(1, 0) = matrix(1, 0);
		Element(1, 1) = matrix(1, 1);
		Element(1, 2) = matrix(1, 2);

		Element(2, 0) = matrix(2, 0);
		Element(2, 1) = matrix(2, 1);
		Element(2, 2) = matrix(2, 2);
	}


	//********************************************************************************
	// Build a transformation from a quaternion.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromQuaternion(const Quaternion& quaternion)
	{
		Matrix44 temp;

		return (*this = temp.FromQuaternion(quaternion));
	}


	//********************************************************************************
	// Convert this matrix into a quaternion representing the same rotation.
	//********************************************************************************

	Quaternion Matrix33of44::ToQuaternion(void) const
	{
		return Quaternion(*this);
	}


	//********************************************************************************
	// Build a transformation representing a rotation of angle radians anti-clockwise
	// about the axis defined by axis. Angles of zero will produce no transform.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromAxisAngle(const Vector3x& axis, F32 angle)
	{
		Matrix44 temp;

		return (*this = temp.FromAxisAngle(axis, angle));
	}


	//********************************************************************************
	// Decompose the rotation part of this matrix into a rotation of radians
	// anti-clockwise about an axis.
	//********************************************************************************

	void Matrix33of44::ToAxisAngle(Vector3x& axis, F32& angle) const
	{
		Matrix44 temp(*this);

		temp.ToAxisAngle(axis, angle);
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromDirectionRH(const Vector3x& direction, const Vector3x& up)
	{
		Matrix44 temp;

		return (*this = temp.FromDirectionRH(direction, up));
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromDirectionLH(const Vector3x& direction, const Vector3x& up)
	{
		Matrix44 temp;

		return (*this = temp.FromDirectionLH(direction, up));
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the x axis.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromXAngle(F32 angle)
	{
		Matrix44 temp;

		return (*this = temp.FromXAngle(angle));
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the y axis.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromYAngle(F32 angle)
	{
		Matrix44 temp;

		return (*this = temp.FromYAngle(angle));
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the z axis.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromZAngle(F32 angle)
	{
		Matrix44 temp;

		return (*this = temp.FromZAngle(angle));
	}


	//********************************************************************************
	// Rotate this matrix about X then Y then Z by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromXYZAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromXYZAngles(angles));
	}


	//********************************************************************************
	// Rotate this matrix about X then Z then Y by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromXZYAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromXZYAngles(angles));
	}


	//********************************************************************************
	// Rotate this matrix about Y then X then Z by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromYXZAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromYXZAngles(angles));
	}


	//********************************************************************************
	// Rotate this matrix about Y then Z then X by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromYZXAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromYZXAngles(angles));
	}


	//********************************************************************************
	// Rotate this matrix about Z then X then Y by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromZXYAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromZXYAngles(angles));
	}

	//********************************************************************************
	// Rotate this matrix about Z then Y then X by the given anti-clockwise angles.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromZYXAngles(const Vector3x& angles)
	{
		Matrix44 temp;

		return (*this = temp.FromZYXAngles(angles));
	}


	//********************************************************************************
	// Build a transformation from a scale value.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromScale(F32 scale)
	{
		Element(0, 0) = scale;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = scale;
		Element(1, 2) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = scale;

		return *this;
	}


	//********************************************************************************
	// Build a transformation from separate x, y and z scale values.
	//********************************************************************************

	Matrix33of44& Matrix33of44::FromScale(F32 x, F32 y, F32 z)
	{
		Element(0, 0) = x;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = y;
		Element(1, 2) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = z;

		return *this;
	}


	//********************************************************************************
	// Scale this matrix uniformly.
	//********************************************************************************

	Matrix33of44& Matrix33of44::Scale(F32 scale)
	{
		for (S32 i = 0; i < 3; i++)
		{
			Element(0, i) *= scale;
			Element(1, i) *= scale;
			Element(2, i) *= scale;
		}

		return *this;
	}


	//********************************************************************************
	// Scale this matrix non-uniformly.
	//********************************************************************************

	Matrix33of44& Matrix33of44::Scale(F32 x, F32 y, F32 z)
	{
		for (S32 i = 0; i < 3; i++)
		{
			Element(0, i) *= x;
			Element(1, i) *= y;
			Element(2, i) *= z;
		}

		return *this;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation.
	//********************************************************************************

	Matrix33of44& Matrix33of44::Invert(void)
	{
		*this = Inverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation.
	//********************************************************************************

	Matrix33of44 Matrix33of44::Inverse(void) const
	{
		Matrix33of44 result(*this);

		F32 cofactor0 = Element(1, 1) * Element(2, 2) - Element(1, 2) * Element(2, 1);
		F32 cofactor3 = Element(0, 2) * Element(2, 1) - Element(0, 1) * Element(2, 2);
		F32 cofactor6 = Element(0, 1) * Element(1, 2) - Element(0, 2) * Element(1, 1);

		F32 determinant = Element(0, 0) * cofactor0 + Element(1, 0) * cofactor3 + Element(2, 0) * cofactor6;

		if (Abs(determinant) >= Epsilon)
		{
			F32 invDeterminant = 1.0f / determinant;

			result(0, 0) = invDeterminant * cofactor0;
			result(0, 1) = invDeterminant * cofactor3;
			result(0, 2) = invDeterminant * cofactor6;

			result(1, 0) = invDeterminant * (Element(1, 2) * Element(2, 0) - Element(1, 0) * Element(2, 2));
			result(1, 1) = invDeterminant * (Element(0, 0) * Element(2, 2) - Element(0, 2) * Element(2, 0));
			result(1, 2) = invDeterminant * (Element(0, 2) * Element(1, 0) - Element(0, 0) * Element(1, 2));

			result(2, 0) = invDeterminant * (Element(1, 0) * Element(2, 1) - Element(1, 1) * Element(2, 0));
			result(2, 1) = invDeterminant * (Element(0, 1) * Element(2, 0) - Element(0, 0) * Element(2, 1));
			result(2, 2) = invDeterminant * (Element(0, 0) * Element(1, 1) - Element(0, 1) * Element(1, 0));
		}
		else
		{
			LIBMATHS_ASSERT(0 && "Singular matrix could not be inverted");
		}

		return result;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation (assumes this matrix
	// is an orthonormal transformation).
	//********************************************************************************

	Matrix33of44& Matrix33of44::OrthonormalInvert(void)
	{
		*this = OrthonormalInverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation (assumes this matrix is an orthonormal transformation).
	//********************************************************************************

	Matrix33of44 Matrix33of44::OrthonormalInverse(void) const
	{
		F32 scale = Element(0, 0) * Element(0, 0) + Element(0, 1) * Element(0, 1) + Element(0, 2) * Element(0, 2);
		Matrix33of44 result;

		scale = 1.0f / scale;

		result(0, 0) = Element(0, 0) * scale;
		result(0, 1) = Element(1, 0) * scale;
		result(0, 2) = Element(2, 0) * scale;
		result(1, 0) = Element(0, 1) * scale;
		result(1, 1) = Element(1, 1) * scale;
		result(1, 2) = Element(2, 1) * scale;
		result(2, 0) = Element(0, 2) * scale;
		result(2, 1) = Element(1, 2) * scale;
		result(2, 2) = Element(2, 2) * scale;

		return result;
	}


	//********************************************************************************
	// Orthonormalise this matrix using the Gram-Schmidt algorithm.
	// NOTE: Only the upper 3x3 sub-matrix is affected.
	//
	// M = [x|y|z], then orthonormal output matrix is Q = [nx|ny|nz],
	//
	// nx = x/|x|
	// ny = (y-nx(nx*y))/|y-nx(nx*y)|
	// nz = nx^ny
	//
	// Where |V| indicates length of vector V and A*B indicates dot product of vectors
	// A and B.
	//********************************************************************************

	Matrix33of44& Matrix33of44::Orthonormalise(void)
	{
		Vector3x nx(XAxis().Normalised());
		Vector3x y(YAxis());
		Vector3x ny((y - nx * (nx * y)).Normalised());

		XAxis(nx);
		YAxis(ny);
		ZAxis(nx ^ ny);

		return *this;
	}


	//********************************************************************************
	// Get the determinant of this matrix.
	//********************************************************************************

	F32 Matrix33of44::Determinant(void) const
	{
		return 	Element(0, 0) * (Element(1, 1) * Element(2, 2) - Element(1, 2) * Element(2, 1)) +
				Element(1, 0) * (Element(0, 2) * Element(2, 1) - Element(0, 1) * Element(2, 2)) +
				Element(2, 0) * (Element(0, 1) * Element(1, 2) - Element(0, 2) * Element(1, 1));
	}


	//********************************************************************************
	// Build this matrix as an interpolation between the two given matrices,
	// with possible extra 360 degree spins.
	//********************************************************************************

	Matrix33of44& Matrix33of44::Lerp(F32 t, Matrix33of44& m0, Matrix33of44& m1, U32 numSpins)
	{
		Quaternion rotation;

		rotation.SlerpShortest(t, Quaternion(m0), Quaternion(m1), numSpins);

		return FromQuaternion(rotation);
	}

// namespace libmaths
}
