//********************************************************************************
//
// 4x4 matrix class. Used for homogenous 3D transformations with 3 and 4
// component vectors.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "matrix44.h"
#include "matrix33of44.h"
#include "math.h"
#include "vector4.h"


namespace libmaths
{

	//********************************************************************************
	// Construct from a 3x3 matrix.
	// The fourth column is initialised to (0, 0, 0, 1)
	// The translation is initialised to (0, 0, 0)
	//********************************************************************************

	Matrix44::Matrix44(const Matrix33of44& matrix)
	{
		Element(0, 0) = matrix(0, 0);
		Element(0, 1) = matrix(0, 1);
		Element(0, 2) = matrix(0, 2);
		Element(0, 3) = 0.0f;

		Element(1, 0) = matrix(1, 0);
		Element(1, 1) = matrix(1, 1);
		Element(1, 2) = matrix(1, 2);
		Element(1, 3) = 0.0f;

		Element(2, 0) = matrix(2, 0);
		Element(2, 1) = matrix(2, 1);
		Element(2, 2) = matrix(2, 2);
		Element(2, 3) = 0.0f;

		Element(3, 0) = 0.0f;
		Element(3, 1) = 0.0f;
		Element(3, 2) = 0.0f;
		Element(3, 3) = 1.0f;
	}


	//********************************************************************************
	// Build a transformation from a quaternion.
	//********************************************************************************

	Matrix44& Matrix44::FromQuaternion(const Quaternion& quaternion)
	{
		F32 w2 = quaternion.W() * 2.0f;

		F32 d1 = (quaternion.X() * quaternion.X()) * 2.0f;
		F32 d2 = (quaternion.Y() * quaternion.Y()) * 2.0f;
		F32 d3 = (quaternion.Z() * quaternion.Z()) * 2.0f;

		F32 d4 = quaternion.X() * w2;
		F32 d5 = quaternion.Y() * w2;
		F32 d6 = quaternion.Z() * w2;

		F32 d7 = (quaternion.X() * quaternion.Y()) * 2.0f;
		F32 d8 = (quaternion.X() * quaternion.Z()) * 2.0f;
		F32 d9 = (quaternion.Y() * quaternion.Z()) * 2.0f;

		Element(0, 0) = 1.0f - d2 - d3;
		Element(1, 0) = d7 - d6;
		Element(2, 0) = d8 + d5;

		Element(0, 1) = d7 + d6;
		Element(1, 1) = 1.0f - d1 - d3;
		Element(2, 1) = d9 - d4;

		Element(0, 2) = d8 - d5;
		Element(1, 2) = d9 + d4;
		Element(2, 2) = 1.0f - d1 - d2;

		Element(3, 0) = Element(3, 1) = Element(3, 2) = 0.0f;
		Element(0, 3) = Element(1, 3) = Element(2, 3) = 0.0f;
		Element(3, 3) = 1.0f;

		return (*this);
	}


	//********************************************************************************
	// Convert this matrix into a quaternion representing the same rotation.
	//********************************************************************************

	Quaternion Matrix44::ToQuaternion(void) const
	{
		return Quaternion(*this);
	}


	//********************************************************************************
	// Build a transformation representing a rotation of angle radians anti-clockwise
	// about the axis defined by axis. Angles of zero will produce no transform.
	//********************************************************************************

	Matrix44& Matrix44::FromAxisAngle(const Vector3x& axis, F32 angle)
	{
		F32 cosine;
		F32 sine;

		CosSin(angle, cosine, sine);

		F32 oneMinusCos = 1.0f - cosine;

		F32 x2 = axis.X() * axis.X();
		F32 y2 = axis.Y() * axis.Y();
		F32 z2 = axis.Z() * axis.Z();

		F32 xym = axis.X() * axis.Y() * oneMinusCos;
		F32 xzm = axis.X() * axis.Z() * oneMinusCos;
		F32 yzm = axis.Y() * axis.Z() * oneMinusCos;

		F32 xSin = axis.X() * sine;
		F32 ySin = axis.Y() * sine;
		F32 zSin = axis.Z() * sine;

		Element(0, 0) = x2 * oneMinusCos + cosine;
		Element(0, 1) = xym + zSin;
		Element(0, 2) = xzm - ySin;
		Element(0, 3) = 0.0f;

		Element(1, 0) = xym - zSin;
		Element(1, 1) = y2 * oneMinusCos + cosine;
		Element(1, 2) = yzm + xSin;
		Element(1, 3) = 0.0f;

		Element(2, 0) = xzm + ySin;
		Element(2, 1) = yzm - xSin;
		Element(2, 2) = z2 * oneMinusCos + cosine;
		Element(2, 3) = 0.0f;

		Element(3, 0) = 0.0f;
		Element(3, 1) = 0.0f;
		Element(3, 2) = 0.0f;
		Element(3, 3) = 1.0f;

		return *this;
	}


	//********************************************************************************
	// Decompose the rotation part of this matrix into a rotation of radians
	// anti-clockwise about an axis.
	//********************************************************************************

	void Matrix44::ToAxisAngle(Vector3x& axis, F32& angle) const
	{
		F32 trace = 0.5f * ((Element(0, 0) + Element(1, 1) + Element(2, 2)) - 1.0f);
		F32 cosine = Clamp(trace, -1.0f, 1.0f);

		angle = Acos(cosine);

		if (angle > 0.0f)
		{
			if (angle < PI)
			{
				axis.XYZ(Element(1, 2) - Element(2, 1), Element(2, 0) - Element(0, 2), Element(0, 1) - Element(1, 0));
				axis.Normalise();
			}
			else
			{
				F32 halfInverse;

				if (Element(0, 0) >= Element(1, 1))
				{
					if (Element(0, 0) >= Element(2, 2))
					{
						axis.X(0.5f * Sqrt(Element(0, 0) - Element(1, 1) - Element(2, 2) + 1.0f));
						halfInverse = 0.5f / axis.X();
						axis.Y(halfInverse * Element(1, 0));
						axis.Z(halfInverse * Element(2, 0));
					}
					else
					{
						axis.Z(0.5f * Sqrt(Element(2, 2) - Element(0, 0) - Element(1, 1) + 1.0f));
						halfInverse = 0.5f / axis.Z();
						axis.X(halfInverse * Element(2, 0));
						axis.Y(halfInverse * Element(2, 1));
					}
				}
				else
				{
					if (Element(1, 1) >= Element(2, 2))
					{
						axis.Y(0.5f * Sqrt(Element(1, 1) - Element(0, 0) - Element(2, 2) + 1.0f));
						halfInverse = 0.5f / axis.Y();
						axis.X(halfInverse * Element(1, 0));
						axis.Z(halfInverse * Element(2, 1));
					}
					else
					{
						axis.Z(0.5f * Sqrt(Element(2, 2) - Element(0, 0) - Element(1, 1) + 1.0f));
						halfInverse = 0.5f / axis.Z();
						axis.X(halfInverse * Element(2, 0));
						axis.Y(halfInverse * Element(2, 1));
					}
				}
			}
		}
		else
		{
			// The angle is 0 and the matrix is the identity.  Any axis will
			// work, so just use the x-axis.

			axis.XYZ(1.0f, 0.0f, 0.0f);
		}
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	// NOTE: This assumes right-handed orientation.
	//********************************************************************************

	Matrix44& Matrix44::FromDirectionRH(Vector3x direction, Vector3x up)
	{
		Identity();

		Vector3x right;

		direction.Normalise();
		direction.RightUpVectors(right, up);

		XAxis(right);
		YAxis(direction);
		ZAxis(up);

		return *this;
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	// NOTE: This assumes left-handed orientation.
	//********************************************************************************

	Matrix44& Matrix44::FromDirectionLH(Vector3x direction, Vector3x up)
	{
		Identity();

		Vector3x right;

		direction.Normalise();
		direction.RightUpVectors(right, up);

		XAxis(-right);
		YAxis(up);
		ZAxis(direction);

		return *this;
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the x axis.
	//********************************************************************************

	Matrix44& Matrix44::FromXAngle(F32 angle)
	{
		F32 c, s;

		CosSin(angle, c, s);

		*this = Matrix44(1.0f, 0.0f, 0.0f, 0.0f, c, s, 0.0f, -s, c);

		return *this;
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the y axis.
	//********************************************************************************

	Matrix44& Matrix44::FromYAngle(F32 angle)
	{
		F32 c, s;

		CosSin(angle, c, s);

		*this = Matrix44(c, 0.0f, -s, 0.0f, 1.0f, 0.0f, s, 0.0f, c);

		return *this;
	}


	//********************************************************************************
	// Build a transformation from an anti-clockwise angle around the z axis.
	//********************************************************************************

	Matrix44& Matrix44::FromZAngle(F32 angle)
	{
		F32 c, s;

		CosSin(angle, c, s);

		*this = Matrix44(c, s, 0.0f, -s, c, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f);

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about X then Y then Z by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromXYZAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44(cy*cz, cy*sz, -sy, ((sx*sy)*cz) + (cx*-sz), ((sx*sy)*sz) + (cx*cz), sx*cy, ((cx*sy)*cz) + (-sx*-sz), ((cx*sy)*sz) + (-sx*cz), cx*cy);

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about X then Z then Y by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromXZYAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44(cz*cy, sz, cz*-sy, ((cx*-sz)*cy) + (sx*sy), cx*cz, ((cx*-sz)*-sy) + (sx*cy), ((-sx*-sz)*cy) + (cx*sy), -sx*cz, ((-sx*-sz)*-sy) + (cx*cy));

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about Y then X then Z by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromYXZAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44((cy*cz) + ((-sy*-sx)*-sz), (cy*sz) + ((-sy*-sx)*cz), -sy*cx, cx*-sz, cx*cz, sx, (sy*cz) + ((cy*-sx)*-sz), (sy*sz) + ((cy*-sx)*cz), cy*cx);

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about Y then Z then X by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromYZXAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44(cy*cz, ((cy*sz)*cx) + (-sy*-sx), ((cy*sz)*sx) + (-sy*cx), -sz, cz*cx, cz*sx, sy*cz, ((sy*sz)*cx) + (cy*-sx), ((sy*sz)*sx) + (cy*cx));

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about Z then X then Y by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromZXYAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44((cz*cy) + ((sz*sx)*sy), sz*cx, (cz*-sy) + ((sz*sx)*cy), (-sz*cy) + ((cz*sx)*sy), cz*cx, (-sz*-sy) + ((cz*sx)*cy), cx*sy, -sx, cx*cy);

		return *this;
	}


	//********************************************************************************
	// Rotate this matrix about Z then Y then X by the given anti-clockwise angles.
	//********************************************************************************

	Matrix44& Matrix44::FromZYXAngles(const Vector3x& angles)
	{
		F32 cx, cy, cz, sx, sy, sz;

		CosSin(angles.X(), cx, sx);
		CosSin(angles.Y(), cy, sy);
		CosSin(angles.Z(), cz, sz);

		*this = Matrix44(cz*cy, (sz*cx) + ((cz*-sy)*-sx), (sz*sx) + ((cz*-sy)*cx), -sz*cy, (cz*cx) + ((-sz*-sy)*-sx), (cz*sx) + ((-sz*-sy)*cx), sy, cy*-sx, cy*cx);

		return *this;
	}


	//********************************************************************************
	// Build a transformation from a scale value.
	//********************************************************************************

	Matrix44& Matrix44::FromScale(F32 scale)
	{
		Element(0, 0) = scale;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;
		Element(0, 3) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = scale;
		Element(1, 2) = 0.0f;
		Element(1, 3) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = scale;
		Element(2, 3) = 0.0f;

		Element(3, 0) = 0.0f;
		Element(3, 1) = 0.0f;
		Element(3, 2) = 0.0f;
		Element(3, 3) = 1.0f;

		return *this;
	}


	//********************************************************************************
	// Build a transformation from separate x, y, z and w scale values.
	//********************************************************************************

	Matrix44& Matrix44::FromScale(F32 x, F32 y, F32 z, F32 w)
	{
		Element(0, 0) = x;
		Element(0, 1) = 0.0f;
		Element(0, 2) = 0.0f;
		Element(0, 3) = 0.0f;

		Element(1, 0) = 0.0f;
		Element(1, 1) = y;
		Element(1, 2) = 0.0f;
		Element(1, 3) = 0.0f;

		Element(2, 0) = 0.0f;
		Element(2, 1) = 0.0f;
		Element(2, 2) = z;
		Element(2, 3) = 0.0f;

		Element(3, 0) = 0.0f;
		Element(3, 1) = 0.0f;
		Element(3, 2) = 0.0f;
		Element(3, 3) = w;

		return *this;
	}


	//********************************************************************************
	// Scale this matrix uniformly.
	//********************************************************************************

	Matrix44& Matrix44::Scale(F32 scale)
	{
		for (S32 i = 0; i < 3; i++)
		{
			Element(0, i) *= scale;
			Element(1, i) *= scale;
			Element(2, i) *= scale;
			Element(3, i) *= scale;
		}

		return *this;
	}


	//********************************************************************************
	// Scale this matrix non-uniformly.
	//********************************************************************************

	Matrix44& Matrix44::Scale(F32 x, F32 y, F32 z, F32 w)
	{
		for (S32 i = 0; i < 3; i++)
		{
			Element(0, i) *= x;
			Element(1, i) *= y;
			Element(2, i) *= z;
			Element(3, i) *= w;
		}

		return *this;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation.
	//********************************************************************************

	Matrix44& Matrix44::Invert(void)
	{
		*this = Inverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation.
	//********************************************************************************

	Matrix44 Matrix44::Inverse(void) const
	{
		Matrix44 result;
		Matrix44 subMatrix;

		// Get determinant of the matrix

		F32 determinant = DeterminantSubMatrix(4);

		if (determinant == 0.0f)
		{
			result = IDENTITY();
		}
		else
		{
			F32 invDeterminant = 1.0f / determinant;

			for (S32 cell = 0; cell < 16; cell++)
			{
				S32 XSelect = cell & 3;
				S32 YSelect = cell >> 2;

				S32 XWrite = 0;
				S32 YWrite = 0;

				for (S32 x = 0; x < 4; x++)
				{
					if (x != XSelect)
					{
						YWrite = 0;

						for (S32 y = 0; y < 4; y++)
						{
							if (y != YSelect)
							{
								subMatrix.Element(XWrite, YWrite++) = Element(x, y);
							}
						}

						XWrite++;
					}
				}

				if ((XSelect + YSelect) & 1)
				{
					result.Element(YSelect, XSelect) = -(subMatrix.DeterminantSubMatrix(3) * invDeterminant);
				}
				else
				{
					result.Element(YSelect, XSelect) = (subMatrix.DeterminantSubMatrix(3) * invDeterminant);
				}
			}
		}

		return result;
	}


	//********************************************************************************
	// Invert this matrix, producing an inverse transformation (assumes this matrix
	// is an orthonormal transformation).
	//********************************************************************************

	Matrix44& Matrix44::OrthonormalInvert(void)
	{
		*this = OrthonormalInverse();

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from the inverse of this matrix, producing an inverse
	// transformation (assumes this matrix is an orthonormal transformation).
	//********************************************************************************

	Matrix44 Matrix44::OrthonormalInverse(void) const
	{
		F32 scale = Element(0, 0) * Element(0, 0) + Element(0, 1) * Element(0, 1) + Element(0, 2) * Element(0, 2);
		F32 result[3][3];

		scale = 1.0f / scale;

		result[0][0] = Element(0, 0) * scale;
		result[0][1] = Element(1, 0) * scale;
		result[0][2] = Element(2, 0) * scale;
		result[1][0] = Element(0, 1) * scale;
		result[1][1] = Element(1, 1) * scale;
		result[1][2] = Element(2, 1) * scale;
		result[2][0] = Element(0, 2) * scale;
		result[2][1] = Element(1, 2) * scale;
		result[2][2] = Element(2, 2) * scale;

		return Matrix44
		(
			result[0][0], result[0][1], result[0][2], 0.0f,
			result[1][0], result[1][1], result[1][2], 0.0f,
			result[2][0], result[2][1], result[2][2], 0.0f,
			-(result[0][0] * Element(3, 0) + result[1][0] * Element(3, 1) + result[2][0] * Element(3, 2)),
			-(result[0][1] * Element(3, 0) + result[1][1] * Element(3, 1) + result[2][1] * Element(3, 2)),
			-(result[0][2] * Element(3, 0) + result[1][2] * Element(3, 1) + result[2][2] * Element(3, 2)), 1.0f
		);
	}


	//********************************************************************************
	// Orthonormalise this matrix using the Gram-Schmidt algorithm.
	// NOTE: Only the upper 3x3 sub-matrix is affected.
	//
	// M = [x|y|z], then orthonormal output matrix is Q = [nx|ny|nz],
	//
	// nx = x/|x|
	// ny = (y-nx(nx*y))/|y-nx(nx*y)|
	// nz = nx^ny
	//
	// Where |V| indicates length of vector V and A*B indicates dot product of vectors
	// A and B.
	//********************************************************************************

	Matrix44& Matrix44::Orthonormalise(void)
	{
		Vector3x nx(XAxis().Normalised());
		Vector3x y(YAxis());
		Vector3x ny((y - nx * (nx * y)).Normalised());

		XAxis(nx);
		YAxis(ny);
		ZAxis(nx ^ ny);

		return *this;
	}


	//********************************************************************************
	// Get the determinant of this matrix.
	//********************************************************************************

	F32 Matrix44::Determinant(void) const
	{
		F32 det;
		F32 da = Element(2, 2) * Element(3, 3) - Element(2, 3) * Element(3, 2);
		F32 db = Element(2, 1) * Element(3, 3) - Element(2, 3) * Element(3, 1);
		F32 dc = Element(2, 1) * Element(3, 2) - Element(2, 2) * Element(3, 1);
		F32 dd = Element(2, 0) * Element(3, 3) - Element(2, 3) * Element(3, 0);
		F32 de = Element(2, 0) * Element(3, 2) - Element(2, 2) * Element(3, 0);
		F32 df = Element(2, 0) * Element(3, 1) - Element(2, 1) * Element(3, 0);

		det = Element(0, 0) * (Element(1, 1) * da - Element(1, 2) * db + Element(1, 3) * dc);
		det -= Element(0, 1) * (Element(1, 0) * da - Element(1, 2) * dd + Element(1, 3) * de);
		det += Element(0, 2) * (Element(1, 0) * db - Element(1, 1) * dd + Element(1, 3) * df);
		det -= Element(0, 3) * (Element(1, 0) * dc - Element(1, 1) * de + Element(1, 2) * df);

		return det;
	}


	//********************************************************************************
	// Get the determinant of a sub-matrix of this matrix.
	//********************************************************************************

	F32 Matrix44::DeterminantSubMatrix(S32 size) const
	{
		Matrix44 newMatrix;
		S32 x, y, j, nextColumn;
		F32 determinant = 0;

		if (size > 1)
		{
			// Subdivide matrix

			for (j = 0; j < size; j++)
			{
				nextColumn = 0;

				for (x = 0; x < size; x++)
				{
					if (x != j)
					{
						for (y = 1; y < size; y++)
						{
							newMatrix.Element(nextColumn, y - 1) = Element(x, y);
						}

						nextColumn++;
					}
				}

				if (j & 1)
				{
					determinant += -(newMatrix.DeterminantSubMatrix(size - 1) * Element(j, 0));
				}
				else
				{
					determinant += (newMatrix.DeterminantSubMatrix(size - 1) * Element(j, 0));
				}
			}

			return determinant;
		}
		else
		{
			return Element(0, 0);
		}
	}


	//********************************************************************************
	// Create a matrix that reflects vertices around a plane.
	//********************************************************************************

	Matrix44& Matrix44::FromReflection(const Plane& plane)
	{
		Vector3x normal = plane.Normal().Normalised();

		F32 x = normal.X();
		F32 y = normal.Y();
		F32 z = normal.Z();

		F32 x2 = -2.0f * x;
		F32 y2 = -2.0f * y;
		F32 z2 = -2.0f * z;

		F32 d = plane.D();

		Element(0, 0) = (x2 * x) + 1.0f;
		Element(0, 1) = y2 * x;
		Element(0, 2) = z2 * x;
		Element(0, 3) = 0.0f;
		Element(1, 0) = x2 * y;
		Element(1, 1) = (y2 * y) + 1.0f;
		Element(1, 2) = z2 * y;
		Element(1, 3) = 0.0f;
		Element(2, 0) = x2 * z;
		Element(2, 1) = y2 * z;
		Element(2, 2) = (z2 * z) + 1.0f;
		Element(2, 3) = 0.0f;
		Element(3, 0) = x2 * d;
		Element(3, 1) = y2 * d;
		Element(3, 2) = z2 * d;
		Element(3, 3) = 1.0f;

		return *this;
	}


	//********************************************************************************
	// Build this matrix as an interpolation between the two given matrices,
	// with possible extra 360 degree spins.
	//********************************************************************************

	Matrix44& Matrix44::Lerp(F32 t, Matrix44& m0, Matrix44& m1, U32 numSpins)
	{
		Quaternion rotation;
		Vector3x v0(m0.Translation());
		Vector3x v1(m1.Translation());

		rotation.SlerpShortest(t, Quaternion(m0), Quaternion(m1), numSpins);

		FromQuaternion(rotation);

		Translation(v0 + (t * (v1 - v0)));

		return *this;
	}


	//********************************************************************************
	// Get a new matrix from this matrix multiplied by (concatenated with) another
	// matrix.
	//********************************************************************************

	Matrix44 Matrix44::operator * (const Matrix33of44& rhs)
	{
		Matrix44 result(*this);

		result *= Matrix44(rhs);

		return result;
	}


	//********************************************************************************
	// Multiply this matrix by (concatenated with) another matrix.
	//********************************************************************************

	Matrix44& Matrix44::operator *= (const Matrix33of44& rhs)
	{
		*this *= Matrix44(rhs);

		return *this;
	}

	// namespace libmaths
}
