//********************************************************************************
//
// Amplitude translation classes returning an amplitude for a parametric value.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************

#pragma once


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "parametricconverters.h"
#include "libmaths.h"


namespace libmaths
{

	//********************************************************************************
	// Normalise an input value to between 0 and 1, or > 1 if not looped.
	//********************************************************************************

	F32 ParametricConverter::Normalise(F32 value)
	{
		if (value < 0.0f)
		{
			return ((_looped) ? Fmod(value, _time) * -InvTime() : 0.0f);
		}
		else
		{
			return ((_looped) ? Fmod(value, _time) : value) * InvTime();
		}
	}


	//********************************************************************************
	// A simple linear, attack, sustain, decay amplitude translator.
	//********************************************************************************

	//********************************************************************************
	// Construct a GeAttackSustainDecayAT.
	//********************************************************************************

	GeAttackSustainDecayAT::GeAttackSustainDecayAT(Bool looped, F32 attack, F32 sustain, F32 decay, F32 amplitude) :
		ParametricConverter(looped, attack + sustain + decay),
		m_Amplitude(amplitude),
		m_Attack(attack * InvTime()),
		m_InvAttack(1.0f / (attack * InvTime())),
		m_InvDecay(1.0f / (decay * InvTime())),
		m_DecayThreshold((Time() - decay) * InvTime())
	{
	}


	//********************************************************************************
	// Translate a value into an amplitude.
	//********************************************************************************

	F32 GeAttackSustainDecayAT::Translate(F32 value)
	{
		F32 result = m_Amplitude;

		value = Normalise(value);

		if (value >= m_DecayThreshold)
		{
			result = (value - m_DecayThreshold) * m_InvDecay;
			result = (result > 1.0f) ? 0.0f : (1.0f - result) * m_Amplitude;
		}
		else if (value < m_Attack)
		{
			result *= (value * m_InvAttack);
		}

		return result;
	}


	//********************************************************************************
	// A simple squared curve, attack, sustain, decay amplitude translator.
	//********************************************************************************

	//********************************************************************************
	// Construct a GeAttackSustainDecaySqrAT.
	//********************************************************************************

	GeAttackSustainDecaySqrAT::GeAttackSustainDecaySqrAT(Bool looped, F32 attack, F32 sustain, F32 decay, F32 amplitude) :
		GeAttackSustainDecayAT(looped, attack, sustain, decay, amplitude)
	{
	}


	//********************************************************************************
	// Translate a value into an amplitude.
	//********************************************************************************

	F32 GeAttackSustainDecaySqrAT::Translate(F32 value)
	{
		F32 result = 1.0f;

		value = Normalise(value);

		if (value >= m_DecayThreshold)
		{
			result = (value - m_DecayThreshold) * m_InvDecay;
			result = (result > 1.0f) ? 0.0f : 1.0f - (result * result);
		}
		else if (value < m_Attack)
		{
			result = 1.0f - (value * m_InvAttack);
			result = 1.0f - (result * result);
		}

		return result * m_Amplitude;
	}


	//********************************************************************************
	// A simple ease-out / ease-in cosine-based transition from 0 to 1.
	//********************************************************************************


	//********************************************************************************
	// Translate a value into an amplitude.
	//********************************************************************************

	F32 SineEaseInParametricConverter::Translate(F32 value)
	{
		F32 result = 1.0f;

		value = Normalise(value);

		if (value < 1.0f)
		{
			result = Sin(value * PI * 0.5f);
		}

		return result;
	}


	//********************************************************************************
	// Translate a value into an amplitude.
	//********************************************************************************

	F32 SineEaseOutParametricConverter::Translate(F32 value)
	{
		F32 result = 1.0f;

		value = Normalise(value);

		if (value < 1.0f)
		{
			result = 1.0f - Cos(value * PI * 0.5f);
		}

		return result;
	}


	//********************************************************************************
	// Translate a value into an amplitude.
	//********************************************************************************

	F32 SineEaseInOutParametricConverter::Translate(F32 value)
	{
		F32 result = 1.0f;

		value = Normalise(value);

		if (value < 1.0f)
		{
			result = 0.5f - (Cos(value * PI) * 0.5f);
		}

		return result;
	}

	// namespace libmaths
}
