//********************************************************************************
//
// Perlin Noise class.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "perlinnoise.h"
#include "libmaths.h"


namespace libmaths
{

	//********************************************************************************
	// Initialise the Perlin object with a seed for the noise.
	//********************************************************************************

	void PerlinNoise::Initialise(U32 seed)
	{
		S32 i, j, nSwap;

		_random = seed;

		// Initialise the permutation table

		for (i = 0; i < SIZE; i++)
		{
			_p[i] = U8(i);
		}

		for (i = 0; i < SIZE; i++)
		{
			j = U8((_random * (F32)MASK));

			nSwap = _p[i];
			_p[i] = _p[j];
			_p[j] = U8(nSwap);
		}

		// Generate the gradient lookup tables

		for (i = 0; i < SIZE; i++)
		{
			// Ken Perlin proposes that the gradients are taken from the unit
			// circle/sphere for 2D/3D, but there are no noticeable difference
			// between that and what I'm doing here. For the sake of generality
			// I will not do that.

			_gx[i] = _random ^ 1.0f;
			_gy[i] = _random ^ 1.0f;
			_gz[i] = _random ^ 1.0f;
		}
	}


	//********************************************************************************
	// Get some one dimensional noise.
	//********************************************************************************

	F32 PerlinNoise::Noise1(F32 x)
	{
		// Compute what gradients to use

		S32 qx0 = Floor(x);
		S32 qx1 = qx0 + 1;
		F32 tx0 = x - F32(qx0);
		F32 tx1 = tx0 - 1;

		// Make sure we don't go outside the lookup table

		qx0 = qx0 & MASK;
		qx1 = qx1 & MASK;

		// Compute the dot product between the vectors and the gradients

		F32 v0 = _gx[qx0] * tx0;
		F32 v1 = _gx[qx1] * tx1;

		// Modulate with the weight function

		F32 wx = (3 - 2 * tx0) * tx0 * tx0;
		F32 v = v0 - wx * (v0 - v1);

		return v;
	}


	//********************************************************************************
	// Get some two dimensional noise.
	//********************************************************************************

	F32 PerlinNoise::Noise2(F32 x, F32 y)
	{
		// Compute what gradients to use

		S32 qx0 = Floor(x);
		S32 qx1 = qx0 + 1;
		F32 tx0 = x - F32(qx0);
		F32 tx1 = tx0 - 1;

		S32 qy0 = Floor(y);
		S32 qy1 = qy0 + 1;
		F32 ty0 = y - F32(qy0);
		F32 ty1 = ty0 - 1;

		// Make sure we don't go outside the lookup table

		qx0 = qx0 & MASK;
		qx1 = qx1 & MASK;

		qy0 = qy0 & MASK;
		qy1 = qy1 & MASK;

		// Permutate values to get pseudo randomly chosen gradients

		S32 q00 = _p[(qy0 + _p[qx0]) & MASK];
		S32 q01 = _p[(qy0 + _p[qx1]) & MASK];

		S32 q10 = _p[(qy1 + _p[qx0]) & MASK];
		S32 q11 = _p[(qy1 + _p[qx1]) & MASK];

		// Compute the dot product between the vectors and the gradients

		F32 v00 = _gx[q00] * tx0 + _gy[q00] * ty0;
		F32 v01 = _gx[q01] * tx1 + _gy[q01] * ty0;

		F32 v10 = _gx[q10] * tx0 + _gy[q10] * ty1;
		F32 v11 = _gx[q11] * tx1 + _gy[q11] * ty1;

		// Modulate with the weight function

		F32 wx = (3 - 2 * tx0) * tx0 * tx0;
		F32 v0 = v00 - wx * (v00 - v01);
		F32 v1 = v10 - wx * (v10 - v11);

		F32 wy = (3 - 2 * ty0) * ty0 * ty0;
		F32 v = v0 - wy * (v0 - v1);

		return v;
	}


	//********************************************************************************
	// Get some three dimensional noise.
	//********************************************************************************

	F32 PerlinNoise::Noise3(F32 x, F32 y, F32 z)
	{
		// Compute what gradients to use

		S32 qx0 = Floor(x);
		S32 qx1 = qx0 + 1;
		F32 tx0 = x - F32(qx0);
		F32 tx1 = tx0 - 1;

		S32 qy0 = Floor(y);
		S32 qy1 = qy0 + 1;
		F32 ty0 = y - F32(qy0);
		F32 ty1 = ty0 - 1;

		S32 qz0 = Floor(z);
		S32 qz1 = qz0 + 1;
		F32 tz0 = z - F32(qz0);
		F32 tz1 = tz0 - 1;

		// Make sure we don't go outside the lookup table

		qx0 = qx0 & MASK;
		qx1 = qx1 & MASK;

		qy0 = qy0 & MASK;
		qy1 = qy1 & MASK;

		qz0 = qz0 & MASK;
		qz1 = qz1 & MASK;

		// Permutate values to get pseudo randomly chosen gradients

		S32 q000 = _p[(qz0 + _p[(qy0 + _p[qx0]) & MASK]) & MASK];
		S32 q001 = _p[(qz0 + _p[(qy0 + _p[qx1]) & MASK]) & MASK];

		S32 q010 = _p[(qz0 + _p[(qy1 + _p[qx0]) & MASK]) & MASK];
		S32 q011 = _p[(qz0 + _p[(qy1 + _p[qx1]) & MASK]) & MASK];

		S32 q100 = _p[(qz1 + _p[(qy0 + _p[qx0]) & MASK]) & MASK];
		S32 q101 = _p[(qz1 + _p[(qy0 + _p[qx1]) & MASK]) & MASK];

		S32 q110 = _p[(qz1 + _p[(qy1 + _p[qx0]) & MASK]) & MASK];
		S32 q111 = _p[(qz1 + _p[(qy1 + _p[qx1]) & MASK]) & MASK];

		// Compute the dot product between the vectors and the gradients

		F32 v000 = _gx[q000] * tx0 + _gy[q000] * ty0 + _gz[q000] * tz0;
		F32 v001 = _gx[q001] * tx1 + _gy[q001] * ty0 + _gz[q001] * tz0;

		F32 v010 = _gx[q010] * tx0 + _gy[q010] * ty1 + _gz[q010] * tz0;
		F32 v011 = _gx[q011] * tx1 + _gy[q011] * ty1 + _gz[q011] * tz0;

		F32 v100 = _gx[q100] * tx0 + _gy[q100] * ty0 + _gz[q100] * tz1;
		F32 v101 = _gx[q101] * tx1 + _gy[q101] * ty0 + _gz[q101] * tz1;

		F32 v110 = _gx[q110] * tx0 + _gy[q110] * ty1 + _gz[q110] * tz1;
		F32 v111 = _gx[q111] * tx1 + _gy[q111] * ty1 + _gz[q111] * tz1;

		// Modulate with the weight function

		F32 wx = (3 - 2 * tx0) * tx0 * tx0;
		F32 v00 = v000 - wx * (v000 - v001);
		F32 v01 = v010 - wx * (v010 - v011);
		F32 v10 = v100 - wx * (v100 - v101);
		F32 v11 = v110 - wx * (v110 - v111);

		F32 wy = (3 - 2 * ty0) * ty0 * ty0;
		F32 v0 = v00 - wy * (v00 - v01);
		F32 v1 = v10 - wy * (v10 - v11);

		F32 wz = (3 - 2 * tz0) * tz0 * tz0;
		F32 v = v0 - wz * (v0 - v1);

		return v;
	}

// namespace libcore
}
