//********************************************************************************
//
// Quaternions are a space efficient representation for 3 dimensional rotations.
// They can be converted to and from regular rotation matrices for ease of use.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "quaternion.h"
#include "matrix33of44.h"
#include "matrix44.h"


namespace libmaths
{

	//********************************************************************************
	// Construct from a 3x3 matrix.
	//********************************************************************************

	Quaternion::Quaternion(const Matrix33of44& matrix)
	{
		FromMatrix(matrix);
	}


	//********************************************************************************
	// Construct from a 4x4 matrix.
	// The fourth column and translation are discarded in the process.
	//********************************************************************************

	Quaternion::Quaternion(const Matrix44& matrix)
	{
		FromMatrix(matrix);
	}


	//********************************************************************************
	// Build a transformation representing a rotation of angle radians anti-clockwise
	// about the axis defined by axis.
	//********************************************************************************

	Quaternion& Quaternion::FromAxisAngle(const Vector3x& axis, const F32 angle)
	{
		// The rotation representing the rotation is
		// q = cos(A/2)+sin(A/2)*(x*i+y*j+z*k)

		F32 sin, cos;

		libmaths::CosSin(0.5f * angle, cos, sin);

		if (Abs(cos) < FLOAT_EPSILON)
			cos = 0;

		if (Abs(sin) < FLOAT_EPSILON)
			sin = 0;

		return *this = Quaternion(axis * sin, cos);
	}


	//********************************************************************************
	// Decompose the quaternion into a rotation of radians anti-clockwise about an
	// axis.
	//********************************************************************************

	void Quaternion::ToAxisAngle(Vector3x& axis, F32& angle) const
	{
		// The quaternion representing the rotation is
		// q = cos(A/2)+sin(A/2)*(x*i+y*j+z*k)

		F32 lengthSqr = VectorLengthSqr();

		if (lengthSqr > 0.0f)
		{
			angle = 2.0f * Acos(W());

			F32 invLength = InvSqrt(lengthSqr);

			axis.X(X() * invLength);
			axis.Y(Y() * invLength);
			axis.Z(Z() * invLength);
		}
		else
		{
			// angle is 0 (mod 2*pi), so any axis will do

			angle = 0.0f;

			axis.X(1.0f);
			axis.Y(0.0f);
			axis.Z(0.0f);
		}
	}


	//********************************************************************************
	// Macro to convert from a matrix to a quaternion (used by 3x3 and 4x4 matrices).
	//********************************************************************************

#define FROM_MATRIX(matrix) \
\
	F32 trace = matrix(0, 0) + matrix(1, 1) + matrix(2, 2); \
\
	if (trace > libmaths::Epsilon) \
	{ \
		F32 scalar = Sqrt(trace + 1.0f); \
		F32 invScalar = 0.5f / scalar; \
\
		X((matrix(1, 2) - matrix(2, 1)) * invScalar); \
		Y((matrix(2, 0) - matrix(0, 2)) * invScalar); \
		Z((matrix(0, 1) - matrix(1, 0)) * invScalar); \
		W(0.5f * scalar); \
	} \
	else \
	{ \
		S32 i = 0; \
		static S32 next[3] = \
		{ 1, 2, 0 }; \
\
		if (matrix(1, 1) > matrix(0, 0)) \
		{ \
			i = 1; \
		} \
\
		if (matrix(2, 2) > matrix(i, i)) \
		{ \
			i = 2; \
		} \
\
		S32 j = next[i], k = next[j]; \
		F32 scalar = Sqrt(matrix(i, i) - matrix(j, j) - matrix(k, k) + 1.0f); \
		F32 invScalar = 0.5f / scalar; \
\
		(*this)[i] = 0.5f * scalar; \
		(*this)[j] = (matrix(j, i) + matrix(i, j)) * invScalar; \
		(*this)[k] = (matrix(i, k) + matrix(k, i)) * invScalar; \
		W((matrix(j, k) - matrix(k, j)) * invScalar); \
	}


	//********************************************************************************
	// Set this quaternion from a 3x3 matrix.
	//********************************************************************************

	Quaternion& Quaternion::FromMatrix(const Matrix33of44& matrix)
	{
		FROM_MATRIX(matrix);

		return *this;
	}


	//********************************************************************************
	// Set this quaternion from a 4x4 matrix.
	// The fourth column and translation are discarded in the process.
	//********************************************************************************

	Quaternion& Quaternion::FromMatrix(const Matrix44& matrix)
	{
		FROM_MATRIX(matrix);

		return *this;
	}


	//********************************************************************************
	// Convert this quaternion to a 3x3 matrix.
	//********************************************************************************

	void Quaternion::ToMatrix(Matrix33of44& matrix) const
	{
		matrix = Matrix33of44(*this);
	}


	//********************************************************************************
	// Convert this quaternion to a 4x4 matrix (default last row and column).
	//********************************************************************************

	void Quaternion::ToMatrix(Matrix44& matrix) const
	{
		matrix = Matrix44(*this);
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	// NOTE: This assumes right-handed orientation.
	//********************************************************************************

	Quaternion& Quaternion::FromDirectionRH(const Vector3x& direction, const Vector3x& up)
	{
		Matrix44 matrix;

		FromMatrix(matrix.FromDirectionRH(direction, up));

		return *this;
	}


	//********************************************************************************
	// Build a transformation from a direction vector with a specified "up" vector.
	// NOTE: This assumes left-handed orientation.
	//********************************************************************************

	Quaternion& Quaternion::FromDirectionLH(const Vector3x& direction, const Vector3x& up)
	{
		Matrix44 matrix;

		FromMatrix(matrix.FromDirectionLH(direction, up));

		return *this;
	}


	//********************************************************************************
	// Build a transformation that will rotate one vector to another when multiplied
	// with this quaternion.
	//********************************************************************************

	Quaternion& Quaternion::Reorient(const Vector3x& from, const Vector3x& to)
	{
		LIBMATHS_ASSERT(Approx(from.LengthSqr(), 1.0f));
		LIBMATHS_ASSERT(Approx(to.LengthSqr(), 1.0f));

		F32 d = from * to;
		F32 s = Sqrt((1.0f + d) * 2.0f);

		Vector3x axis;

		if (d < -1.0f + Epsilon())
		{
			if (Abs(from.X()) > Abs(from.Y()))
			{
				F32 scale = InvSqrt(from.Z() * from.Z() + from.X() * from.X());
				axis.XYZ(from.Z() * scale, 0, -from.X() * scale);
			}
			else
			{
				F32 scale = InvSqrt(from.Z() * from.Z() + from.Y() * from.Y());
				axis.XYZ(0, -from.Z() * scale, from.Y() * scale);
			}

			*this = Quaternion(axis, 0);
		}
		else
		{
			axis.Cross(from, to);

			*this = Quaternion(axis / s, s * 0.5f);
		}

		return *this;
	}


	//********************************************************************************
	// Invert this quaternion, producing an inverse transformation.
	//********************************************************************************

	Quaternion& Quaternion::Invert(void)
	{
		F32 normSqr = NormSqr();

		LIBMATHS_ASSERT(normSqr != 0.0f);

		normSqr = 1.0f / normSqr;

		_x *= -normSqr;
		_y *= -normSqr;
		_z *= -normSqr;
		_w *= normSqr;

		return *this;
	}


	//********************************************************************************
	// Get a new quaternion from the inverse of this quaternion, producing an inverse
	// transformation.
	//********************************************************************************

	Quaternion Quaternion::Inverse(void) const
	{
		return Quaternion(*this).Invert();
	}


	//********************************************************************************
	// Invert this quaternion, producing an inverse transformation.
	// Assumes this is a unit rotation.
	//********************************************************************************

	Quaternion Quaternion::UnitInvert(void)
	{
		// We need to ensure that the input has a unit normal or this optimisation
		// is invalid.

		LIBMATHS_ASSERT(Approx(NormSqr(), 1.0f));

		_x = -_x;
		_y = -_y;
		_z = -_z;

		return *this;
	}


	//********************************************************************************
	// Get a new quaternion from the inverse of this quaternion, producing an inverse
	// transformation. Assumes this is a unit rotation.
	//********************************************************************************

	Quaternion Quaternion::UnitInverse(void) const
	{
		return Quaternion(*this).UnitInvert();
	}


	//********************************************************************************
	// Normalise this quaternion returning its normal.
	//********************************************************************************

	F32 Quaternion::Normalise(void)
	{
		F32 length = NormSqr();

		length = Sqrt(length);

		if (length > libmaths::Epsilon)
		{
			F32 r = 1.0f / length;

			_x *= r;
			_y *= r;
			_z *= r;
			_w *= r;
		}

		return length;
	}


	//********************************************************************************
	// Get a new quaternion normalised from this quaternion.
	//********************************************************************************

	Quaternion Quaternion::Normalised(void) const
	{
		Quaternion normalised(*this);

		normalised.Normalise();

		return normalised;
	}


	//********************************************************************************
	// Set this quaternion by spherical linear interpolation between q0 and q1 by an
	// amount t with extra spins. This function takes no account of the possibility
	// that they may lie in opposite hemispheres.
	//********************************************************************************

	void Quaternion::Slerp(const F32 t, const Quaternion& q0, const Quaternion& q1, U32 extraspins)
	{
		F32 alpha = t, beta;
		F32 cosine = q0.Dot(q1);

		if ((1.0f - cosine) < libmaths::Epsilon)
		{
			beta = 1.0f - alpha;
		}
		else
		{
			F32 theta = Acos(Min(cosine, 1.0f));
			F32 phi = theta + (extraspins * PI);
			F32 invSine = 1.0f / Sin(theta);
			beta = Sin(theta - alpha * phi) * invSine;
			alpha = Sin(alpha * phi) * invSine;
		}

		*this = beta * q0 + alpha * q1;

		Normalise();
	}


	//********************************************************************************
	// Set this quaternion by spherical linear interpolation between q0 and q1 by an
	// amount t with extra spins. This function takes into account the possibility
	// that they lie in opposite hemispheres and will always choose the shortest
	// interpolation path between them.
	//********************************************************************************

	void Quaternion::SlerpShortest(const F32 t, const Quaternion& q0, const Quaternion& q1, U32 extraspins)
	{
		F32 flip = 1.0f;
		F32 alpha = t, beta;
		F32 cosine = q0.Dot(q1);

		if (cosine < 0.0f)
		{
			// If q0 is on opposite hemisphere to q1 then interpolate the other way
			// to ensure we get the shortest path.

			flip = -flip;
			cosine = -cosine;
		}

		if ((1.0f - cosine) < libmaths::Epsilon)
		{
			beta = 1.0f - alpha;
		}
		else
		{
			F32 theta = Acos(Min(cosine, 1.0f));
			F32 phi = theta + (extraspins * PI);
			F32 invSine = 1.0f / Sin(theta);
			beta = Sin(theta - alpha * phi) * invSine;
			alpha = Sin(alpha * phi) * invSine;
		}

		*this = beta * q0 + (alpha * flip) * q1;

		Normalise();
	}


	//********************************************************************************
	// Spherical quadratic interpolation.
	//********************************************************************************

	void Quaternion::Squad(const F32 t, const Quaternion& q0, const Quaternion& q1, const Quaternion& q2, const Quaternion& q3)
	{
		Quaternion q4, q5;

		q4.Slerp(t, q0, q3);
		q5.Slerp(t, q1, q2);

		(*this).Slerp(2.0f * t * (1.0f - t), q4, q5);
	}


	//********************************************************************************
	// Get the exponential of this quaternion.
	//********************************************************************************

	Quaternion Quaternion::Exp(void) const
	{
		// If q = A*(x*i+y*j+z*k) where (x,y,z) is unit length, then
		// exp(q) = cos(A)+sin(A)*(x*i+y*j+z*k). If sin(A) is near zero,
		// use exp(q) = cos(A)+A*(x*i+y*j+z*k) since A/sin(A) has limit 1.

		Quaternion result;
		F32 sine, cosine, angle;

		angle = VectorLength();

		LIBMATHS_ASSERT(angle > 0.0f);

		CosSin(angle, cosine, sine);

		F32 coefficient;

		if (Abs(sine) >= Quaternion::Epsilon())
		{
			coefficient = sine / angle;
		}
		else
		{
			coefficient = 1.0f;
		}

		result.X(coefficient * X());
		result.Y(coefficient * Y());
		result.Z(coefficient * Z());

		result.W(cosine);

		return result;
	}


	//********************************************************************************
	// Get the natural Logarithm of this quaternion.
	//********************************************************************************

	Quaternion Quaternion::Log(void) const
	{
		// If q = cos(A)+sin(A)*(x*i+y*j+z*k) where (x,y,z) is unit length, then
		// log(q) = A*(x*i+y*j+z*k). If sin(A) is near zero, use log(q) =
		// sin(A)*(x*i+y*j+z*k) since sin(A)/A has limit 1.

		Quaternion result;

		if (Abs(W()) < 1.0f)
		{
			F32 angle = Acos(W());
			F32 sine = Sin(angle);

			if (Abs(sine) >= Quaternion::Epsilon())
			{
				F32 coefficient = angle/sine;

				result.X(coefficient * X());
				result.Y(coefficient * Y());
				result.Z(coefficient * Z());

				result.W(0.0f);

				return result;
			}
		}

		result.X(X());
		result.Y(Y());
		result.Z(Z());

		result.W(0.0f);

		return result;
	}


	//********************************************************************************
	// Multiply a 3D vector by this quaternion.
	//********************************************************************************

	Vector3x Quaternion::operator * (const Vector3x& vector) const
	{
		return Matrix44(*this) * vector;
	}


	//********************************************************************************
	// Multiply a 3D vector by this quaternion.
	//********************************************************************************

	Vector3 Quaternion::operator * (const Vector3& vector) const
	{
		return Matrix44(*this) * vector;
	}

// namespace libmaths
}
