//********************************************************************************
//
// A class used to define rays in 3D space, often used for detecting intersections
// with other geometric types such as spheres and triangles.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "ray3.h"
#include "capsule.h"
#include "sphere.h"
#include "triangle3.h"
#include "plane.h"
#include "box.h"


namespace libmaths
{

	//********************************************************************************
	// Get the distance of the ray from a point in space. pointOnRay will return the
	// nearest point on the ray in parametric form; 0 .. ~, less than zero before the
	// start of the ray.
	//********************************************************************************

	F32 Ray3::Distance(const Vector3x& point, F32& pointOnRay, Bool clamp) const
	{
		return Sqrt(DistanceSqr(point, pointOnRay, clamp));
	}


	//********************************************************************************
	// Get the squared distance of the ray from a point in space. pointOnRay will
	// return the nearest point on the ray in parametric form; 0 .. ~, less than zero
	// before the start of the ray.
	//********************************************************************************

	F32 Ray3::DistanceSqr(const Vector3x& point, F32& pointOnRay, Bool clamp) const
	{
		Vector3x difference = point - _origin;

		// Calculate the scalar for the nearest point on the line to the point that we
		// are comparing.

		Vector3x lineVector = _direction;
		F32 lengthSqr = lineVector.LengthSqr();

		if (lengthSqr < Epsilon)
		{
			pointOnRay = 0.0f;
		}
		else
		{
			pointOnRay = (lineVector * difference) / lengthSqr;

			if (pointOnRay > 0.0f)
			{
				// Get the nearest point on the ray to the point that we are comparing and
				// return the squared distance between them.

				lineVector *= pointOnRay;

				difference -= lineVector;
			}

			if (clamp)
			{
				pointOnRay = Max(pointOnRay, 0.0f);
			}
		}

		return difference.LengthSqr();
	}


	//********************************************************************************
	// Does the ray intersect a box?
	//********************************************************************************

	Bool Ray3::Intersects(const AxisAlignedBox& box) const
	{
		return box.Intersects(*this);
	}


	//********************************************************************************
	// Does the ray intersect a box?
	//********************************************************************************

	Bool Ray3::Intersects(const AxisAlignedBox& box, Vector3x& point, Vector3x& normal) const
	{
		return box.Intersects(*this, point, normal);
	}


	//********************************************************************************
	// Does the ray intersect a capsule?
	//********************************************************************************

	Bool Ray3::Intersects(const Capsule& capsule) const
	{
		return capsule.Intersects(*this);
	}


	//********************************************************************************
	// Does the ray intersect a capsule?
	//********************************************************************************

	Bool Ray3::Intersects(const Capsule& capsule, Vector3x& point, Vector3x& normal) const
	{
		return capsule.Intersects(*this, point, normal);
	}


	//********************************************************************************
	// Does the ray intersect a sphere?
	//********************************************************************************

	Bool Ray3::Intersects(const Sphere& sphere) const
	{
		return sphere.Intersects(*this);
	}


	//********************************************************************************
	// Does the ray intersect a sphere?
	//********************************************************************************

	Bool Ray3::Intersects(const Sphere& sphere, Vector3x& point, Vector3x& normal) const
	{
		return sphere.Intersects(*this, point, normal);
	}


	//********************************************************************************
	// Does the ray intersect a triangle?
	//********************************************************************************

	Bool Ray3::Intersects(const Triangle3& triangle, Bool twoSided) const
	{
		return triangle.Intersects(*this, twoSided);
	}


	//********************************************************************************
	// Does the ray intersect a triangle?
	//********************************************************************************

	Bool Ray3::Intersects(const Triangle3& triangle, Vector3x& point, Vector3x& normal, Bool twoSided) const
	{
		return triangle.Intersects(*this, point, normal, twoSided);
	}


	//********************************************************************************
	// Does the ray intersect a plane?
	//********************************************************************************

	Bool Ray3::Intersects(const Plane& plane, Bool twoSided) const
	{
		return plane.Intersects(*this, twoSided);
	}


	//********************************************************************************
	// Does the ray intersect a plane?
	//********************************************************************************

	Bool Ray3::Intersects(const Plane& plane, Vector3x& point, Vector3x& normal, Bool twoSided) const
	{
		return plane.Intersects(*this, point, normal, twoSided);
	}

// namespace libmaths
}
