//********************************************************************************
//
// 2D bounding boxes for collision checks.
//
// These can be tested for collisions with other rectangles, points, circles or
// rays through the game world.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "rectangle.h"


namespace libmaths
{

	//********************************************************************************
	// Check if the line intersects this rectangle.
	// Does not allow intersections from within the rectangle.
	//********************************************************************************

	Bool AxisAlignedRectangle::Intersects(const Ray2& ray, Vector2& point, Vector2& normal) const
	{
		F32 tmin, tmax, tymin, tymax;

		Vector2 invDirection(1.0f / ray.Direction().X(), 1.0f / ray.Direction().Y());
		Vector2 topLeft(Minimum());
		Vector2 bottomRight(Maximum());

		if (invDirection.X() >= 0)
		{
			tmin = (topLeft.X() - ray.Origin().X()) * invDirection.X();
			tmax = (bottomRight.X() - ray.Origin().X()) * invDirection.X();
		}
		else
		{
			tmin = (bottomRight.X() - ray.Origin().X()) * invDirection.X();
			tmax = (topLeft.X() - ray.Origin().X()) * invDirection.X();
		}

		if (invDirection.Y() >= 0)
		{
			tymin = (topLeft.Y() - ray.Origin().Y()) * invDirection.Y();
			tymax = (bottomRight.Y() - ray.Origin().Y()) * invDirection.Y();
		}
		else
		{
			tymin = (bottomRight.Y() - ray.Origin().Y()) * invDirection.Y();
			tymax = (topLeft.Y() - ray.Origin().Y()) * invDirection.Y();
		}

		if ((tmin > tymax) || (tymin > tmax))
			return false;

		if (tymin > tmin)
			tmin = tymin;

		if (tymax < tmax)
			tmax = tymax;

		if (tmin <= 0.0f)
			return false;

		point = ray.Origin() + ray.Direction() * tmin;

		// TODO: This is not the surface normal of the side of the rectangle
		// that has been intersected, which is what we ought to really
		// be returning to the caller.

		normal = (point - ray.Origin()).Normalised();

		return true;
	}


	//********************************************************************************
	// Check if the line intersects this rectangle.
	// Does not allow intersections from within the rectangle.
	//********************************************************************************

	Bool AxisAlignedRectangle::Intersects(const LineSegment2& line) const
	{
		// TODO: Write this code.
		LIBMATHS_ASSERT(0 && "Not yet implemented");

		return false;
	}

// namespace libmaths
}
