//********************************************************************************
//
// 3D bounding spheres for collision checks.
//
// Bounding spheres can detect overlaps with other spheres, points and bounding
// boxes.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "box.h"
#include "sphere.h"


namespace libmaths
{

	//********************************************************************************
	// Construct bounding sphere to enclose the given bounding box.
	//********************************************************************************

	Sphere::Sphere(const AxisAlignedBox& box)
	{
		Point(box.Centre());
		Radius((box.Maximum() - box.Minimum()).Length() * 0.5f);
	}


	//********************************************************************************
	// Build bounding sphere enclosing the two given bounding spheres.
	//********************************************************************************

	void Sphere::Surround(const Sphere& sphere0, const Sphere& sphere1)
	{
		AxisAlignedBox b, s1b(sphere0), s2b(sphere1);

		b.Surround(s1b, s2b);

		Point(b.Centre());
		Radius(b.Extent().Length());
	}


	//********************************************************************************
	// Build bounding sphere enclosing the given bounding sphere and bounding box.
	//********************************************************************************

	void Sphere::Surround(const Sphere& sphere, const AxisAlignedBox& box)
	{
		AxisAlignedBox b, b2(sphere);

		b.Surround(box, b2);

		Point(b.Centre());
		Radius(b.Extent().Length());
	}


	//********************************************************************************
	//  Extend the bounding sphere to additionally contain the given bounding sphere.
	//********************************************************************************

	void Sphere::Add(const Sphere& sphere)
	{
		Surround(*this, sphere);
	}


	//********************************************************************************
	// Extend the bounding sphere to additionally contain the given bounding box.
	//********************************************************************************

	void Sphere::Add(const AxisAlignedBox& box)
	{
		Surround(*this, box);
	}


	//********************************************************************************
	// Check if the ray intersects this sphere.
	// Does not allow intersections from within the sphere.
	//********************************************************************************

	Bool Sphere::Intersects(const Ray3& ray) const
	{
		Vector3x distance(ray.Origin() - _sphere);

		F32 b = distance * ray.Direction();
		F32 c = distance.LengthSqr() - RadiusSqr();
		F32 d = (b * b) - c;

		if (d > 0)
		{
			F32 length = -b - Sqrt(d);

			if (length >= 0.0f)
			{
				return true;
			}
		}

		return false;
	}


	//********************************************************************************
	// Check if the ray intersects this sphere.
	// Does not allow intersections from within the sphere.
	//********************************************************************************

	Bool Sphere::Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal) const
	{
		Vector3x distance(ray.Origin() - _sphere);

		F32 b = distance * ray.Direction();
		F32 c = distance.LengthSqr() - RadiusSqr();
		F32 d = (b * b) - c;

		if (d > 0)
		{
			F32 length = -b - Sqrt(d);

			if (length >= 0.0f)
			{
				point = ray.Origin() + (ray.Direction() * length);

				normal = point - _sphere;
				normal.Normalise();

				return true;
			}
		}

		return false;
	}

// namespace libmaths
}
