//********************************************************************************
//
// Defines simple 3D triangles. Intended mainly for collision and ray intersection
// testing.
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "triangle3.h"


namespace libmaths
{

	//********************************************************************************
	// Check if the given point is inside the triangle.
	//********************************************************************************

	Bool Triangle3::Contains(const Vector3x& point) const
	{
		Vector3x v0 = _points[1] - _points[0];
		Vector3x v1 = _points[2] - _points[0];

		F32 a = v0 * v0;
		F32 b = v0 * v1;
		F32 c = v1 * v1;
		F32 acbb = (a * c) - (b * b);

		Vector3x vp = point - _points[0];

		F32 d = vp * v0;
		F32 e = vp * v1;
		F32 x = (d * c) - (e * b);
		F32 y = (e * a) - (d * b);
		F32 z = x + y - acbb;

		return ((((U32&)z & ~((U32&)x | (U32&)y)) & 0x80000000) != 0);

#if 0

		F32 ty, tx;
		const Vector3x* vtx0 = &_points[2];
		const Vector3x* vtx1 = &_points[0];
		S32 yflag0, yflag1, inside_flag = 0, line_flag;

		F32 x = Abs(_plane.Normal().X());
		F32 y = Abs(_plane.Normal().Y());
		F32 z = Abs(_plane.Normal().Z());

		if ((x >= y) && (x >= z))
		{
			tx = point[1]; ty = point[2]; yflag0 = ((*vtx0)[2] >= ty); inside_flag = line_flag = 0;

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[1]-(*vtx1)[1])) >= (((*vtx1)[1]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; line_flag = 1; }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[1];

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[1]-(*vtx1)[1])) >= (((*vtx1)[1]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; if (line_flag) return (inside_flag ? true : false); }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[2];

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[1]-(*vtx1)[1])) >= (((*vtx1)[1]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; }
		}
		else if ((y >= x) && (y >= z))
		{
			tx = point[0]; ty = point[2]; yflag0 = ((*vtx0)[2] >= ty); inside_flag = line_flag = 0;

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; line_flag = 1; }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[1];

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; if (line_flag) return (inside_flag ? true : false); }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[2];

			yflag1 = ((*vtx1)[2] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[2]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[2]-(*vtx1)[2])) ) == yflag1 ) inside_flag ^= 1; }
		}
		else if ((z >= x) && (z >= y))
		{
			tx = point[0]; ty = point[1]; yflag0 = ((*vtx0)[1] >= ty); inside_flag = line_flag = 0;

			yflag1 = ((*vtx1)[1] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[1]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[1]-(*vtx1)[1])) ) == yflag1 ) inside_flag ^= 1; line_flag = 1; }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[1];

			yflag1 = ((*vtx1)[1] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[1]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[1]-(*vtx1)[1])) ) == yflag1 ) inside_flag ^= 1; if (line_flag) return (inside_flag ? true : false); }
			yflag0 = yflag1; vtx0 = vtx1; vtx1 = &_points[2];

			yflag1 = ((*vtx1)[1] >= ty); if (yflag0 != yflag1)
			{ if ( ((((*vtx1)[1]-ty) * ((*vtx0)[0]-(*vtx1)[0])) >= (((*vtx1)[0]-tx) * ((*vtx0)[1]-(*vtx1)[1])) ) == yflag1 ) inside_flag ^= 1; }
		}

		return (inside_flag ? true : false);

#endif // 0

	}


	//********************************************************************************
	// Check if a ray of given origin and direction intersects the triangle.
	// Does not allow intersections on the back side of the triangle.
	//********************************************************************************

	Bool Triangle3::Intersects(const Ray3& ray, Bool twoSided) const
	{
		Vector3x point, normal;

		return (Intersects(ray, point, normal, twoSided));
	}


	//********************************************************************************
	// Check if a ray of given origin and direction intersects the triangle.
	// Does not allow intersections on the back side of the triangle.
	//********************************************************************************

	Bool Triangle3::Intersects(const Ray3& ray, Vector3x& point, Vector3x& normal, Bool twoSided) const
	{

#if 1

		if (_plane.Intersects(ray, point, normal, twoSided) == false)
		{
			return false;
		}

		Vector3x edgeNormal;

		for (S32 i = 0; i < 3; i++)
		{
			static const S32 next[3] =
			{ 1, 2, 0 };

			const Vector3x& v0 = _points[i];
			const Vector3x& v1 = _points[next[i]];

			edgeNormal.Cross(normal, v0 - v1);

			if ((point - v0) * edgeNormal < 0.0f)
			{
				return false;
			}
		}

		return true;

#endif // 1

#if 0

		// This code has been temporarily disabled as it produced erroneous results from
		// time to time.

		if (_plane.RayIntersects(ray, twoSided) == false)
		{
			return false;
		}

		Vector3x ABC = _points[0] - _points[1];
		Vector3x DEF = _points[0] - _points[2];
		Vector3x JKL = _points[0] - ray.Origin();

		const Vector3x& GHI = ray.Direction();

		F32 EIHF = DEF.Y() * GHI.Z() - GHI.Y() * DEF.Z();
		F32 GFDI = GHI.X() * DEF.Z() - DEF.X() * GHI.Z();
		F32 DHEG = DEF.X() * GHI.Y() - DEF.Y() * GHI.X();

		F32 denom = (ABC.X() * EIHF + ABC.Y() * GFDI + ABC.Z() * DHEG);

		if (denom == 0.0f)
		{
			return false;
		}

		F32 beta = (JKL.X() * EIHF + JKL.Y() * GFDI + JKL.Z() * DHEG) / denom;

		if (beta <= 0.0f || beta >= 1.0f)
		{
			return false;
		}

		F32 AKJB = ABC.X() * JKL.Y() - JKL.X() * ABC.Y();
		F32 JCAL = JKL.X() * ABC.Z() - ABC.X() * JKL.Z();
		F32 BLKC = ABC.Y() * JKL.Z() - JKL.Y() * ABC.Z();

		F32 gamma = (GHI.Z() * AKJB + GHI.Y() * JCAL + GHI.X() * BLKC) / denom;

		if (gamma <= 0.0f || beta + gamma >= 1.0f)
		{
			return false;
		}

		F32 t = -(DEF.Z() * AKJB + DEF.Y() * JCAL + DEF.X() * BLKC) / denom;

		if (t <= 0.0f)
		{
			return false;
		}

		point = ray.Origin() + (ray.Direction() * t);
		normal = _plane.Normal();

		return true;

#endif // 0

	}


	//********************************************************************************
	// Squared distance of point from triangle. Closest point is returned in
	// intersection.
	//********************************************************************************

	F32 Triangle3::DistanceSqr(const Vector3x& point, Vector3x& intersection) const
	{
		Vector3x edge0 = _points[1] - _points[0];
		Vector3x edge1 = _points[2] - _points[0];
		Vector3x diff = _points[0] - point;
		F32 A00 = edge0.LengthSqr();
		F32 A01 = edge0 * edge1;
		F32 A11 = edge1.LengthSqr();
		F32 B0 = diff * edge0;
		F32 B1 = diff * edge1;
		F32 C = diff.LengthSqr();
		F32 determinant = Abs(A00 * A11 - A01 * A01);
		F32 s = A01*B1-A11*B0;
		F32 t = A01*B0-A00*B1;
		F32 distanceSqr;

		if (s + t <= determinant)
		{
			if (s < 0.0f)
			{
				// region 4
				if (t < 0.0f)
				{
					if (B0 < 0.0f)
					{
						t = 0.0f;

						if (-B0 >= A00)
						{
							s = 1.0f;
							distanceSqr = A00 + 2.0f * B0 + C;
						}
						else
						{
							s = -B0 / A00;
							distanceSqr = B0 * s + C;
						}
					}
					else
					{
						s = 0.0f;

						if (B1 >= 0.0f)
						{
							t = 0.0f;
							distanceSqr = C;
						}
						else if (-B1 >= A11)
						{
							t = 1.0f;
							distanceSqr = A11 + 2.0f * B1 + C;
						}
						else
						{
							t = -B1 / A11;
							distanceSqr = B1 * t + C;
						}
					}
				}
				// region 3
				else
				{
					s = 0.0f;

					if (B1 >= 0.0f)
					{
						t = 0.0f;
						distanceSqr = C;
					}
					else if (-B1 >= A11)
					{
						t = 1;
						distanceSqr = A11 + 2.0f * B1 + C;
					}
					else
					{
						t = -B1 / A11;
						distanceSqr = B1 * t + C;
					}
				}
			}
			// region 5
			else if (t < 0.0f)
			{
				t = 0.0f;

				if (B0 >= 0.0f)
				{
					s = 0.0f;
					distanceSqr = C;
				}
				else if (-B0 >= A00)
				{
					s = 1.0f;
					distanceSqr = A00 + 2.0f * B0 + C;
				}
				else
				{
					s = -B0 / A00;
					distanceSqr = B0 * s + C;
				}
			}
			// region 0
			else
			{
				// minimum at interior point
				F32 invDeterminant = 1.0f / determinant;
				s *= invDeterminant;
				t *= invDeterminant;
				distanceSqr = s * (A00 * s + A01 * t + 2.0f * B0) + t * (A01 * s + A11 * t + 2.0f * B1) + C;
			}
		}
		else
		{
			F32 Tmp0, Tmp1, difference, denominator;

			// region 2
			if (s < 0.0f)
			{
				Tmp0 = A01 + B0;
				Tmp1 = A11 + B1;

				if (Tmp1 > Tmp0)
				{
					difference = Tmp1 - Tmp0;
					denominator = A00 - 2.0f * A01 + A11;

					if (difference >= denominator)
					{
						s = 1.0f;
						t = 0.0f;
						distanceSqr = A00 + 2.0f * B0 + C;
					}
					else
					{
						s = difference / denominator;
						t = 1.0f - s;
						distanceSqr = s * (A00 * s + A01 * t + 2.0f * B0) + t * (A01 * s + A11 * t + 2.0f * B1) + C;
					}
				}
				else
				{
					s = 0.0f;

					if (Tmp1 <= 0.0f)
					{
						t = 1.0f;
						distanceSqr = A11 + 2.0f * B1 + C;
					}
					else if (B1 >= 0.0f)
					{
						t = 0.0f;
						distanceSqr = C;
					}
					else
					{
						t = -B1 / A11;
						distanceSqr = B1 * t + C;
					}
				}
			}
			// region 6
			else if (t < 0.0f)
			{
				Tmp0 = A01 + B1;
				Tmp1 = A00 + B0;

				if (Tmp1 > Tmp0)
				{
					difference = Tmp1 - Tmp0;
					denominator = A00 - 2.0f * A01 + A11;

					if (difference >= denominator)
					{
						t = 1.0f;
						s = 0.0f;
						distanceSqr = A11 + 2.0f * B1 + C;
					}
					else
					{
						t = difference / denominator;
						s = 1.0f - t;
						distanceSqr = s * (A00 * s + A01 * t + 2.0f * B0) + t * (A01 * s + A11 * t + 2.0f * B1) + C;
					}
				}
				else
				{
					t = 0.0f;

					if (Tmp1 <= 0.0f)
					{
						s = 1.0f;
						distanceSqr = A00 + 2.0f * B0 + C;
					}
					else if (B0 >= 0.0f)
					{
						s = 0.0f;
						distanceSqr = C;
					}
					else
					{
						s = -B0 / A00;
						distanceSqr = B0 * s + C;
					}
				}
			}
			// region 1
			else
			{
				difference = A11 + B1 - A01 - B0;

				if (difference <= 0.0f)
				{
					s = 0.0f;
					t = 1.0f;
					distanceSqr = A11 + 2.0f * B1 + C;
				}
				else
				{
					denominator = A00 -2.0f * A01 + A11;

					if (difference >= denominator)
					{
						s = 1.0f;
						t = 0.0f;
						distanceSqr = A00 + 2.0f * B0 + C;
					}
					else
					{
						s = difference / denominator;
						t = 1.0f - s;
						distanceSqr = s * (A00 * s + A01 * t + 2.0f * B0) + t * (A01 * s + A11 * t + 2.0f * B1) + C;
					}
				}
			}
		}

		// Closest point
		intersection = _points[0] + s * edge0 + t * edge1;

		return Abs(distanceSqr);
	}


	//********************************************************************************
	// Distance of point from triangle. Closest point is returned in intersection.
	//********************************************************************************

	F32 Triangle3::Distance(const Vector3x& point, Vector3x& intersection) const
	{
		return libmaths::Sqrt(DistanceSqr(point, intersection));
	}

// namespace libmaths
}
