//********************************************************************************
//
// 3D vector class.
// Most commonly used for 3D work - in 3 and 4 component flavours (4th component
// is ignored but is present for performance reasons).
//
// Original author: Rob Baker.
// Current maintainer: Rob Baker.
//
//********************************************************************************


//********************************************************************************
// Include the required header files.
//********************************************************************************

#include "vector3.h"
#include "matrix44.h"
#include "spherical3lh.h"
#include "spherical3rh.h"


namespace libmaths
{

	//********************************************************************************
	// Assignment from a spherical coordinate, resulting in a right-handed vector.
	//********************************************************************************

	Vector3x& Vector3x::operator = (const Spherical3RH& p)
	{
		F32 cz, sz, ca, sa;

		CosSin(p.Azimuth(), ca, sa);
		CosSin(p.Inclination(), cz, sz);

		x = p.Radius() * sz * ca;
		y = p.Radius() * sz * sa;
		z = p.Radius() * cz;
		w = 0.0f;

		return *this;
	}


	//********************************************************************************
	// Assignment from a spherical coordinate, resulting in a right-handed vector.
	//********************************************************************************

	Vector3& Vector3::operator = (const Spherical3RH& p)
	{
		F32 cz, sz, ca, sa;

		CosSin(p.Azimuth(), ca, sa);
		CosSin(p.Inclination(), cz, sz);

		x = p.Radius() * sz * ca;
		y = p.Radius() * sz * sa;
		z = p.Radius() * cz;

		return *this;
	}


	//********************************************************************************
	// Assignment from a spherical coordinate, resulting in a left-handed vector.
	//********************************************************************************

	Vector3x& Vector3x::operator = (const Spherical3LH& p)
	{
		F32 cz, sz, ca, sa;

		CosSin(p.Azimuth(), ca, sa);
		CosSin(p.Inclination(), cz, sz);

		x = p.Radius() * sz * ca;
		z = -p.Radius() * sz * sa;
		y = -p.Radius() * cz;
		w = 0.0f;

		return *this;
	}


	//********************************************************************************
	// Assignment from a spherical coordinate, resulting in a left-handed vector.
	//********************************************************************************

	Vector3& Vector3::operator = (const Spherical3LH& p)
	{
		F32 cz, sz, ca, sa;

		CosSin(p.Azimuth(), ca, sa);
		CosSin(p.Inclination(), cz, sz);

		x = p.Radius() * sz * ca;
		z = -p.Radius() * sz * sa;
		y = -p.Radius() * cz;

		return *this;
	}


	//********************************************************************************
	// Get a new vector from this vector multiplied (transformed) by a quaternion.
	//********************************************************************************

	Vector3x Vector3x::operator * (const Quaternion& rhs) const
	{
		return Vector3x(Matrix44(rhs) * (*this));
	}


	//********************************************************************************
	// Get a new vector from this vector multiplied (transformed) by a quaternion.
	//********************************************************************************

	Vector3 Vector3::operator * (const Quaternion& rhs) const
	{
		return Vector3(Matrix44(rhs) * (*this));
	}


	//********************************************************************************
	// Multiply (transform) this vector by a quaternion.
	//********************************************************************************

	Vector3x& Vector3x::operator *= (const Quaternion& rhs)
	{
		(*this) *= Matrix44(rhs);

		return *this;
	}


	//********************************************************************************
	// Multiply (transform) this vector by a quaternion.
	//********************************************************************************

	Vector3& Vector3::operator *= (const Quaternion& rhs)
	{
		(*this) *= Matrix44(rhs);

		return *this;
	}

// namespace libmaths
}
